#include <config.h>
#include "gpu_block_solver_config.h"

module gpu_block_solver_class
  use iso_c_binding
  use comlib
  use lattice_class
  use error_class
  use field_gauge_class
  use field_fermion_class
  use quark_clover_class
  use gpu_bind_class
  implicit none
  private

  public :: gpu_block_solver
  public :: new,delete
  public :: assign_inv_mult_eoprec_wd
  public :: assign_inv_mult_Qee_shift
  public :: assign_inv_mult_Oee

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  integer, save, allocatable :: GPU_TASK_TABLE(:,:)
  integer, parameter :: D_GAUGE_SIZE = COL*COL*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)*NDIM
  integer, parameter :: D_FERMI_SIZE = COL*SPIN*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)

  integer, parameter :: S_GAUGE_SIZE = COL*(COL-1)*NTH*NZ*NY*NX*NDIM
  integer, parameter :: S_FERMI_SIZE = COL*SPIN*NTH*NZ*NY*NX

#ifdef _CLOVER
  integer, parameter :: D_CLV_SIZE = CLSPH*2*NTH*NZ*NY*NX
  integer, parameter :: S_CLV_SIZE = CLSPH*2*NTH*NZ*NY*NX
#endif

  type gpu_block_solver
    private
    complex(C_FLOAT), allocatable ::  swe(:,:)
    complex(C_FLOAT), allocatable ::  sve(:,:)
    complex(C_FLOAT), allocatable ::   su(:,:)
#ifdef _CLOVER
    complex(C_FLOAT), allocatable :: sclv(:,:)
#endif
    integer, private :: num_gpu
    integer, public :: ieo
    integer, public :: num_pf
    real(DP), public :: tol
    real(DP), public :: err_norm
    integer,  public :: iter
    integer(C_INT),  allocatable :: iter_all(:)
    integer,  public :: print_flag = 0
  end type

  interface new
    module procedure new_gpu_blk_slv_f
  end interface

  interface delete
    module procedure delete_gpu_blk_slv_f
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_blk_gpu
  end interface

  interface assign_inv_mult_Qee_shift
    module procedure assign_inv_mult_Qee_shift_blk_gpu
  end interface

  interface assign_inv_mult_Oee
    module procedure assign_inv_mult_Oee_blk_gpu
  end interface

  interface

!=================================================================
! for GPU# 1
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g1(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g1")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g1(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g1")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g1(print_flag) BIND(C,name="delete_gpu_solver_g1")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine gpu_blk_solver_g1(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g1")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

    subroutine gpu_blk_cg_g1(tol,iter,kappa,ye,dye,ieo,shift,do_shift) BIND(C,name="cg_g1")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: shift
      integer(C_INT),    intent(in)    :: do_shift
    end subroutine

    subroutine assign_inv_mult_Oee_g1(tol,iter,kappa,ye,dye,ieo) BIND(C,name="assign_inv_mult_Oee_g1")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

!=================================================================
! for GPU# 2
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g2(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g2")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g2(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g2")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g2(print_flag) BIND(C,name="delete_gpu_solver_g2")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine gpu_blk_solver_g2(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g2")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

    subroutine gpu_blk_cg_g2(tol,iter,kappa,ye,dye,ieo,shift,do_shift) BIND(C,name="cg_g2")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: shift
      integer(C_INT),    intent(in)    :: do_shift
    end subroutine

    subroutine assign_inv_mult_Oee_g2(tol,iter,kappa,ye,dye,ieo) BIND(C,name="assign_inv_mult_Oee_g2")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

!=================================================================
! for GPU# 3
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g3(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g3")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g3(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g3")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g3(print_flag) BIND(C,name="delete_gpu_solver_g3")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine gpu_blk_solver_g3(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g3")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

    subroutine gpu_blk_cg_g3(tol,iter,kappa,ye,dye,ieo,shift,do_shift) BIND(C,name="cg_g3")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: shift
      integer(C_INT),    intent(in)    :: do_shift
    end subroutine

    subroutine assign_inv_mult_Oee_g3(tol,iter,kappa,ye,dye,ieo) BIND(C,name="assign_inv_mult_Oee_g3")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

!=================================================================
! for GPU# 4
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g4(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g4")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g4(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g4")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g4(print_flag) BIND(C,name="delete_gpu_solver_g4")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine gpu_blk_solver_g4(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g4")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

    subroutine gpu_blk_cg_g4(tol,iter,kappa,ye,dye,ieo,shift,do_shift) BIND(C,name="cg_g4")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: shift
      integer(C_INT),    intent(in)    :: do_shift
    end subroutine

    subroutine assign_inv_mult_Oee_g4(tol,iter,kappa,ye,dye,ieo) BIND(C,name="assign_inv_mult_Oee_g4")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

!====================================================================
! precision and array conversion routines for GPU<=>CPU
!====================================================================
    subroutine conv_q_d2s(vec,fye,ieo,norm) BIND(C,name="conv_q_vec_d2s_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_FERMI_SIZE,S_FERMI_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: vec(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: fye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(inout) :: norm
    end subroutine

    subroutine conv_q_s2d(fye,vec,ieo,norm) BIND(C,name="conv_q_vec_s2d_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_FERMI_SIZE,S_FERMI_SIZE
      implicit none
      complex(C_FLOAT),  intent(in)    :: fye(S_FERMI_SIZE)
      complex(C_DOUBLE), intent(inout) :: vec(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: norm
    end subroutine

    subroutine conv_u(due,fue,ieo) BIND(C,name="conv_u_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_GAUGE_SIZE,S_GAUGE_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: due(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
      complex(C_FLOAT),  intent(inout) :: fue(S_GAUGE_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

#ifdef _CLOVER
    subroutine conv_clv(dclve,fclve,ieo) BIND(C,name="conv_clv_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_CLV_SIZE,S_CLV_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: dclve(CLSPH,2,NTH,NZ,NY,NX)
      complex(C_FLOAT),  intent(inout) :: fclve(S_CLV_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine
#endif

  end interface

contains

#ifdef _CLOVER
subroutine new_gpu_blk_slv_f(this,u,ieo,fclinv)
#else
subroutine new_gpu_blk_slv_f(this,u,ieo)
#endif
  use field_gauge_class
#ifdef _CLOVER
  use field_clover_class
#endif
  implicit none
  type(gpu_block_solver),  intent(inout) :: this
  type(vfield_gluon_wg),   intent(in)    :: u
  integer, intent(in)                    :: ieo
#ifdef _CLOVER
  type(field_clover_term), intent(in)    :: fclinv
  integer :: ieoxyz,ics
#endif
  complex(DP), allocatable :: uu(:,:,:,:,:,:,:)
  integer :: ic,jc,ith,iz,iy,ix,mu
  integer :: num_tot_gpus
  integer :: num_use_gpus
  integer :: ioe,igpu

#ifdef _DEBUG
  this%print_flag = 1
#else
  this%print_flag = 0
#endif

  this%iter = 0
  if (allocated(this%iter_all)) deallocate(this%iter_all)
  allocate(this%iter_all(this%num_pf))
  this%iter_all(:) = 0

  this%ieo = ieo
  ioe = 1-ieo

  if (allocated(this%su)) deallocate(this%su)
  allocate(this%su(S_GAUGE_SIZE,0:1))

  allocate(uu(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM))
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,ith,ic,jc)
  do mu=1,NDIM
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1
  do ith=0,NTH
  do jc=1,COL
  do ic=1,COL
    uu(ic,jc,ith,iz,iy,ix,mu) = u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc)
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  call conv_u(uu,this%su(:,ieo),ieo)

!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,ith,ic,jc)
  do mu=1,NDIM
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1
  do ith=0,NTH
  do jc=1,COL
  do ic=1,COL
    uu(ic,jc,ith,iz,iy,ix,mu) = u%eo(ioe)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc)
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  call conv_u(uu,this%su(:,ioe),ioe)
  deallocate(uu)

#ifdef _CLOVER
  if (allocated(this%sclv)) deallocate(this%sclv)
  allocate(this%sclv(S_CLV_SIZE,0:1))
  call conv_clv(fclinv%eo(ieo)%l,this%sclv(:,ieo),ieo)
  call conv_clv(fclinv%eo(ioe)%l,this%sclv(:,ioe),ioe)
#endif

  this%num_gpu = _NUM_GPUS
  call bind_cpu_to_gpu(num_tot_gpus,this%num_gpu)
  gpu_mult_count(:) = 0
  call set_gpu_task_table(this%num_pf,this%num_gpu)

!$OMP PARALLEL DO PRIVATE(igpu)
  do igpu=1,this%num_gpu
    call switch_gpu_device(igpu-1)
#ifdef _CLOVER
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g1(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g2(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g3(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g4(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    end select
#else
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g1(this%su(:,ieo),this%su(:,ioe),ieo)
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g2(this%su(:,ieo),this%su(:,ioe),ieo)
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g3(this%su(:,ieo),this%su(:,ioe),ieo)
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call new_gpu_blk_solver_g4(this%su(:,ieo),this%su(:,ioe),ieo)
    end select
#endif
  enddo

  if (allocated(this%su)) deallocate(this%su)
#ifdef _CLOVER
  if (allocated(this%sclv)) deallocate(this%sclv)
#endif

  return
end subroutine

subroutine delete_gpu_blk_slv_f(this)
  implicit none
  type(gpu_block_solver), intent(inout) :: this
  integer :: igpu
  gpu_mult_count(:) = 0
  if (allocated(this%iter_all)) deallocate(this%iter_all)
  if (allocated(this%su))  deallocate(this%su)
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
#ifdef _CLOVER
  if (allocated(this%sclv)) deallocate(this%sclv)
#endif
!$OMP PARALLEL DO PRIVATE(igpu)
  do igpu=1,this%num_gpu
    call switch_gpu_device(igpu-1)
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call delete_gpu_blk_solver_g1(this%print_flag)
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call delete_gpu_blk_solver_g2(0)
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call delete_gpu_blk_solver_g3(0)
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) &
 &      call delete_gpu_blk_solver_g4(0)
    end select
  enddo
  return
end subroutine

subroutine set_gpu_task_table(num_pf,ngpu)
!
! set GPU and PF-field assignment table
!
  implicit none
  integer, intent(in) :: num_pf
  integer, intent(in) :: ngpu
  integer :: ipf,igpu
  external :: get_num_gpu   ! function defined in GPUSolverLIB(CUDA function)

  if (num_pf <= 0) then
    call error_stop("Number of PF field should be positive.")
  endif

  if (allocated(GPU_TASK_TABLE)) deallocate(GPU_TASK_TABLE)
  allocate(GPU_TASK_TABLE(ngpu,3))

  GPU_TASK_TABLE(:,:)=0

  !=======================================
  ! count ps-fields for each GPU.
  !=======================================
  do ipf=1,num_pf
    igpu = mod(ipf-1,ngpu)+1
    GPU_TASK_TABLE(igpu,3) = GPU_TASK_TABLE(igpu,3) + 1
  enddo

  igpu=1
  if (GPU_TASK_TABLE(igpu,3) > 0) then
    GPU_TASK_TABLE(igpu,1) = 1
    GPU_TASK_TABLE(igpu,2) = GPU_TASK_TABLE(igpu,1)+GPU_TASK_TABLE(igpu,3)-1
  endif
  do igpu=2,ngpu
    if (GPU_TASK_TABLE(igpu,3) > 0) then
      GPU_TASK_TABLE(igpu,1) = GPU_TASK_TABLE(igpu-1,2)+1
      GPU_TASK_TABLE(igpu,2) = GPU_TASK_TABLE(igpu,1)+GPU_TASK_TABLE(igpu,3)-1
    endif
  enddo

#ifdef _DEBUG
  if (0==nodeid) then
    write(*,'("# GPU START_PF END_PF NUM_PF")')
    do igpu=1,ngpu
      write(*,'("# ",I3,2X,3I7)')igpu,GPU_TASK_TABLE(igpu,1),&
 &                                    GPU_TASK_TABLE(igpu,2),&
 &                                    GPU_TASK_TABLE(igpu,3)
    enddo
  endif
#endif

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_blk_gpu(this,kappa,x,y)
!
! x = D \ y
!
  use field_fermion_class, only : mult_iter
  use counter_class
  implicit none
  type(gpu_block_solver), intent(inout) :: this
  real(DP),    intent(in)    :: kappa
  complex(DP), intent(inout) :: x(:,:),y(:,:)

  real(C_DOUBLE), allocatable :: norm(:)
  real(C_DOUBLE) :: tol
  complex(DP), allocatable :: vv(:)
  integer :: igpu,ipf,ieo,num_pf,j
  integer :: isite

  ieo = this%ieo
  num_pf = this%num_pf
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
#ifdef _DEBUG
  write(*,'("% target_tol, residual, gpu_tol:",3E24.15)')this%tol,this%err_norm,tol
#endif

  allocate(norm(num_pf))
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  allocate(this%swe(S_FERMI_SIZE,num_pf))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
!$OMP PARALLEL DO PRIVATE(isite)
    do isite=1,S_FERMI_SIZE
      vv(isite) = y(ipf,isite)
    enddo
!!$OMP PARALLEL WORKSHARE
!    vv(:) = y(ipf,:)
!!$OMP END PARALLEL WORKSHARE
    call conv_q_vec_d2s(vv,this%swe(:,ipf),ieo,norm(ipf))
  enddo
  deallocate(vv)

  allocate(this%sve(S_FERMI_SIZE,num_pf))

  !        v
  !  sve = Dee \ swe
  !
!$OMP PARALLEL DO PRIVATE(igpu,ipf)
  do igpu=1,this%num_gpu
    call switch_gpu_device(igpu-1)
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call gpu_blk_solver_g1(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call gpu_blk_solver_g2(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call gpu_blk_solver_g3(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call gpu_blk_solver_g4(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    end select
  enddo

  deallocate(this%swe)

  do igpu=1,this%num_gpu
    call inc(mult_iter,gpu_mult_count(igpu))
  enddo
  gpu_mult_count(:) = 0
  this%iter = SUM(this%iter_all(:))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
    call conv_q_vec_s2d(this%sve(:,ipf),vv,ieo,norm(ipf))
!$OMP PARALLEL DO PRIVATE(isite)
    do isite=1,S_FERMI_SIZE
      x(ipf,isite) = vv(isite)
    enddo
!!$OMP PARALLEL WORKSHARE
!    x(ipf,:) = vv(:)
!!$OMP END PARALLEL WORKSHARE
  enddo
  deallocate(vv)

  deallocate(this%sve)
  deallocate(norm)

  return
end subroutine

subroutine assign_inv_mult_Qee_shift_blk_gpu(this,kappa,x,y,shift)
!
! x = (Qee - shift) \ y
!
  use field_fermion_class, only : mult_iter
  use counter_class
  implicit none
  type(gpu_block_solver),  intent(inout) :: this
  real(DP),                intent(in)    :: kappa
  type(field_quark_eo_wg), intent(inout) :: y(:)
  type(field_quark_eo_wg), intent(inout) :: x(:)
  real(DP), optional,      intent(in)    :: shift(:)

  real(C_DOUBLE), allocatable :: norm(:)
  real(C_DOUBLE) :: tol
  real(C_DOUBLE) :: dummy
  logical :: is_shift
  complex(DP), allocatable :: vv(:)
  integer :: igpu,ipf,ieo,num_pf

  is_shift = present(shift)
  dummy = 0.0_DP

  num_pf = this%num_pf

  do ipf=1,num_pf
    ieo = y(ipf)%ieo
    call new(x(ipf),ieo)
  enddo
  
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
#ifdef _DEBUG
  write(*,'("% target_tol, residual, gpu_tol:",3E24.15)')this%tol,this%err_norm,tol
  if (is_shift) then
    write(*,'("% with shift")')
  else
    write(*,'("% without shift")')
  endif
#endif

  allocate(norm(num_pf))
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  allocate(this%swe(S_FERMI_SIZE,num_pf))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
    call pack(y(ipf),vv)
    call conv_q_vec_d2s(vv,this%swe(:,ipf),ieo,norm(ipf))
  enddo
  deallocate(vv)

  allocate(this%sve(S_FERMI_SIZE,num_pf))

  !         
  !  sve = (Qee - shift) \ swe
  !
!$OMP PARALLEL DO PRIVATE(igpu,ipf)
  do igpu=1,this%num_gpu
    call switch_gpu_device(igpu-1)
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          if (is_shift) then
            call gpu_blk_cg_g1(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,shift(ipf),1)
          else
            call gpu_blk_cg_g1(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,dummy,0)
          endif
        enddo
      endif
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          if (is_shift) then
            call gpu_blk_cg_g2(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,shift(ipf),1)
          else
            call gpu_blk_cg_g2(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,dummy,0)
          endif
        enddo
      endif
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          if (is_shift) then
            call gpu_blk_cg_g3(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,shift(ipf),1)
          else
            call gpu_blk_cg_g3(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,dummy,0)
          endif
        enddo
      endif
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          if (is_shift) then
            call gpu_blk_cg_g4(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,shift(ipf),1)
          else
            call gpu_blk_cg_g4(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo,dummy,0)
          endif
        enddo
      endif
    end select
  enddo

  deallocate(this%swe)

  do igpu=1,this%num_gpu
    call inc(mult_iter,gpu_mult_count(igpu))
  enddo
  gpu_mult_count(:) = 0
  this%iter = SUM(this%iter_all(:))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
    call conv_q_vec_s2d(this%sve(:,ipf),vv,ieo,norm(ipf))
    call unpack(vv,x(ipf))
  enddo
  deallocate(vv)

  deallocate(this%sve)
  deallocate(norm)

  return
end subroutine

subroutine assign_inv_mult_Oee_blk_gpu(this,kappa,x,y)
!
! x(j) = Oee \ y(j)
!
!             v   v
! with  Oee = Dee Dee^dag
!   v
!   Dee = 1 - kappa^2 Fee Meo Foo Moe
!
  use field_fermion_class, only : mult_iter
  use counter_class
  implicit none
  type(gpu_block_solver),  intent(inout) :: this
  real(DP),                intent(in)    :: kappa
  type(field_quark_eo_wg), intent(inout) :: x(:)
  type(field_quark_eo_wg), intent(inout) :: y(:)

  real(C_DOUBLE), allocatable :: norm(:)
  real(C_DOUBLE) :: tol
  real(C_DOUBLE) :: dummy
  complex(DP), allocatable :: vv(:)
  integer :: igpu,ipf,ieo,num_pf

  num_pf = this%num_pf

  do ipf=1,num_pf
    ieo = y(ipf)%ieo
    call new(x(ipf),ieo)
  enddo
  
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
#ifdef _DEBUG
  write(*,'("% target_tol, residual, gpu_tol:",3E24.15)')this%tol,this%err_norm,tol
  write(*,'("% Blocked Solver for Oee xe(j) = be(j).")')
#endif

  allocate(norm(num_pf))
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  allocate(this%swe(S_FERMI_SIZE,num_pf))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
    call pack(y(ipf),vv)
    call conv_q_vec_d2s(vv,this%swe(:,ipf),ieo,norm(ipf))
  enddo
  deallocate(vv)

  allocate(this%sve(S_FERMI_SIZE,num_pf))

  !         
  !  sve = Oee \ swe
  !
!$OMP PARALLEL DO PRIVATE(igpu,ipf)
  do igpu=1,this%num_gpu
    call switch_gpu_device(igpu-1)
    select case(igpu)
    case(1)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call assign_inv_mult_Oee_g1(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(2)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call assign_inv_mult_Oee_g2(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(3)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call assign_inv_mult_Oee_g3(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    case(4)
      if (0 < GPU_TASK_TABLE(igpu,3)) then
        do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
          call assign_inv_mult_Oee_g4(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
        enddo
      endif
    end select
  enddo

  deallocate(this%swe)

  do igpu=1,this%num_gpu
    call inc(mult_iter,gpu_mult_count(igpu))
  enddo
  gpu_mult_count(:) = 0
  this%iter = SUM(this%iter_all(:))

  allocate(vv(S_FERMI_SIZE))
  do ipf=1,num_pf
    call conv_q_vec_s2d(this%sve(:,ipf),vv,ieo,norm(ipf))
    call unpack(vv,x(ipf))
  enddo
  deallocate(vv)

  deallocate(this%sve)
  deallocate(norm)

  return
end subroutine

end module
