void
cuda_assign_inv_mult_Qee_shift_cg(const double   *kappa,
                                  const double   *tol,
                                  int            *iter,
                                  f_wqf_eo       *yde,
                                  f_wqf_eo       *ye,
                                  const double   *shift = NULL) {
//
// MAIN GPU CGNE Solver for
//
//  (Qee - shift) yde = ye
//
//  with Hermitian-squared even/odd site preconditioned Wilson-Clover-Dirac opeartor
//        v       v
//  Qee = Dee^dag Dee
//    v
//    Dee = 1 - kappa^2 Fee Meo Foo Moe
//

  //===========================================================
  // host variables
  //===========================================================
  const int niter = 5000;
  int ieo;
  float err;
  float mkappa2 = -(float)((*kappa)*(*kappa));
  float    tol2 =  (float)((*tol)*(*tol));
  float sshift;
  float alp, nalp, bet, rho0, rho1, rtmp0;
  cuComplex ctmp0;

  //===========================================================
  // device variables
  //===========================================================
  float_gpu  mkappa2d;
  float_gpu  shiftd;
  float_gpu  alpd, nalpd, betd;
  cu_wqf_eo *yed,*yded,*ped,*qed,*red;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
  }

  startTimer(solv_timer);

#ifdef _DEBUG_
  fflush(stdout);
  if (is_clover) {
    printf("%%S_CG_CLOVER_QEE_SHIFT(MPI+GPU-PARALLEL)\n");
  } else {
    printf("%%S_CG_WILSON_QEE_SHIFT(MPI+GPU-PARALLEL)\n");
  }
  fflush(stdout);
#endif

   ieo = (*ye).ieo;

  yded = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   yed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ped = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   qed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   red = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);

  (*yded).ieo = ieo;
   (*yed).ieo = ieo;
   (*ped).ieo = ieo;
   (*qed).ieo = ieo;
   (*red).ieo = ieo;

  (*yded).ipeo = _ipeo_;
   (*yed).ipeo = _ipeo_;
   (*ped).ipeo = _ipeo_;
   (*qed).ipeo = _ipeo_;
   (*red).ipeo = _ipeo_;

  //======================================
  // allocate device memory
  //======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaMalloc((void**)    &(alpd.g[igx]),sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nalpd.g[igx]),sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(betd.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&(mkappa2d.g[igx]),sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&(  shiftd.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**) &((*yded).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*yed).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*ped).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*qed).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*red).g[igx]),sizeof(cu_wqf_eo_blk)));

    new_q((*yded).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*yed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*ped).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*qed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*red).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
  }

#ifdef _DEBUG_
  printf("%% cudaMalloc OK.\n");
#endif

//======================================
// Send data from host to device
//======================================
  startTimer(ysend_timer);
  assign(mkappa2d,mkappa2);
  if ( NULL != shift ) {
    sshift = (float)(*shift);
    assign(  shiftd, sshift);
  }
  assign(yed,ye);
  stopTimer(ysend_timer);

///////////////////////////////////////////////////////////////////////////////
  //
  // set x_0 = 0 (yde=0)
  //
  clear(yded,ieo);

  //
  // r  = b ( = y : rhs of equation )
  // rho0 = |re|^2
  // p  = r
  //
  assign(red,yed);
  assign(ped,red);
  rtmp0 = abs2(red);
  solv_flop_count += QABS2;
  rho0 = rtmp0;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%% S_TOL=%24.15e\n",sqrt(tol2));
  if ( 0 == _nodeid_ ) printf("@%9d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    if ( NULL != shift ) {

      //
      // q = (Qee - shift) p 
      //
      if (is_clover) {
        assign_mult_Qee_shift(mkappa2d,shiftd,qed,ped,ud,fclinvd);
      } else {
        assign_mult_Qee_shift(mkappa2d,shiftd,qed,ped,ud);
      }
      solv_flop_count += MULT_MB_PRE*2 + QACCUM_ADD_RMULT;

    } else {

      //
      // q = Qee p 
      //
      if (is_clover) {
        assign_mult_Qee(      mkappa2d,       qed,ped,ud,fclinvd);
      } else {
        assign_mult_Qee(      mkappa2d,       qed,ped,ud);
      }
      solv_flop_count += MULT_MB_PRE*2;

    }
    *iter = *iter + 2;

    //
    // alp = rho0/<p|q>
    //
    ctmp0 = prod(ped,qed);
    rtmp0 = ctmp0.x;
     alp = rho0/rtmp0;
    nalp = -alp;
    assign( alpd, alp);
    assign(nalpd,nalp);
    solv_flop_count += QPROD;

    //
    // x = x + alp p
    // r = r - alp q
    //
    accum_add_mult(yded, alpd,ped);
    accum_add_mult( red,nalpd,qed);
    solv_flop_count += QACCUM_ADD_RMULT*2;

    rtmp0 = abs2(red);
    solv_flop_count += QABS2;
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%9d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < tol2) break;

    //
    // rho1 = |r|^2
    // bet  = rho1/rho0
    // rho0 = rho1
    //
    rho1 = rtmp0;
    bet  = rho1/rho0;
    rho0 = rho1;
    assign( betd, bet);

    // p = p * bet + r
    accum_mult_add(ped, betd , red);
    solv_flop_count += QACCUM_RMULT_ADD;

  } // end while

  //======================================
  // check iteration counts (max iter)
  // if iter exceeds max iteration
  // flip the sign.
  //======================================
  if (*iter >= niter) {
    rtmp0 = abs2(red);
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    if ( 0 == _nodeid_ ) printf("@%s %d : Maximum iteration limit was reached. Increase niter=%d in %s.\n",__FILE__,__LINE__,niter,__FILE__);
    fflush(stdout);
    *iter = -(*iter);
  }

  //======================================
  // receive result from device to host
  //======================================
  startTimer(yrecv_timer);
  assign(yde,yded);
  stopTimer(yrecv_timer);

#ifdef _DEBUG_
  //
  // Check residual
  //
  // ped = (Qee - shift) yded
  //
  if (is_clover) {

    if ( NULL != shift ) {
      assign_mult_Qee_shift(mkappa2d,shiftd,ped,yded,ud,fclinvd);
    } else {
      assign_mult_Qee(      mkappa2d,       ped,yded,ud,fclinvd);
    }

  } else {

    if ( NULL != shift ) {
      assign_mult_Qee_shift(mkappa2d,shiftd,ped,yded,ud);
    } else {
      assign_mult_Qee(      mkappa2d,       ped,yded,ud);
    }

  }
  //
  // red = yed - ped
  //
  assign_sub(red,yed,ped);

  rtmp0 = abs2(red);
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%%%9d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

//======================================
// deallocate device memory
//======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaFree(    alpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nalpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    betd.g[igx]));

    CUDA_SAFE_CALL(cudaFree(mkappa2d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  shiftd.g[igx]));

    CUDA_SAFE_CALL(cudaFree(  (*yed).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*yded).g[igx]));

    CUDA_SAFE_CALL(cudaFree(  (*ped).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*qed).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*red).g[igx]));

  }

  _mm_free(yded);
  _mm_free( yed);
  _mm_free( ped);
  _mm_free( qed);
  _mm_free( red);

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solv_timer);
}

extern "C" void
cuda_assign_inv_mult_Qee_cg_(const double   *kappa,
                             const double   *tol,
                             int            *iter,
                             f_wqf_eo       *yde,
                             f_wqf_eo       *ye) {
//
// yde = Qee \ ye
//
// with 
//         v       v
//   Qee = Dee^dag Dee
//
  cuda_assign_inv_mult_Qee_shift_cg(kappa,tol,iter,yde,ye);
}

extern "C" void
cuda_assign_inv_mult_Qee_shift_cg_(const double   *kappa,
                                   const double   *tol,
                                   int            *iter,
                                   f_wqf_eo       *yde,
                                   f_wqf_eo       *ye,
                                   const double   *shift) {
//
// yde = (Qee - shift) \ ye
//
// with 
//         v       v
//   Qee = Dee^dag Dee
//
  cuda_assign_inv_mult_Qee_shift_cg(kappa,tol,iter,yde,ye,shift);
}
