#include <cuda.h>
#include <omp.h>
#ifdef _QMX_
#include <qmx_lib.h>
#else
#include <mpi.h>
#endif

#include <cassert>
#include <mm_malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cutil.h>

#include "cconfig.h"
#include "cgpu_config.h"
#include "cgpu_solver_config.h"
#include "cudaTimer.h"
#include "cuparam-size.h"
#include "cuda_bind_cpu_to_gpu.h"
#include "cuMyComplex.h"
#include "flop_table.h"

namespace cuda_gpu_solver_class {

//
// Fortran Module Variables
//
#include "cufortran_vars.h"

//
// Texture fetching for the link variables and
// the input fermion fields.
//
#ifdef _USE_TEXFETCH_
static texture<float4,1,cudaReadModeElementType> tex_ueri;
static texture<float4,1,cudaReadModeElementType> tex_uori;
static texture<float4,1,cudaReadModeElementType> tex_yori;
#endif
static texture<float2,1,cudaReadModeElementType> tex_Complex;
static texture<float ,1,cudaReadModeElementType> tex_Float;

//
// list vectors
//
// HOST side
static int4  phys_blk[_PHYS_BLOCK_];
static int4    in_blk[_IN_BLOCK_];
static int4   out_blk[_OUT_BLOCK_];
// GPU side
static int4_gpu phys_blk_d;
static int4_gpu   in_blk_d;
static int4_gpu  out_blk_d;

//
// parallel nodeid 4-D index
//
static int4 ipsite;

//
// MPI_Allreduce for reduction operations
//
#ifndef _singlePU
#ifdef _QMX_
static void sumcast(float     *f){
//  float ff = *f;
  qmx_sumcast_float(1,f);
}
static void sumcast(cuComplex *c){  
  float f[2];
  f[0] = (*c).x;
  f[1] = (*c).y;
  qmx_sumcast_float(2,f);
  (*c).x = f[0];
  (*c).y = f[1];
}
#else
static void sumcast(float     *f){
  float f2;
//  float ff = *f;
  MPI_Allreduce(f,&f2,1,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
  *f=f2;
//  printf("%%SUMCAST_FLOAT[%d]%14.6e =>%14.6e\n",_nodeid_,ff,*f);
}

static void sumcast(cuComplex *c){
  cuComplex c2;
//  cuComplex cc = *c;
  MPI_Allreduce(c,&c2,2,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
  *c=c2;
//  printf("%%SUMCAST_CMPLX[%d]%14.6e%14.6e =>%14.6e%14.6e\n",_nodeid_,cc.x,cc.y,(*c).x,(*c).y);
}
#endif
#endif

//
// Timer names
//
static cudaTimer_t ycopy_timer;
static cudaTimer_t gsend_timer;
static cudaTimer_t ysend_timer;
static cudaTimer_t yrecv_timer;
static cudaTimer_t  solv_timer;
static cudaTimer_t  mult_timer;
static cudaTimer_t  prec_timer;
static cudaTimer_t  binv_timer;

//
// flop/copy counters
//
static double ycopy_count;  // counts MPI_sendrecv calls
static double ycopy_dsize;  // counts MPI_sendrecv total data size
static double solv_flop_count;  // counts total flop in the solver
static double mult_flop_count;  // counts total flop for mult operation

//
// kernel switch
//
static bool is_clover = false;

#include "YMACRO.h"
#include "cuMatVec.h"
#include "cuWqfAlgebra.h"
#include "cu_mult_forw_t.h"
#include "cu_mult_forw_z.h"
#include "cu_mult_forw_y.h"
#include "cu_mult_forw_x.h"
#include "cu_mult_back_t.h"
#include "cu_mult_back_z.h"
#include "cu_mult_back_y.h"
#include "cu_mult_back_x.h"
#include "cu_mult_clover_term.h"
#include "copy_boundary.h"
#include "assign_mult_hop_kernel.h"
#include "assign_mult_hop_in_kernel.h"
#include "assign_mult_gamma5.h"
#include "assign_mult_clover_term.h"
#include "assign_mult_eoprec_wd_eo.h"
#include "assign_mult_Qee_shift.h"
#include "assign_inv_mult_eoprec_wd_ras.h"

//
// global device memory for LINK/CLOVER fields.
//
static cu_gvf    *ud      = NULL;
static cu_clvf   *fclinvd = NULL;

void create_list_vectors(){

  //
  // create list vector for physical sites 
  //
  int jblk=0;
  for(int ibx=0;ibx<_NBX;ibx++) {
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    phys_blk[jblk].x = ibx;
    phys_blk[jblk].y = iby;
    phys_blk[jblk].z = ibz;
    phys_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }

  //
  // create list vector for physical bulk(internal) sites
  //
  jblk=0;
  for(int ibx=1;ibx<_NBX-1;ibx++) {
  for(int iby=1;iby<_NBY-1;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    in_blk[jblk].x = ibx;
    in_blk[jblk].y = iby;
    in_blk[jblk].z = ibz;
    in_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }

  //
  // create list vector for physical surface(edge) sites
  //
  jblk=0;
  int ibx=0;
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
  for(    ibx=1;ibx<_NBX-1;ibx++) {
  for(int iby=0;iby<_NBY;iby=iby+_NBY-1) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }
  ibx=_NBX-1;
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
#ifdef _CHECK_LIST_VECTOR_
  if ( 0 == _nodeid_ ) {
    for (int iblk=0;iblk<_PHYS_BLOCK_;iblk++) {
      printf("PHYS:%6d%6d%3d%3d%3d\n",iblk,phys_blk[iblk].w,phys_blk[iblk].z,phys_blk[iblk].y,phys_blk[iblk].x);
    }
    for (int iblk=0;iblk<_IN_BLOCK_;iblk++) {
      printf("IN:%6d%6d%3d%3d%3d\n",iblk,in_blk[iblk].w,in_blk[iblk].z,in_blk[iblk].y,in_blk[iblk].x);
    }
    for (int iblk=0;iblk<_OUT_BLOCK_;iblk++) {
      printf("OUT:%6d%6d%3d%3d%3d\n",iblk,out_blk[iblk].w,out_blk[iblk].z,out_blk[iblk].y,out_blk[iblk].x);
    }
  }
#endif
  fflush(stdout);
}

static void 
new_gpu_solver_cuda(f_gvf  *u,
                    f_clvf *fclinv = NULL) {

#ifdef _DEBUG_
  printf("%%      nodeid=%d\n",_nodeid_);
  printf("%%        ipeo=%d\n",_ipeo_);
  for (int mu=0;mu < NDIM-1; mu++) printf("%%nodeidup[%d]=%d\n",mu,_nodeidup_(mu));
  for (int mu=0;mu < NDIM-1; mu++) printf("%%nodeiddn[%d]=%d\n",mu,_nodeiddn_(mu));
  fflush(stdout);
#endif

#ifdef _DEBUG_
  if ( 0 == _nodeid_ ) {
    printf("%%SIZEOF(CHAR)    : %ld\n",sizeof(char));
    printf("%%SIZEOF(SHORT)   : %ld\n",sizeof(short));
    printf("%%SIZEOF(INT)     : %ld\n",sizeof(int));
    printf("%%SIZEOF(FLOAT)   : %ld\n",sizeof(float));
    printf("%%SIZEOF(LINKeo)  : %ld\n",sizeof(cu_gvf_eo_blk));
    printf("%%SIZEOF(CLOVEReo): %ld\n",sizeof(cu_clvf_eo_blk));
    fflush(stdout);
  }
#endif

  if ( 0 == _nodeid_ ) fflush(stdout);

//
// Initialize GPU solver and
//   send LINK/CLOVER data from HOST to GPU
//

  //
  // Initialize timers and counters.
  //
  createTimer(&ycopy_timer);
  createTimer(&gsend_timer);
  createTimer(&ysend_timer);
  createTimer(&yrecv_timer);
  createTimer( &solv_timer);
  createTimer( &mult_timer);
  createTimer( &prec_timer);
  createTimer( &binv_timer);

  ycopy_count     = 0.0;
  ycopy_dsize     = 0.0;
  solv_flop_count = 0.0;
  mult_flop_count = 0.0;

  //
  // create list vectors
  //
  create_list_vectors();

  //
  // create ghost site copy buffers
  //
  new_copy_boundary();

  startTimer(gsend_timer);
  //
  // create device memory and send LINK/CLOVER data to GPU
  //
  ud = (cu_gvf *)_mm_malloc(sizeof(cu_gvf),16);
  if ( NULL == ud ) {
    printf(" _mm_malloc for ud failed.: %s at %d.\n",__FILE__,__LINE__);
  }
  (*ud).eo[0].ipeo = _ipeo_;
  (*ud).eo[1].ipeo = _ipeo_;
  (*ud).eo[0].ieo = 0;
  (*ud).eo[1].ieo = 1;
  if (is_clover) {
    fclinvd = (cu_clvf *)_mm_malloc(sizeof(cu_clvf),16);
    if ( NULL == fclinvd ) {
      printf(" _mm_malloc for fclinvd failed.: %s at %d.\n",__FILE__,__LINE__);
    }
    (*fclinvd).eo[0].ipeo = _ipeo_;
    (*fclinvd).eo[1].ipeo = _ipeo_;
    (*fclinvd).eo[0].ieo = 0;
    (*fclinvd).eo[1].ieo = 1;
  }

  ipsite.x = _ipsite_x_ ;
  ipsite.y = _ipsite_y_ ;
  ipsite.z = _ipsite_z_ ;
  ipsite.w = _ipsite_t_ ;

  ipeo_d.cpu_val   = _ipeo_;
  ieo_d[0].cpu_val = 0;
  ieo_d[1].cpu_val = 1;
  ipsite_d.cpu_val = ipsite;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    //
    // constants
    //
    CUDA_SAFE_CALL(cudaMalloc((void**)&(ipsite_d.g[igx]),sizeof(int4)));
    CUT_CHECK_ERROR("cudaMalloc ipsite_d.g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&(  ipeo_d.g[igx]),sizeof(int)));
    CUT_CHECK_ERROR("cudaMalloc ipeo_d.g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&(   igx_d.g[igx]),sizeof(int)));
    CUT_CHECK_ERROR("cudaMalloc igx_d.g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&(ieo_d[0].g[igx]),sizeof(int)));
    CUT_CHECK_ERROR("cudaMalloc ieo_d[0].g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&(ieo_d[1].g[igx]),sizeof(int)));
    CUT_CHECK_ERROR("cudaMalloc ieo_d[1].g[igx] faild");

    CUDA_SAFE_CALL(cudaMemcpy(ipeo_d.g[igx],&_ipeo_,sizeof(int),cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy( igx_d.g[igx],   &igx,sizeof(int),cudaMemcpyHostToDevice));
    int EVN = 0;
    int ODD = 1;
    CUDA_SAFE_CALL(cudaMemcpy(ieo_d[0].g[igx],   &EVN,sizeof(int) ,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(ieo_d[1].g[igx],   &ODD,sizeof(int) ,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(ipsite_d.g[igx],&ipsite,sizeof(int4),cudaMemcpyHostToDevice));

    //
    // list vectors
    //
    CUDA_SAFE_CALL(cudaMalloc((void**)&(phys_blk_d.g[igx]),sizeof(int4)*_PHYS_BLOCK_));
    CUT_CHECK_ERROR("cudaMalloc phys_blk_d.g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&(  in_blk_d.g[igx]),sizeof(int4)*_IN_BLOCK_));
    CUT_CHECK_ERROR("cudaMalloc in_blk_d.g[igx] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&( out_blk_d.g[igx]),sizeof(int4)*_OUT_BLOCK_));
    CUT_CHECK_ERROR("cudaMalloc out_blk_d.g[igx] faild");

    CUDA_SAFE_CALL(cudaMemcpy(phys_blk_d.g[igx],phys_blk,sizeof(int4)*_PHYS_BLOCK_,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(  in_blk_d.g[igx],  in_blk,sizeof(int4)*_IN_BLOCK_  ,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy( out_blk_d.g[igx], out_blk,sizeof(int4)*_OUT_BLOCK_ ,cudaMemcpyHostToDevice));

    //
    // link fields
    //
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*ud).eo[0].g[igx]),sizeof(cu_gvf_eo_blk)));
    CUT_CHECK_ERROR("cudaMalloc ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*ud).eo[1].g[igx]),sizeof(cu_gvf_eo_blk)));
    CUT_CHECK_ERROR("cudaMalloc ud.eo[1] faild");

    CUDA_SAFE_CALL(cudaMemcpy((*ud).eo[0].g[igx],&((*u).eo[0].g[igx]),sizeof(cu_gvf_eo_blk),cudaMemcpyHostToDevice));
    CUT_CHECK_ERROR("cudaMemcpy ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaMemcpy((*ud).eo[1].g[igx],&((*u).eo[1].g[igx]),sizeof(cu_gvf_eo_blk),cudaMemcpyHostToDevice));
    CUT_CHECK_ERROR("cudaMemcpy ud.eo[1] faild");

    new_u((*ud).eo[0].g[igx],ipeo_d.g[igx],ieo_d[0].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_u((*ud).eo[1].g[igx],ipeo_d.g[igx],ieo_d[1].g[igx],igx_d.g[igx],ipsite_d.g[igx]);

    //
    // clover fields
    //
    if (is_clover) {
      CUDA_SAFE_CALL(cudaMalloc((void**)&((*fclinvd).eo[0].g[igx]),sizeof(cu_clvf_eo_blk)));
      CUT_CHECK_ERROR("cudaMalloc fclinvd.eo[0] faild");

      CUDA_SAFE_CALL(cudaMalloc((void**)&((*fclinvd).eo[1].g[igx]),sizeof(cu_clvf_eo_blk)));
      CUT_CHECK_ERROR("cudaMalloc fclinvd.eo[1] faild");

      CUDA_SAFE_CALL(cudaMemcpy((*fclinvd).eo[0].g[igx],&((*fclinv).eo[0].g[igx]),sizeof(cu_clvf_eo_blk),cudaMemcpyHostToDevice));
      CUT_CHECK_ERROR("cudaMemcpy fclinvd.eo[0] faild");

      CUDA_SAFE_CALL(cudaMemcpy((*fclinvd).eo[1].g[igx],&((*fclinv).eo[1].g[igx]),sizeof(cu_clvf_eo_blk),cudaMemcpyHostToDevice));
      CUT_CHECK_ERROR("cudaMemcpy fclinvd.eo[1] faild");

      new_clv((*fclinvd).eo[0].g[igx],ipeo_d.g[igx],ieo_d[0].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
      new_clv((*fclinvd).eo[1].g[igx],ipeo_d.g[igx],ieo_d[1].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    }

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
  }

  stopTimer(gsend_timer);
}

extern "C" void
new_gpu_solver_wilson_(f_gvf *u){
  is_clover = false;
  new_gpu_solver_cuda(u);
#ifdef _DEBUG_
  printf("%%NEW_GPU_SOLVER_WILSON\n");
  fflush(stdout);
#endif
}

extern "C" void
new_gpu_solver_clover_(f_gvf *u, f_clvf *fclinv){
  is_clover = true;
  new_gpu_solver_cuda(u,fclinv);
#ifdef _DEBUG_
  printf("%%NEW_GPU_SOLVER_CLOVER\n");
  fflush(stdout);
#endif
}

extern "C" void
delete_gpu_solver_(const int *print_flag) {
//
// Delete GPU solver and summarize GPU statistics
//
// *pring_flag == 1 : print runtime statistics
// *pring_flag != 1 : no output
//

  //
  // delete ghost site copy buffers
  //
  delete_copy_boundary();

  //
  // delete device memorys
  //
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaFree(ipsite_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  ipeo_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   igx_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(ieo_d[0].g[igx]));
    CUDA_SAFE_CALL(cudaFree(ieo_d[1].g[igx]));

    CUDA_SAFE_CALL(cudaFree(phys_blk_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  in_blk_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree( out_blk_d.g[igx]));

    CUDA_SAFE_CALL(cudaFree((*ud).eo[0].g[igx]));
    CUT_CHECK_ERROR("cudaFree ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaFree((*ud).eo[1].g[igx]));
    CUT_CHECK_ERROR("cudaFree ud.eo[1] faild");

    if (is_clover) {
      CUDA_SAFE_CALL(cudaFree((*fclinvd).eo[0].g[igx]));
      CUT_CHECK_ERROR("cudaFree fclinvd.eo[0] faild");
      CUDA_SAFE_CALL(cudaFree((*fclinvd).eo[1].g[igx]));
      CUT_CHECK_ERROR("cudaFree fclinvd.eo[1] faild");
    }

#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
  }

  _mm_free(ud);
  ud = NULL;
  if (is_clover) {
    _mm_free(fclinvd);
    fclinvd = NULL;
  }

  //
  // summarize GPU statistics
  //
  double etime_gsend = ((double)getTimerValue(gsend_timer))*1.0e-3;
  double etime_ysend = ((double)getTimerValue(ysend_timer))*1.0e-3;
  double etime_yrecv = ((double)getTimerValue(yrecv_timer))*1.0e-3;
  double etime_ycopy = ((double)getTimerValue(ycopy_timer))*1.0e-3;
  double etime_solv  = ((double)getTimerValue( solv_timer))*1.0e-3;
  double etime_mult  = ((double)getTimerValue( mult_timer))*1.0e-3;
  double etime_prec  = ((double)getTimerValue( prec_timer))*1.0e-3;
  double etime_binv  = ((double)getTimerValue( binv_timer))*1.0e-3;

  deleteTimer(gsend_timer);
  deleteTimer(ysend_timer);
  deleteTimer(yrecv_timer);
  deleteTimer(ycopy_timer);
  deleteTimer( solv_timer);
  deleteTimer( mult_timer);
  deleteTimer( prec_timer);
  deleteTimer( binv_timer);


  double HNSITE = (double)(_NT*_NZ*_NY*_NX/2);
  double solv_tflop_count = HNSITE*solv_flop_count;
  double mult_tflop_count = HNSITE*mult_flop_count;
  double solv_gflops = 0.0;
  double mult_gflops = 0.0;
  if (etime_solv > 0.0) {
    solv_gflops = ((HNSITE*solv_flop_count)/etime_solv)*1.0e-9;
  }
  if (etime_mult > 0.0) {
    mult_gflops = ((HNSITE*mult_flop_count)/etime_mult)*1.0e-9;
  }

  if ( 0 == _nodeid_ && 1 == *print_flag ) {
    printf("%%==============================================================================\n");
    printf("%% Solve_GFLOPS_SP     : %15.7F\n",solv_gflops);
    printf("%% Solve_FLOP_SP       : %15lld\n",(long long int)solv_tflop_count);
    printf("%% Solve_SP  ETIME(sec): %15.7F\n",etime_solv);

    printf("%% Mult_GFLOPS_SP      : %15.7F\n",mult_gflops);
    printf("%% Mult_FLOP_SP        : %15lld\n",(long long int)mult_tflop_count);
    printf("%% Mult_SP   ETIME(sec): %15.7F\n",etime_mult);

    printf("%% Prec_SP   ETIME(sec): %15.7F\n",etime_prec);
    printf("%% Binv_SP   ETIME(sec): %15.7F\n",etime_binv);
  }

  double gsend_dsize;
  if (is_clover) {
    gsend_dsize =  (((double)sizeof(cu_gvf_eo_blk)*2
                   +(double)sizeof(cu_clvf_eo_blk)*2)*(_NGX))/1024.0/1024.0;
  } else {
    gsend_dsize =  (((double)sizeof(cu_gvf_eo_blk)*2)*(_NGX))/1024.0/1024.0;
  }
  double ysend_dsize = ((double)sizeof(cu_wqf_eo_blk)*(_NGX))/1024.0/1024.0;

  ycopy_dsize = ycopy_dsize/1024.0/1024.0;

  if ( 0 == _nodeid_  && 1 == *print_flag ) {
    if (etime_gsend > 0.0) {
      printf("%% GPU_GSend ETIME(sec): %15.7F\n",etime_gsend);
      printf("%% GPU_GSend  SIZE(MiB): %15.7F\n",gsend_dsize);
      printf("%% GPU_GSend  (MiB/sec): %15.7F\n",gsend_dsize/etime_gsend);
    }
    if (etime_ysend > 0.0) {
      printf("%% GPU_YSend ETIME(sec): %15.7F\n",etime_ysend);
      printf("%% GPU_YSend  SIZE(MiB): %15.7F\n",ysend_dsize);
      printf("%% GPU_YSend  (MiB/sec): %15.7F\n",ysend_dsize/etime_ysend);
    }
    if (etime_yrecv > 0.0) {
      printf("%% GPU_YRecv ETIME(sec): %15.7F\n",etime_yrecv);
      printf("%% GPU_YRecv  SIZE(MiB): %15.7F\n",ysend_dsize);
      printf("%% GPU_YRecv  (MiB/sec): %15.7F\n",ysend_dsize/etime_yrecv);
    }
    if (etime_ycopy > 0.0) {
      printf("%% GPU_YCopy ETIME(sec): %15.7F\n",etime_ycopy);
      printf("%% GPU_YCopy    Called : %15.7F\n",ycopy_count);
      printf("%% GPU_YCopy  SIZE(MiB): %15.7F\n",ycopy_dsize);
      printf("%% GPU_YCopy  (MiB/sec): %15.7F\n",ycopy_dsize/etime_ycopy);
    }
    printf("%%==============================================================================\n");
    fflush(stdout);
  }

  is_clover = false;
}


#include "assign_inv_mult_eoprec_wd_bicgstab.h"

#include "assign_inv_mult_eoprec_wd_cg.h"

#include "assign_inv_mult_Qee_shift_cg.h"

}
