///////////////////////////////////////////////////////////////////
// MIC native solver interfaces.
// This calles the MIC native solver code in the offload region
// The interfaces can be called from Fortran
///////////////////////////////////////////////////////////////////

#include <immintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include "cconfig.h"
#include "cmic_config.h"

#include "mic_nodeinfo.h"
#include "mic_param_size.h"

//////////////////////////////////////////////////
// node infor in the host side
// The value must be set in the fortran module.
//////////////////////////////////////////////////
extern "C" mic_nodeinfo m_host_nodeinfo;

// interface routines for offloaded solver
#include <mpi.h>
#include "signal.h"

///////////////////////////////////////////////
// host offload solver data structure
//   "host_ras_solver" is defined
///////////////////////////////////////////////
#include "mic_offload_solver_class.h"

#pragma offload_attribute(push, target(mic))
#include "scif_interface.h"
#pragma offload_attribute(pop)

///////////////////////////////////////////////
// reverse offload interface 
///////////////////////////////////////////////
#include "mic_reverse_offload.h"

/////////////////////////////////////////////////////////////
// MIC native code prototypes
/////////////////////////////////////////////////////////////
extern "C" __declspec(target(mic)) void 
#ifdef _CLOVER
new_mic_ras_solver(const mic_nodeinfo    *ni,
                   const mic_gvf_soa     *u,
                   const mic_ras_gvf_soa *ru,
                   const mic_clvf        *fclinv,
                   const mic_ras_clvf    *rfclinv);
#else
new_mic_ras_solver(const mic_nodeinfo    *ni,
                   const mic_gvf_soa     *u,
                   const mic_ras_gvf_soa *ru);
#endif

extern "C" __declspec(target(mic)) void
delete_mic_ras_solver();

extern "C" __declspec(target(mic)) void
assign_inv_mult_eoprec_wd_bicgstab_mic(const float  *skappa,
                                       const float  *stol,
                                       int          *iter,
                                             mic_wqf_eo   *restrict yde,
                                       const mic_wqf_eo   *restrict  ye);

extern "C" __declspec(target(mic)) void
#ifdef _CLOVER
assign_mult_hop_eo_mic(      mic_wqf_eo      *restrict mye,
                       const mic_wqf_eo      *restrict  yo,
                       const mic_gvf_eo_soa  *restrict  ue,
                       const mic_gvf_eo_soa  *restrict  uo,
                       const mic_clvf_eo      *fclinve);
#else
assign_mult_hop_eo_mic(      mic_wqf_eo      *restrict mye,
                       const mic_wqf_eo      *restrict  yo,
                       const mic_gvf_eo_soa  *restrict  ue,
                       const mic_gvf_eo_soa  *restrict  uo);
#endif

extern "C" __declspec(target(mic)) int get_mult_count();

/////////////////////////////////////////////////////////////
// END MIC native code prototypes
/////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
// OFFLOAD INTERFACES
/////////////////////////////////////////////////////////////
__declspec(target(mic))
static host_ras_solver *mic_ras_ptr = NULL;

static int m_is_scif_initialized = 0;

extern "C" int m_mic_mult_count;

extern "C" void
new_offl_ras_solver(const mic_gvf_soa     *u,
                    const mic_ras_gvf_soa *u_ras,
                    const mic_clvf        *fclinv,
                    const mic_ras_clvf    *fclinv_ras)
{
    mic_nodeinfo ni = m_host_nodeinfo;
    mic_gvf_soa     *u1;
    mic_ras_gvf_soa *u_ras1;
    mic_clvf        *fclinv1;
    mic_ras_clvf    *fclinv_ras1;

    if ( 0 == m_is_scif_initialized ) {
      start_scif();
      m_is_scif_initialized = 1;
    }

    // Allocate the target struct and its pointers
    #pragma offload_transfer target(mic:mic_targetid) \
      nocopy(mic_ras_ptr[0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(u1         [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(u_ras1     [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(fclinv1    [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(fclinv_ras1[0:1] : alloc_if(1) free_if(0) targetptr)

    // Transfer the data to the card
    #pragma offload target(mic:mic_targetid) \
        in(mic_ras_ptr[0:0] :                        alloc_if(0) free_if(0) targetptr) \
        in(u          [0:1] : into(u1         [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(u_ras      [0:1] : into(u_ras1     [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv     [0:1] : into(fclinv1    [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv_ras [0:1] : into(fclinv_ras1[0:1]) alloc_if(0) free_if(0) targetptr) \
        in(ni)
    {
        mic_ras_ptr->ni         = ni;
        mic_ras_ptr->u          = u1; // store gauge field device pointers to the mic_ras_ptr.
        mic_ras_ptr->u_ras      = u_ras1;
        mic_ras_ptr->fclinv     = fclinv1;
        mic_ras_ptr->fclinv_ras = fclinv_ras1;

        new_mic_ras_solver(&(mic_ras_ptr->ni),
                             mic_ras_ptr->u,
                             mic_ras_ptr->u_ras,
                             mic_ras_ptr->fclinv,
                             mic_ras_ptr->fclinv_ras);

    }

}


extern "C" void
delete_offl_ras_solver()
{
    void *u1, *u_ras1, *fclinv1, *fclinv_ras1;
    mic_gvf_soa     *u;
    mic_ras_gvf_soa *u_ras;
    mic_clvf        *fclinv;
    mic_ras_clvf    *fclinv_ras;

    // Get the device pointers
    #pragma offload target(mic:mic_targetid) \
        in(mic_ras_ptr[0:0] : alloc_if(0) free_if(0) targetptr)
    {
        u1          = (void *)mic_ras_ptr->u;  // restore gauge field device pointers from the mic_ras_ptr
        u_ras1      = (void *)mic_ras_ptr->u_ras;
        fclinv1     = (void *)mic_ras_ptr->fclinv;
        fclinv_ras1 = (void *)mic_ras_ptr->fclinv_ras;
        delete_mic_ras_solver();
    }

    u          = (mic_gvf_soa *)u1;
    u_ras      = (mic_ras_gvf_soa *)u_ras1;
    fclinv     = (mic_clvf *)fclinv1;
    fclinv_ras = (mic_ras_clvf *)fclinv_ras1;

    // free the device memory
    #pragma offload_transfer target(mic:mic_targetid) \
        in(mic_ras_ptr[0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u          [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u_ras      [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv     [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv_ras [0:0] : alloc_if(0) free_if(1) targetptr)

}

#include "commands.h"

extern "C" void
assign_inv_mult_offl_ras_solver( const float      *kappa,
                                 const float      *stol,
                                       int        *iter,
                                 const mic_wqf_eo *swe,
                                       mic_wqf_eo *sve)
{

    int offload_signal = 0xF;
    const float skappa1 = *kappa;
    const float stol1 = *stol;
    int iter1 = *iter;
    int mult_count;

    mult_count = 0;

    ////////////////////////////////////////////////////////////////////////
    // Asynchronous offloading of native code
    //   proxy server is running parallel to the native code.
    ////////////////////////////////////////////////////////////////////////
    #pragma offload target(mic:mic_targetid) \
        nocopy(scif_mic) \
        in( mic_ras_ptr[0:0] : alloc_if(0) free_if(0) targetptr) \
        in( swe[0:1] : alloc_if(1) free_if(1)) \
        out(sve[0:1] : alloc_if(1) free_if(1)) \
        in(skappa1,stol1) \
        inout(iter1) \
        out(mult_count) \
        signal(offload_signal)
    {

        assign_inv_mult_eoprec_wd_bicgstab_mic(&skappa1,&stol1,&iter1,sve,swe);

        mult_count = get_mult_count();

#ifndef _singlePU
        //////////////////////////////////////
        // send Stop-message to the proxy 
        //////////////////////////////////////
        Command cmd;
        cmd.cmd = Command::CMD_Stop;
        cmd.cmdlen = sizeof(Command);
        scif_mic->enqueue(&cmd, sizeof(cmd), true);
        scif_mic->qempty(true);
#endif

    }

    ////////////////////////////////////////////////
    // process proxy 
    // receive requests from the native code located
    // in the above offload region.
    ////////////////////////////////////////////////
    #ifndef _singlePU
    process_cmds();
    #endif

    ////////////////////////////////////////////////
    // wait for finishing of offloading
    ////////////////////////////////////////////////
    #pragma offload_wait target(mic:mic_targetid) wait(offload_signal)

    m_mic_mult_count = mult_count;
    *iter = iter1;
}


////////////////////////////////////////////////////////////////
/* test code
extern "C" void
assign_mult_hop_eo_offl(      mic_wqf_eo      *restrict mye,
                        const mic_wqf_eo      *restrict  yo)
{
    #pragma offload target(mic:mic_targetid) \
        in( mic_ras_ptr[0:0] : alloc_if(0) free_if(0) targetptr) \
        in(  yo[0:1] : alloc_if(1) free_if(1)) \
        out(mye[0:1] : alloc_if(1) free_if(1)) 
    {
      int ioe = yo->ieo;
      int ieo = 1-ioe;
      assign_mult_hop_eo_mic(mye, yo, 
                             &(mic_ras_ptr->u->eo[ieo]),
                             &(mic_ras_ptr->u->eo[ioe]),
                             &(mic_ras_ptr->fclinv->eo[ieo]));
    }
}
*/

