#include "config.h"
#include "mic_config.h"

module mic_solver_class
!
! Single precision solver API used in the host program.
!
  use ISO_C_BINDING
  use comlib
  use lattice_class
  use counter_class
  use field_gauge_class
  use field_fermion_class
  use field_clover_class
  implicit none
  private

  public :: mic_solver
  public :: mic_ras_solver
  public :: get_name
  public :: new
  public :: delete
  public :: assign_inv_mult_eoprec_wd

  !========================================
  ! RAS overlap region depth
  !========================================
#if _NDIMZ != 1
  integer, parameter :: NDEPTHZ = (_NDEPTH)
#else
  integer, parameter :: NDEPTHZ = 0
#endif
#if _NDIMY != 1
  integer, parameter :: NDEPTHY = (_NDEPTH)
#else
  integer, parameter :: NDEPTHY = 0
#endif
#if _NDIMX != 1
  integer, parameter :: NDEPTHX = (_NDEPTH)
#else
  integer, parameter :: NDEPTHX = 0
#endif

  integer, parameter :: CSP=C_FLOAT
!  integer, parameter :: DP=KIND(1.0d0)

  integer, parameter :: VLEN=4
  character(len=*), parameter :: archname="ARCH=MIC-OFFLOAD:"

  integer, parameter :: NTHT=NTH/VLEN

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  !========================================
  ! node information for C-lang solver
  !========================================
  type, BIND(C) :: nodeinfo
    integer(C_INT) :: mic_nodeid
    integer(C_INT) :: mic_nodeid_x_up
    integer(C_INT) :: mic_nodeid_y_up
    integer(C_INT) :: mic_nodeid_z_up
    integer(C_INT) :: mic_nodeid_x_dn
    integer(C_INT) :: mic_nodeid_y_dn
    integer(C_INT) :: mic_nodeid_z_dn
    integer(C_INT) :: mic_nodeid_x
    integer(C_INT) :: mic_nodeid_y
    integer(C_INT) :: mic_nodeid_z
    integer(C_INT) :: mic_ipdeo;
    integer(C_INT) :: mic_ipeo
  end type

  type(nodeinfo), BIND(C), save :: m_host_nodeinfo

  !=================================
  ! module initialization status
  !=================================
  logical, save :: m_is_mic_ras_solver_initialized = .FALSE.

  integer(C_INT), BIND(C), save :: m_mic_mult_count = 0

  !==========================
  ! MPI buffer
  !==========================
  type buffer_obj
    complex(SP), pointer :: buff_up(:,:) => NULL() 
    complex(SP), pointer :: buff_dn(:,:) => NULL() 
    type(comlib_data)    :: id_sendup
    type(comlib_data)    :: id_senddn
    integer :: size = 0
    integer :: dummy1 = 0
    integer :: dummy2 = 0
    integer :: dummy3 = 0
  end type 

  type(buffer_obj), save :: wqhf_buff(NDIM-1)
  type(buffer_obj), save ::  wqf_buff(NDIM-1)
  type(buffer_obj), save ::  gvf_buff(NDIM-1)
  type(buffer_obj), save :: clvf_buff(NDIM-1)

  !==================================
  ! Field Definitions
  !==================================
  type, BIND(C) :: half_spinor
    real(CSP) :: y1r(VLEN)
    real(CSP) :: y2r(VLEN)
    real(CSP) :: y1i(VLEN)
    real(CSP) :: y2i(VLEN)
  end type

  type, BIND(C) :: mic_su3_spinor
    type(half_spinor) :: y(COL,2)
  end type

  type, BIND(C) :: mic_wqf_eo
    type(mic_su3_spinor) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_wqf_eo
    type(mic_su3_spinor) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_wqf
    type(mic_wqf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_wqf
    type(mic_ras_wqf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: su3mat
    real(CSP) :: u11r(VLEN)
    real(CSP) :: u12r(VLEN)
    real(CSP) :: u11i(VLEN)
    real(CSP) :: u12i(VLEN)

    real(CSP) :: u21r(VLEN)
    real(CSP) :: u22r(VLEN)
    real(CSP) :: u21i(VLEN)
    real(CSP) :: u22i(VLEN)

    real(CSP) :: u31r(VLEN)
    real(CSP) :: u32r(VLEN)
    real(CSP) :: u31i(VLEN)
    real(CSP) :: u32i(VLEN)
  end type

  type, BIND(C) :: mic_gsf_eo
    type(su3mat) :: s(NTHT,NZ,NY,NX)
  end type
  type, BIND(C) :: mic_gsf_eo_soa
    type(su3mat) :: v(NDIM)
  end type

  type, BIND(C) :: mic_ras_gsf_eo
    type(su3mat) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
  end type
  type, BIND(C) :: mic_ras_gsf_eo_soa
    type(su3mat) :: v(NDIM)
  end type

  type, BIND(C) :: mic_gvf_eo
    type(mic_gsf_eo) :: v(NDIM)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type
  type, BIND(C) :: mic_gvf_eo_soa
    type(mic_gsf_eo_soa) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_gvf_eo
    type(mic_ras_gsf_eo) :: v(NDIM)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type
  type, BIND(C) :: mic_ras_gvf_eo_soa
    type(mic_ras_gsf_eo_soa) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_gvf
    type(mic_gvf_eo) :: eo(0:1)
  end type
  type, BIND(C) :: mic_gvf_soa
    type(mic_gvf_eo_soa) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_gvf
    type(mic_ras_gvf_eo) :: eo(0:1)
  end type
  type, BIND(C) :: mic_ras_gvf_soa
    type(mic_ras_gvf_eo_soa) :: eo(0:1)
  end type

  type, BIND(C) :: half_clover_term
    real(CSP) ::  f1r(VLEN)
    real(CSP) :: f16r(VLEN)
    real(CSP) ::  f4r(VLEN)
    real(CSP) ::  f4i(VLEN)

    real(CSP) ::  f2r(VLEN)
    real(CSP) :: f17r(VLEN)
    real(CSP) ::  f2i(VLEN)
    real(CSP) :: f17i(VLEN)

    real(CSP) ::  f5r(VLEN)
    real(CSP) ::  f9r(VLEN)
    real(CSP) ::  f5i(VLEN)
    real(CSP) ::  f9i(VLEN)

    real(CSP) ::  f3r(VLEN)
    real(CSP) :: f18r(VLEN)
    real(CSP) ::  f3i(VLEN)
    real(CSP) :: f18i(VLEN)

    real(CSP) ::  f6r(VLEN)
    real(CSP) :: f13r(VLEN)
    real(CSP) ::  f6i(VLEN)
    real(CSP) :: f13i(VLEN)

    real(CSP) ::  f7r(VLEN)
    real(CSP) :: f19r(VLEN)
    real(CSP) :: f10r(VLEN)
    real(CSP) :: f10i(VLEN)

    real(CSP) ::  f8r(VLEN)
    real(CSP) :: f20r(VLEN)
    real(CSP) ::  f8i(VLEN)
    real(CSP) :: f20i(VLEN)

    real(CSP) :: f11r(VLEN)
    real(CSP) :: f14r(VLEN)
    real(CSP) :: f11i(VLEN)
    real(CSP) :: f14i(VLEN)

    real(CSP) :: f12r(VLEN)
    real(CSP) :: f21r(VLEN)
    real(CSP) :: f15r(VLEN)
    real(CSP) :: f15i(VLEN)
  end type

  type, BIND(C) :: mic_clover
    type(half_clover_term) :: c(2)
  end type

  type, BIND(C) :: mic_clvf_eo
    type(mic_clover) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_clvf_eo
    type(mic_clover) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_clvf
    type(mic_clvf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_clvf
    type(mic_ras_clvf_eo) :: eo(0:1)
  end type

  !======================================
  ! MIC Kernel solver
  !======================================
  type mic_solver
    character(len=CHARLEN), private :: name = ""
  end type

  type, extends(mic_solver) :: mic_ras_solver
    private
    type(mic_wqf_eo),      allocatable :: swe
    type(mic_wqf_eo),      allocatable :: sve
    type(mic_gvf_soa),     allocatable :: u
    type(mic_ras_gvf_soa), allocatable :: u_ras
    type(mic_clvf),        allocatable :: fclinv
    type(mic_ras_clvf),    allocatable :: fclinv_ras
    real(DP), public :: err_norm   = 0.0_DP
    real(DP), public :: tol        = 1.0e-4_DP
    integer,  public :: iter       = 0
    integer,  public :: print_flag = 0
    integer,  public :: ieo = 0
  end type

  interface

#ifdef _CLOVER
    subroutine new_offl_ras_solver(u,u_ras,fclinv,fclinv_ras) BIND(C)
#else
    subroutine new_offl_ras_solver(u,u_ras) BIND(C)
#endif
      import
      type(mic_gvf_soa),     intent(in) :: u
      type(mic_ras_gvf_soa), intent(in) :: u_ras
#ifdef _CLOVER
      type(mic_clvf),        intent(in) :: fclinv
      type(mic_ras_clvf),    intent(in) :: fclinv_ras
#endif
    end subroutine

    subroutine delete_offl_ras_solver() BIND(C)
    end subroutine

    subroutine assign_inv_mult_offl_ras_solver(skappa,stol,iter,swe,sve) BIND(C)
      import
      real(CSP),        intent(in)    :: skappa
      integer(C_INT),   intent(inout) :: iter
      type(mic_wqf_eo), intent(in)    :: swe
      type(mic_wqf_eo), intent(out)   :: sve
    end subroutine


!-----------------
! test code 
!    subroutine assign_mult_hop_eo_offl(mye,yo) BIND(C)
!      import
!      type(mic_wqf_eo), intent(inout) :: mye
!      type(mic_wqf_eo), intent(in)    ::  yo
!    end subroutine
!-----------------

  end interface

  interface new
    module procedure new_mic
  end interface

  interface delete
    module procedure delete_mic
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_mic
  end interface

  interface get_name
    module procedure get_name_mic_solver
  end interface

contains

function get_name_mic_solver(this) result(name)
  class(mic_solver), intent(in) :: this
  character(len=CHARLEN) :: name
  name = this%name
  return
end function

#ifdef _CLOVER
subroutine new_mic(this,u,ieo,fclinv)
#else
subroutine new_mic(this,u,ieo)
#endif
  use error_class
  implicit none
  type(mic_ras_solver),    intent(inout) :: this
  type(vfield_gluon_wg),   intent(in)    :: u
#ifdef _CLOVER
  type(field_clover_term), intent(in)    :: fclinv
#endif
  integer,                 intent(in)    :: ieo
  character(len=CHARLEN) :: str
  integer :: ioe
  integer :: mu, nthmodvlen

  this%name = "ARCH=MIC-OFFLOAD:"

  m_host_nodeinfo%mic_nodeid = nodeid
  m_host_nodeinfo%mic_nodeid_x = ipsite(1)
  m_host_nodeinfo%mic_nodeid_y = ipsite(2)
  m_host_nodeinfo%mic_nodeid_z = ipsite(3)
  m_host_nodeinfo%mic_nodeid_x_up = nodeidup(1)
  m_host_nodeinfo%mic_nodeid_y_up = nodeidup(2)
  m_host_nodeinfo%mic_nodeid_z_up = nodeidup(3)
  m_host_nodeinfo%mic_nodeid_x_dn = nodeiddn(1)
  m_host_nodeinfo%mic_nodeid_y_dn = nodeiddn(2)
  m_host_nodeinfo%mic_nodeid_z_dn = nodeiddn(3)
  m_host_nodeinfo%mic_ipeo = ipeo

  nthmodvlen = mod(NTH,VLEN)

#ifdef _DEBUG_
  if (0 == nodeid) then
    write(*,'("%",80("="))')
    write(*,'("% ",A)')archname
    write(*,'("% VLEN=",I3)')VLEN
    write(*,'("% mod(NTH,VLEN)=",I3)') nthmodvlen
  endif
#endif

  if ( 0 /= nthmodvlen ) then
    write(str,'("Stop by NT error! NTT must be a multple of (VLEN*2).")')
    call error_stop(str)
  endif

#ifndef _singlePU

#if _NDIMZ != 1
   wqf_buff(3)%size = CLSP         *VLEN*NTHT*NY*NX*NDEPTHZ
   gvf_buff(3)%size = COL*(COL-1)  *VLEN*NTHT*NY*NX*NDEPTHZ*NDIM
  clvf_buff(3)%size = (CLSPH-COL)*2*VLEN*NTHT*NY*NX*NDEPTHZ
#endif
#if _NDIMY != 1
   wqf_buff(2)%size = CLSP         *VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY
   gvf_buff(2)%size = COL*(COL-1)  *VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY*NDIM
  clvf_buff(2)%size = (CLSPH-COL)*2*VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY
#endif
#if _NDIMX != 1
   wqf_buff(1)%size = CLSP         *VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX
   gvf_buff(1)%size = COL*(COL-1)  *VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX*NDIM
  clvf_buff(1)%size = (CLSPH-COL)*2*VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX
#endif

  do mu=1,NDIM-1
    allocate( wqf_buff(mu)%buff_up( wqf_buff(mu)%size,2))
    allocate( wqf_buff(mu)%buff_dn( wqf_buff(mu)%size,2))
    allocate( gvf_buff(mu)%buff_up( gvf_buff(mu)%size,2))
    allocate( gvf_buff(mu)%buff_dn( gvf_buff(mu)%size,2))
    allocate(clvf_buff(mu)%buff_up(clvf_buff(mu)%size,2))
    allocate(clvf_buff(mu)%buff_dn(clvf_buff(mu)%size,2))
  enddo

  do mu=1,NDIM-1
    call comlib_make2(wqf_buff(mu)%id_sendup,nodeidup(mu),  &
                      wqf_buff(mu)%buff_up(:,1),  &
                      wqf_buff(mu)%buff_up(:,2),wqf_buff(mu)%size)
    call comlib_make2(wqf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      wqf_buff(mu)%buff_dn(:,1),  &
                      wqf_buff(mu)%buff_dn(:,2),wqf_buff(mu)%size)

    call comlib_make2(gvf_buff(mu)%id_sendup,nodeidup(mu),  &
                      gvf_buff(mu)%buff_up(:,1),  &
                      gvf_buff(mu)%buff_up(:,2),gvf_buff(mu)%size)
    call comlib_make2(gvf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      gvf_buff(mu)%buff_dn(:,1),  &
                      gvf_buff(mu)%buff_dn(:,2),gvf_buff(mu)%size)

    call comlib_make2(clvf_buff(mu)%id_sendup,nodeidup(mu),  &
                      clvf_buff(mu)%buff_up(:,1),  &
                      clvf_buff(mu)%buff_up(:,2),clvf_buff(mu)%size)
    call comlib_make2(clvf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      clvf_buff(mu)%buff_dn(:,1),  &
                      clvf_buff(mu)%buff_dn(:,2),clvf_buff(mu)%size)
  enddo

#endif

  if (allocated(this%swe))        deallocate(this%swe)
  if (allocated(this%sve))        deallocate(this%sve)
  if (allocated(this%u))          deallocate(this%u)
  if (allocated(this%u_ras))      deallocate(this%u_ras)
#ifdef _CLOVER
  if (allocated(this%fclinv))     deallocate(this%fclinv)
  if (allocated(this%fclinv_ras)) deallocate(this%fclinv_ras)
#endif

  allocate(this%swe)
  allocate(this%sve)
  allocate(this%u)
  allocate(this%u_ras)
#ifdef _CLOVER
  allocate(this%fclinv)
  allocate(this%fclinv_ras)
#endif

  m_mic_mult_count = 0

  this%ieo      = ieo
  this%iter     = 0
  this%err_norm = 0.0_DP
  this%tol      = 1.0_DP

  !===========================================
  !  send data to mic
  !===========================================

  call conv_u_d2s_eo(this%u%eo(0),u%eo(0))
  call conv_u_d2s_eo(this%u%eo(1),u%eo(1))
  call assign_u_ras_eo(this%u_ras%eo(0),this%u%eo(0))
  call assign_u_ras_eo(this%u_ras%eo(1),this%u%eo(1))

#ifdef _CLOVER
  call conv_fclv_d2s_eo(this%fclinv%eo(0),fclinv%eo(0))
  call conv_fclv_d2s_eo(this%fclinv%eo(1),fclinv%eo(1))
  call assign_fclv_ras_eo(this%fclinv_ras%eo(0),this%fclinv%eo(0))
  call assign_fclv_ras_eo(this%fclinv_ras%eo(1),this%fclinv%eo(1))
#endif

#ifdef _CLOVER
  call new_offl_ras_solver( this%u, this%u_ras, this%fclinv, this%fclinv_ras)
#else
  call new_offl_ras_solver( this%u, this%u_ras)
#endif

  if (allocated(this%u))          deallocate(this%u)
  if (allocated(this%u_ras))      deallocate(this%u_ras)
#ifdef _CLOVER
  if (allocated(this%fclinv))     deallocate(this%fclinv)
  if (allocated(this%fclinv_ras)) deallocate(this%fclinv_ras)
#endif

  m_is_mic_ras_solver_initialized = .TRUE.

  return
end subroutine

subroutine delete_mic(this)
  implicit none
  type(mic_ras_solver), intent(inout) :: this
  integer :: mu

  if (m_is_mic_ras_solver_initialized) then

    m_mic_mult_count = 0

    call delete_offl_ras_solver()

#ifndef _singlePU
    do mu=1,NDIM-1
      if (associated( wqf_buff(mu)%buff_up)) deallocate( wqf_buff(mu)%buff_up)
      if (associated( wqf_buff(mu)%buff_dn)) deallocate( wqf_buff(mu)%buff_dn)
      if (associated( gvf_buff(mu)%buff_up)) deallocate( gvf_buff(mu)%buff_up)
      if (associated( gvf_buff(mu)%buff_dn)) deallocate( gvf_buff(mu)%buff_dn)
#ifdef _CLOVER
      if (associated(clvf_buff(mu)%buff_up)) deallocate(clvf_buff(mu)%buff_up)
      if (associated(clvf_buff(mu)%buff_dn)) deallocate(clvf_buff(mu)%buff_dn)
#endif
    enddo
#endif

    if (allocated(this%swe))        deallocate(this%swe)
    if (allocated(this%sve))        deallocate(this%sve)
    if (allocated(this%u))          deallocate(this%u)
    if (allocated(this%u_ras))      deallocate(this%u_ras)
#ifdef _CLOVER
    if (allocated(this%fclinv))     deallocate(this%fclinv)
    if (allocated(this%fclinv_ras)) deallocate(this%fclinv_ras)
#endif

    m_is_mic_ras_solver_initialized = .FALSE.

  endif

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_mic(this,kappa,dye,ye)
  use omp_lib
  implicit none
  type(mic_ras_solver), intent(inout) :: this
  real(DP),             intent(in)    :: kappa
  complex(DP),          intent(inout) :: dye(:)  ! packed data
  complex(DP),          intent(inout) ::  ye(:)  ! packed data
  complex(DP) :: dyo(SIZE(ye))
  type(mic_wqf_eo) :: myo,mye,syo,sye
  real(DP) :: t0,t1
  real(DP) :: rnorm
  real(CSP) :: stol,skappa
  integer(C_INT) :: ieo
  integer(C_INT) :: iter

#ifdef _TIMING_
  t0 = omp_get_wtime()
#endif

  skappa = kappa
  iter = this%iter
  stol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)

  m_mic_mult_count = 0

  ieo = this%ieo
  this%swe%ieo = ieo
  this%sve%ieo = ieo

  call unpack_d2s_eo(ye,this%swe,rnorm)

  call assign_inv_mult_offl_ras_solver(skappa,stol,iter,this%swe,this%sve)
  this%iter = iter

  call pack_s2d_eo(this%sve,dye,rnorm)

  if (this%iter < 0) then
    call delete(this)
    if (0 == nodeid) then
      write(*,'("%assign_inv_mult_eoprec_wd_bicgstab_mic : maximum iteration reached.")')
    endif
  endif

  call inc(mult_iter,m_mic_mult_count)

#ifdef _TIMING_
  t1 = omp_get_wtime()
  if (0 == nodeid) then
    write(*,'("#ETIME_=",ES24.15)')(t1-t0)
  endif
#endif

  return
end subroutine

subroutine unpack_d2s_eo(vec,sqe,rnorm)
!
!  rnorm = |dqe|
!  sqe = dqe/rnorm
!
  implicit none
  complex(DP),             intent(in)    :: vec(:)
  type(mic_wqf_eo),        intent(inout) :: sqe
  real(DP), optional,      intent(out)   :: rnorm
  integer :: ieo,ix,iy,iz,ieoxyz,ith,itht,ivec,ith0
  integer :: ic,is,isite
  real(DP) :: rtmp
  complex(DP) :: yy(COL,SPIN)

  if (present(rnorm)) then
    rnorm = 0.0d0
!$OMP PARALLEL DO PRIVATE(isite) REDUCTION(+:rnorm)
   do isite = 1, SIZE(vec)
     rnorm = rnorm + REAL(vec(isite))**2 + AIMAG(vec(isite))**2
   enddo
#ifndef _singlePU
    call comlib_sumcast(rnorm)
#endif
    rnorm = SQRT(rnorm)
    rtmp = 1.0_DP/rnorm
  else
    rtmp = 1.0_DP
  endif
!  print *,'rnorm,rtmp',rnorm,rtmp

  ieo = sqe%ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,is,itht,ivec,ith0,isite,yy) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1
      ith0 = ith + ieoxyz

      do is=1,SPIN
      do ic=1,COL
        isite = ic  &
 &        +    (is-1)*COL            &
 &        +  (ith0-1)*CLSP           &
 &        +    (iz-1)*CLSP*NTH       &
 &        +    (iy-1)*CLSP*NTH*NZ    &
 &        +    (ix-1)*CLSP*NTH*NZ*NY
        yy(ic,is) = vec(isite)*rtmp
      enddo
      enddo

      do ic=1,COL
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y1r(ivec) =  real(yy(ic,1),kind=DP)
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y1i(ivec) = aimag(yy(ic,1))
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y2r(ivec) =  real(yy(ic,2),kind=DP)
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y2i(ivec) = aimag(yy(ic,2))
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y1r(ivec) =  real(yy(ic,3),kind=DP)
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y1i(ivec) = aimag(yy(ic,3))
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y2r(ivec) =  real(yy(ic,4),kind=DP)
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y2i(ivec) = aimag(yy(ic,4))
      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine pack_s2d_eo(sqe,vec,rnorm)
  implicit none
  type(mic_wqf_eo),    intent(in)    :: sqe
  complex(DP),         intent(inout) :: vec(:)
  real(DP),  optional, intent(in)    :: rnorm
  integer :: ieo
  integer :: ix,iy,iz,ieoxyz,ith,itht,ivec,isite,ith0
  integer :: ic,is
  complex(DP) :: ctmp,yy(COL,SPIN)
  real(DP) :: rtmp

  if (present(rnorm)) then
    rtmp = rnorm
  else
    rtmp = 1.0_DP
  endif

  ieo = sqe%ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,is,itht,ivec,ith0,isite,yy) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1
      ith0 = ith + ieoxyz

      do ic=1,COL
        yy(ic,1) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,1)%y1r(ivec),  &
  &                      sqe%s(itht,iz,iy,ix)%y(ic,1)%y1i(ivec),kind=DP)
        yy(ic,2) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,1)%y2r(ivec),  &
  &                      sqe%s(itht,iz,iy,ix)%y(ic,1)%y2i(ivec),kind=DP)
        yy(ic,3) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,2)%y1r(ivec),  &
  &                      sqe%s(itht,iz,iy,ix)%y(ic,2)%y1i(ivec),kind=DP)
        yy(ic,4) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,2)%y2r(ivec),  &
  &                      sqe%s(itht,iz,iy,ix)%y(ic,2)%y2i(ivec),kind=DP)
      enddo

      do is=1,SPIN
      do ic=1,COL
        isite = ic  &
 &        +    (is-1)*COL            &
 &        +  (ith0-1)*CLSP           &
 &        +    (iz-1)*CLSP*NTH       &
 &        +    (iy-1)*CLSP*NTH*NZ    &
 &        +    (ix-1)*CLSP*NTH*NZ*NY

        vec(isite) = yy(ic,is)*rtmp
      enddo
      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine conv_u_d2s_eo(sue,due)
  implicit none
  type(mic_gvf_eo_soa),     intent(inout) :: sue
  type(vfield_gluon_eo_wg), intent(in)    :: due
  integer :: ieo,ix,iy,iz,ieoxyz,ith,itht,ivec
  integer :: ic,jc,mu

  ieo = due%ieo
  sue%ieo = ieo

!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,ieoxyz,ith,itht,ivec) COLLAPSE(4)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do mu=1,NDIM
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1

      sue%s(itht,iz,iy,ix)%v(mu)%u11r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(1,1),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u11i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(1,1))
      sue%s(itht,iz,iy,ix)%v(mu)%u21r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(2,1),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u21i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(2,1))
      sue%s(itht,iz,iy,ix)%v(mu)%u31r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(3,1),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u31i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(3,1))
      sue%s(itht,iz,iy,ix)%v(mu)%u12r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(1,2),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u12i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(1,2))
      sue%s(itht,iz,iy,ix)%v(mu)%u22r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(2,2),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u22i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(2,2))
      sue%s(itht,iz,iy,ix)%v(mu)%u32r(ivec) =  real(due%mu(mu)%s(ith,iz,iy,ix)%u(3,2),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u32i(ivec) = aimag(due%mu(mu)%s(ith,iz,iy,ix)%u(3,2))

    enddo
  enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_u_ras_eo(ras_ue,ue)
  implicit none
  type(mic_ras_gvf_eo_soa), intent(inout) :: ras_ue
  type(mic_gvf_eo_soa),     intent(in)    :: ue
  integer :: ix,iy,iz,itht
  integer :: ic,jc,mu
  integer :: ieo

  ieo  = ue%ieo
  ras_ue%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itht=1,NTHT
      ras_ue%s(itht,iz,iy,ix)%v(mu) = ue%s(itht,iz,iy,ix)%v(mu)
    enddo
  enddo
  enddo
  enddo
  enddo

  call copy_boundary_ras_gvf_eo(ras_ue)

  return
end subroutine

subroutine conv_fclv_d2s_eo(sfclve,dfclve)
  implicit none
  type(mic_clvf_eo),          intent(inout) :: sfclve
  type(field_clover_term_eo), intent(in)    :: dfclve
  integer :: ieo,ix,iy,iz,ieoxyz,ith,ith0,itht,ivec
  integer :: is

  ieo = dfclve%ieo
  sfclve%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ith0,itht,ivec,is) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1
      ith0 = ith + ieoxyz

      do is=1,2
        sfclve%s(itht,iz,iy,ix)%c(is)%f1r(ivec)  =  real(dfclve%l( 1,is,ith0,iz,iy,ix),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f2r(ivec)  =  real(dfclve%l( 2,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f2i(ivec)  = aimag(dfclve%l( 2,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f3r(ivec)  =  real(dfclve%l( 3,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f3i(ivec)  = aimag(dfclve%l( 3,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f4r(ivec)  =  real(dfclve%l( 4,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f4i(ivec)  = aimag(dfclve%l( 4,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f5r(ivec)  =  real(dfclve%l( 5,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f5i(ivec)  = aimag(dfclve%l( 5,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f6r(ivec)  =  real(dfclve%l( 6,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f6i(ivec)  = aimag(dfclve%l( 6,is,ith0,iz,iy,ix))

        sfclve%s(itht,iz,iy,ix)%c(is)%f7r(ivec)  =  real(dfclve%l( 7,is,ith0,iz,iy,ix),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f8r(ivec)  =  real(dfclve%l( 8,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f8i(ivec)  = aimag(dfclve%l( 8,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f9r(ivec)  =  real(dfclve%l( 9,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f9i(ivec)  = aimag(dfclve%l( 9,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f10r(ivec) =  real(dfclve%l(10,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f10i(ivec) = aimag(dfclve%l(10,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f11r(ivec) =  real(dfclve%l(11,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f11i(ivec) = aimag(dfclve%l(11,is,ith0,iz,iy,ix))

        sfclve%s(itht,iz,iy,ix)%c(is)%f12r(ivec) =  real(dfclve%l(12,is,ith0,iz,iy,ix),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f13r(ivec) =  real(dfclve%l(13,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f13i(ivec) = aimag(dfclve%l(13,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f14r(ivec) =  real(dfclve%l(14,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f14i(ivec) = aimag(dfclve%l(14,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f15r(ivec) =  real(dfclve%l(15,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f15i(ivec) = aimag(dfclve%l(15,is,ith0,iz,iy,ix))

        sfclve%s(itht,iz,iy,ix)%c(is)%f16r(ivec) =  real(dfclve%l(16,is,ith0,iz,iy,ix),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f17r(ivec) =  real(dfclve%l(17,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f17i(ivec) = aimag(dfclve%l(17,is,ith0,iz,iy,ix))
        sfclve%s(itht,iz,iy,ix)%c(is)%f18r(ivec) =  real(dfclve%l(18,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f18i(ivec) = aimag(dfclve%l(18,is,ith0,iz,iy,ix))

        sfclve%s(itht,iz,iy,ix)%c(is)%f19r(ivec) =  real(dfclve%l(19,is,ith0,iz,iy,ix),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f20r(ivec) =  real(dfclve%l(20,is,ith0,iz,iy,ix),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f20i(ivec) = aimag(dfclve%l(20,is,ith0,iz,iy,ix))

        sfclve%s(itht,iz,iy,ix)%c(is)%f21r(ivec) =  real(dfclve%l(21,is,ith0,iz,iy,ix),kind=DP)
!=========

      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_fclv_ras_eo(ras_fclve,fclve)
  implicit none
  type(mic_ras_clvf_eo), intent(inout) :: ras_fclve
  type(mic_clvf_eo),     intent(in)    :: fclve
  integer :: ix,iy,iz,itht,ieo
  integer :: is

  ieo  = fclve%ieo
  ras_fclve%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht) COLLAPSE(4)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itht=1,NTHT
      ras_fclve%s(itht,iz,iy,ix) = fclve%s(itht,iz,iy,ix)
    enddo
  enddo
  enddo
  enddo

  call copy_boundary_ras_clvf_eo(ras_fclve)

  return
end subroutine

subroutine copy_boundary_ras_wqf_eo(qe)
  type(mic_ras_wqf_eo), intent(inout) :: qe
  integer :: ix,iy,iz,itht,ivec,ic,is,ibuff
  integer, parameter :: NDAT = VLEN*COL*SPIN
  complex(4) :: ctmp(NDAT)
  type(mic_su3_spinor) :: su3tmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      wqf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,NZ-NDEPTHZ+1+iz,iy,ix),mold=ctmp)
      wqf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,           1+iz,iy,ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHZ*NTHT*NX*NY
!  if (0 == nodeid) write(*,'("F:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(3)%size
  call comlib_sendrecv(wqf_buff(3)%id_sendup)
  call comlib_sendrecv(wqf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      qe%s(itht,1-NDEPTHZ+iz,iy,ix) = TRANSFER(wqf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,     NZ+1+iz,iy,ix) = TRANSFER(wqf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      wqf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,NY-NDEPTHY+1+iy,ix),mold=ctmp)
      wqf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,           1+iy,ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHY*NTHT*NZ*(2*NDEPTHZ + NZ)
!  if (0 == nodeid) write(*,'("F:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(2)%size
  call comlib_sendrecv(wqf_buff(2)%id_sendup)
  call comlib_sendrecv(wqf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      qe%s(itht,iz,1-NDEPTHY+iy,ix) = TRANSFER(wqf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,iz,     NY+1+iy,ix) = TRANSFER(wqf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      wqf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,iy,NX-NDEPTHX+1+ix),mold=ctmp)
      wqf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,iy,           1+ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHX*NTHT*(2*NDEPTHY + NY)*(2*NDEPTHZ + NZ)
!  if (0 == nodeid) write(*,'("F:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(1)%size
  call comlib_sendrecv(wqf_buff(1)%id_sendup)
  call comlib_sendrecv(wqf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      qe%s(itht,iz,iy,1-NDEPTHX+ix) = TRANSFER(wqf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,iz,iy,     NX+1+ix) = TRANSFER(wqf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine


subroutine copy_boundary_ras_gvf_eo(ue)
  type(mic_ras_gvf_eo_soa), intent(inout) :: ue
  integer :: mu,ix,iy,iz,itht,ivec,ic,jc,ibuff
  integer, parameter :: NDAT = VLEN*COL*(COL-1)
  complex(4) :: ctmp(NDAT)
  type(su3mat) :: su3tmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1 + NX*(mu-1)))))
      gvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,NZ-NDEPTHZ+1+iz,iy,ix)%v(mu),mold=ctmp)
      gvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,           1+iz,iy,ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(3)%size
  call comlib_sendrecv(gvf_buff(3)%id_sendup)
  call comlib_sendrecv(gvf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1 + NX*(mu-1)))))
      ue%s(itht,1-NDEPTHZ+iz,iy,ix)%v(mu) = TRANSFER(gvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,     NZ+1+iz,iy,ix)%v(mu) = TRANSFER(gvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1 + NX*(mu-1)))))
      gvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,NY-NDEPTHY+1+iy,ix)%v(mu),mold=ctmp)
      gvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,           1+iy,ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(2)%size
  call comlib_sendrecv(gvf_buff(2)%id_sendup)
  call comlib_sendrecv(gvf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1 + NX*(mu-1)))))
      ue%s(itht,iz,1-NDEPTHY+iy,ix)%v(mu) = TRANSFER(gvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,iz,     NY+1+iy,ix)%v(mu) = TRANSFER(gvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*(ix + NDEPTHX*(mu-1)))))
      gvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,iy,NX-NDEPTHX+1+ix)%v(mu),mold=ctmp)
      gvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,iy,           1+ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(1)%size
  call comlib_sendrecv(gvf_buff(1)%id_sendup)
  call comlib_sendrecv(gvf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*(ix + NDEPTHX*(mu-1)))))
      ue%s(itht,iz,iy,1-NDEPTHX+ix)%v(mu) = TRANSFER(gvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,iz,iy,     NX+1+ix)%v(mu) = TRANSFER(gvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine

subroutine copy_boundary_ras_clvf_eo(fclve)
  type(mic_ras_clvf_eo), intent(inout) :: fclve
  integer :: mu,ix,iy,iz,itht,ivec,is,ibuff,ic
  integer, parameter :: NDAT=VLEN*(CLSPH-COL)*2
  complex(4) :: ctmp(NDAT)
  type(mic_clover) :: clvtmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      clvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,NZ-NDEPTHZ+1+iz,iy,ix),mold=ctmp)
      clvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,           1+iz,iy,ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(3)%size
  call comlib_sendrecv(clvf_buff(3)%id_sendup)
  call comlib_sendrecv(clvf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      fclve%s(itht,1-NDEPTHZ+iz,iy,ix) = TRANSFER(clvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,     NZ+1+iz,iy,ix) = TRANSFER(clvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      clvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,NY-NDEPTHY+1+iy,ix),mold=ctmp)
      clvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,           1+iy,ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(2)%size
  call comlib_sendrecv(clvf_buff(2)%id_sendup)
  call comlib_sendrecv(clvf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      fclve%s(itht,iz,1-NDEPTHY+iy,ix) = TRANSFER(clvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,iz,     NY+1+iy,ix) = TRANSFER(clvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      clvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,iy,NX-NDEPTHX+1+ix),mold=ctmp)
      clvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,iy,           1+ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(1)%size
  call comlib_sendrecv(clvf_buff(1)%id_sendup)
  call comlib_sendrecv(clvf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      fclve%s(itht,iz,iy,1-NDEPTHX+ix) = TRANSFER(clvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,iz,iy,     NX+1+ix) = TRANSFER(clvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine

subroutine mic_dump(tag,vec,ieo)
  implicit none
  character(len=1) :: tag
  complex(DP), intent(in) :: vec(:)
  integer,     intent(in) :: ieo
  integer :: ix,iy,iz,ieoxyz,ith,ith0
  integer :: ic,is,isite

  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ith0 = ith + ieoxyz

      do is=1,SPIN
      do ic=1,COL
        isite = ic  &
 &        +    (is-1)*COL            &
 &        +  (ith0-1)*CLSP           &
 &        +    (iz-1)*CLSP*NTH       &
 &        +    (iy-1)*CLSP*NTH*NZ    &
 &        +    (ix-1)*CLSP*NTH*NZ*NY
        write(*,'(A,":",I10,7I3,2E24.15)')tag,isite,ieo,ix,iy,iz,ith,ic,is,vec(isite)
      enddo
      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine

end module
