#ifndef _MIC_SOLVER_TIMER_H
#define _MIC_SOLVER_TIMER_H

///////////////////////////////////////////////////////////////
// Timers and counters to measure the MIC Kernel performance
///////////////////////////////////////////////////////////////
typedef struct {
  micTimer_t     ycopy_timer;  // MPI copy timer
  micTimer_t ycopy_wait_timer; // MPI copy receive wait timer
  micTimer_t ras_ycopy_timer;  // RAS MPI copy timer
  micTimer_t     gsend_timer;  // HOST=>MIC gauge fields send timer
  micTimer_t     ysend_timer;  // HOST=>MIC fermion fields send timer
  micTimer_t     yrecv_timer;  // MIC=>HOST fermion fields send timer
  micTimer_t      solv_timer;  // solver timer
  micTimer_t      mult_timer;  // normal mult timer
  micTimer_t  mult_w_comm_timer;  // normal mult + comm timer
  micTimer_t  mult_pre_timer;  // pre computation for mult on surface sites timer
  micTimer_t   mult_in_timer;  // normal mult on interior sites timer
  micTimer_t  mult_pst_timer;  // post computation for mult on surface sites timer
  micTimer_t  sap_mult_timer;  // SAP mult timer
  micTimer_t  ras_mult_timer;  // RAS mult timer
  micTimer_t  sap_prec_timer;  // SAP preconditioner timer
  micTimer_t  ras_prec_timer;  // RAS preconditioner timer
  micTimer_t      binv_timer;  // Domain block solver timer

  micTimer_t      prod_timer;
  micTimer_t      abs2_timer;
  micTimer_t    update_timer;

  micTimer_t     mem_timer;     // timing the malloc and free on MIC 
  micTimer_t     assign_timer;  // timing the assignment
  micTimer_t     reduct_timer;  // timing the MPI reduction

  long long int         mult_count;  // counts hopping mult calls
  long long int     ras_mult_count;  // counts RAS hopping mult calls
  long long int     sap_mult_count;  // counts SAP hopping mult calls
  long long int        ycopy_count;  // counts MPI_sendrecv calls
  double               ycopy_dsize;  // counts MPI_sendrecv total data size
  long long int    ras_ycopy_count;  // counts MPI_sendrecv calls
  double           ras_ycopy_dsize;  // counts MPI_sendrecv total data size
  long long int        gsend_count;  // HOST=>MIC gauge fields send count
  double               gsend_dsize;  // HOST=>MIC gauge fields send size
  long long int        ysend_count;  // HOST=>MIC fermion fields send count
  double               ysend_dsize;  // HOST=>MIC fermion fields send size
  long long int        yrecv_count;  // MIC=>HOST fermion fields send count
  double               yrecv_dsize;  // MIC=>HOST fermion fields send size

  double           solv_flop_count;  // counts total flop in the solver
  double           mult_flop_count;  // counts total flop for mult operation
  double       mult_pre_flop_count;  // counts pre computation flop for mult surface  operation
  double        mult_in_flop_count;  // counts flop for mult interior operation
  double       mult_pst_flop_count;  // counts post computation flop for mult surface  operation
  double       sap_mult_flop_count;  // counts total flop for SAP mult operation
  double       ras_mult_flop_count;  // counts total flop for RAS mult operation

  double           prod_ld_dsize;    // counts total loaded data size for prod
  double           abs2_ld_dsize;    // counts total loaded data size for abs2

  double           prod_flop_count;  // counts total flop for prod
  double           abs2_flop_count;  // counts total flop for abs2
  double         update_flop_count;  // counts total flop for update
} mic_timer;

void new_timer(mic_timer *timer) {
  createTimer(timer->solv_timer);
  createTimer(timer->mult_timer);
  createTimer(timer->mult_w_comm_timer);
  createTimer(timer->ycopy_timer);
  createTimer(timer->ycopy_wait_timer);
  createTimer(timer->gsend_timer);
  createTimer(timer->ysend_timer);
  createTimer(timer->yrecv_timer);

  createTimer(timer->mult_pre_timer);
  createTimer(timer->mult_in_timer);
  createTimer(timer->mult_pst_timer);

  createTimer(timer->binv_timer);
  createTimer(timer->sap_mult_timer);
  createTimer(timer->ras_mult_timer);
  createTimer(timer->sap_prec_timer);
  createTimer(timer->ras_prec_timer);
  createTimer(timer->ras_ycopy_timer);

  createTimer(timer->prod_timer);
  createTimer(timer->abs2_timer);
  createTimer(timer->update_timer);
  createTimer(timer->mem_timer);
  createTimer(timer->assign_timer);
  createTimer(timer->reduct_timer);

  timer->mult_count  = 0;
  timer->ycopy_count = 0;
  timer->ycopy_dsize = 0.0;
  timer->gsend_count = 0;
  timer->gsend_dsize = 0.0;
  timer->ysend_count = 0;
  timer->ysend_dsize = 0.0;
  timer->yrecv_count = 0;
  timer->yrecv_dsize = 0.0;

  timer->sap_mult_count = 0;
  timer->ras_mult_count = 0;
  timer->ras_ycopy_count = 0;
  timer->ras_ycopy_dsize = 0.0;

  timer->solv_flop_count = 0.0;
  timer->mult_flop_count = 0.0;
  timer->mult_pre_flop_count = 0.0;
  timer->mult_in_flop_count  = 0.0;
  timer->mult_pst_flop_count = 0.0;
  timer->sap_mult_flop_count = 0.0;
  timer->ras_mult_flop_count = 0.0;

  timer->prod_ld_dsize = 0.0;
  timer->abs2_ld_dsize = 0.0;
  timer->prod_flop_count = 0.0;
  timer->abs2_flop_count = 0.0;
  timer->update_flop_count = 0.0;

}

void delete_timer(mic_timer *timer) {

  double etime_solv  = getTimerValue(timer->solv_timer);
  double etime_mult  = getTimerValue(timer->mult_timer);
  double etime_mult_w_comm  = getTimerValue(timer->mult_w_comm_timer);
  double etime_mult_pre = getTimerValue(timer->mult_pre_timer);
  double etime_mult_in  = getTimerValue(timer->mult_in_timer);
  double etime_mult_pst = getTimerValue(timer->mult_pst_timer);
  double etime_ycopy = getTimerValue(timer->ycopy_timer);
  double etime_ycopy_wait = getTimerValue(timer->ycopy_wait_timer);
  double etime_gsend = getTimerValue(timer->gsend_timer);
  double etime_ysend = getTimerValue(timer->ysend_timer);
  double etime_yrecv = getTimerValue(timer->yrecv_timer);

  double etime_binv  = getTimerValue(timer->binv_timer);

  double etime_sap_mult = getTimerValue(timer->sap_mult_timer);
  double etime_ras_mult = getTimerValue(timer->ras_mult_timer);

  double etime_sap_prec = getTimerValue(timer->sap_prec_timer);
  double etime_ras_prec = getTimerValue(timer->ras_prec_timer);

  double etime_ras_ycopy = getTimerValue(timer->ras_ycopy_timer);

  double etime_prod = getTimerValue(timer->prod_timer);
  double etime_abs2 = getTimerValue(timer->abs2_timer);
  double etime_updt = getTimerValue(timer->update_timer);

  double etime_mem    = getTimerValue(timer->mem_timer);
  double etime_assign = getTimerValue(timer->assign_timer);
  double etime_reduct = getTimerValue(timer->reduct_timer);

  if ( 0 == _nodeid_ ) {
    printf("%%==============================================================================\n");
    printf("%% MIC Kernel Performance / MPI process \n");
    printf("%%------------------------------------------------------------------------------\n");
    printf("%% Solve_GFLOPS_SP          : %15.7F\n",(timer->solv_flop_count/etime_solv)*1.0e-9);
    printf("%% Solve_FLOP_SP            : %15lld\n",(long long int)timer->solv_flop_count);
    printf("%% Solve_SP       ETIME(sec): %15.7F\n",etime_solv);

    printf("%% Mult_W_COMM_GFLOPS_SP    : %15.7F\n",(timer->mult_flop_count/etime_mult_w_comm)*1.0e-9);
    printf("%% Mult_W_COMM_SP ETIME(sec): %15.7F\n",etime_mult);
    printf("%% Mult_GFLOPS_SP           : %15.7F\n",(timer->mult_flop_count/etime_mult)*1.0e-9);
    printf("%% Mult_FLOP_SP             : %15lld\n",(long long int)timer->mult_flop_count);
    printf("%% Mult_SP           Called : %15lld\n",timer->mult_count);
    printf("%% Mult_SP        ETIME(sec): %15.7F\n",etime_mult);
#ifndef _singlePU
    printf("%% Mult_PRE_GFLOPS_SP       : %15.7F\n",(timer->mult_pre_flop_count/etime_mult_pre)*1.0e-9);
    printf("%% Mult_PRE_FLOP_SP         : %15lld\n",(long long int)timer->mult_pre_flop_count);
    printf("%% Mult_PRE_SP    ETIME(sec): %15.7F\n",etime_mult_pre);
    printf("%% Mult_IN_GFLOPS_SP        : %15.7F\n",(timer->mult_in_flop_count/etime_mult_in)*1.0e-9);
    printf("%% Mult_IN_FLOP_SP          : %15lld\n",(long long int)timer->mult_in_flop_count);
    printf("%% Mult_IN_SP     ETIME(sec): %15.7F\n",etime_mult_in);
    printf("%% Mult_PST_GFLOPS_SP       : %15.7F\n",(timer->mult_pst_flop_count/etime_mult_pst)*1.0e-9);
    printf("%% Mult_PST_FLOP_SP         : %15lld\n",(long long int)timer->mult_pst_flop_count);
    printf("%% Mult_PST_SP    ETIME(sec): %15.7F\n",etime_mult_pst);
#endif

    if ( 0.0 < etime_ras_prec ) {
      printf("%% Mult_RAS_GFLOPS_SP       : %15.7F\n",(timer->ras_mult_flop_count/etime_ras_mult)*1.0e-9);
      printf("%% Mult_RAS_FLOP_SP         : %15lld\n",(long long int)timer->ras_mult_flop_count);
      printf("%% Mult_RAS_SP       Called : %15lld\n",timer->ras_mult_count);
      printf("%% Mult_RAS_SP    ETIME(sec): %15.7F\n",etime_ras_mult);
   
      printf("%% Prec_SP        ETIME(sec): %15.7F\n",etime_ras_prec);
      printf("%% Binv_SP        ETIME(sec): %15.7F\n",etime_binv);

      printf("%% RAS_YCopy      ETIME(sec): %15.7F\n",etime_ras_ycopy);
      printf("%% RAS_YCopy         Called : %15lld\n",timer->ras_ycopy_count);
      printf("%% RAS_YCopy       SIZE(MiB): %15.7F\n",timer->ras_ycopy_dsize/1024/1024);
      printf("%% RAS_YCopy       (MiB/sec): %15.7F\n",timer->ras_ycopy_dsize/etime_ras_ycopy/1024/1024);

    }    

    if ( 0.0 < etime_sap_prec ) {
      printf("%% Mult_SAP_GFLOPS_SP       : %15.7F\n",(timer->sap_mult_flop_count/etime_sap_mult)*1.0e-9);
      printf("%% Mult_SAP_FLOP_SP         : %15lld\n",(long long int)timer->sap_mult_flop_count);
      printf("%% Mult_SAP_SP       Called : %15lld\n",timer->sap_mult_count);
      printf("%% Mult_SAP_SP    ETIME(sec): %15.7F\n",etime_sap_mult);
   
      printf("%% Prec_SP        ETIME(sec): %15.7F\n",etime_sap_prec);
      printf("%% Binv_SP        ETIME(sec): %15.7F\n",etime_binv);
    }

    printf("%% Prod_GFLOPS_SP           : %15.7F\n",(timer->prod_flop_count/etime_prod)*1.0e-9);
    printf("%% Prod_FLOP_SP             : %15lld\n",(long long int)timer->prod_flop_count);
    printf("%% Prod         LD SIZE(MiB): %15.7F\n",timer->prod_ld_dsize/1024/1024);
    printf("%% Prod       LD BW(MiB/sec): %15.7F\n",timer->prod_ld_dsize/1024/1024/etime_prod);
    printf("%% Prod           ETIME(sec): %15.7F\n",etime_prod);

    printf("%% Abs2_GFLOPS_SP           : %15.7F\n",(timer->abs2_flop_count/etime_abs2)*1.0e-9);
    printf("%% Abs2_FLOP_SP             : %15lld\n",(long long int)timer->abs2_flop_count);
    printf("%% Abs2         LD SIZE(MiB): %15.7F\n",timer->abs2_ld_dsize/1024/1024);
    printf("%% Abs2       LD BW(MiB/sec): %15.7F\n",timer->abs2_ld_dsize/1024/1024/etime_abs2);
    printf("%% Abs2           ETIME(sec): %15.7F\n",etime_abs2);

    printf("%% Update_GFLOPS_SP         : %15.7F\n",(timer->update_flop_count/etime_updt)*1.0e-9);
    printf("%% Update_FLOP_SP           : %15lld\n",(long long int)timer->update_flop_count);
    printf("%% Update         ETIME(sec): %15.7F\n",etime_updt);

    printf("%% MPI Reduct     ETIME(sec): %15.7F\n",etime_reduct);

    if ( 0.0 < etime_gsend ) {
      printf("%% GSend          ETIME(sec): %15.7F\n",etime_gsend);
      printf("%% GSend             Called : %15lld\n",timer->gsend_count);
      printf("%% GSend           SIZE(MiB): %15.7F\n",timer->gsend_dsize/1024/1024);
      printf("%% GSend           (MiB/sec): %15.7F\n",timer->gsend_dsize/etime_gsend/1024/1024);
    }

    if ( 0.0 < etime_ysend ) {
      printf("%% YSend          ETIME(sec): %15.7F\n",etime_ysend);
      printf("%% YSend             Called : %15lld\n",timer->ysend_count);
      printf("%% YSend           SIZE(MiB): %15.7F\n",timer->ysend_dsize/1024/1024);
      printf("%% YSend           (MiB/sec): %15.7F\n",timer->ysend_dsize/etime_ysend/1024/1024);
    }

    if ( 0.0 < etime_yrecv ) {
      printf("%% YRecv          ETIME(sec): %15.7F\n",etime_yrecv);
      printf("%% YRecv             Called : %15lld\n",timer->yrecv_count);
      printf("%% YRecv           SIZE(MiB): %15.7F\n",timer->yrecv_dsize/1024/1024);
      printf("%% YRecv           (MiB/sec): %15.7F\n",timer->yrecv_dsize/etime_yrecv/1024/1024);
    }

    if ( 0.0 < etime_ycopy ) {
      printf("%% YCopy          ETIME(sec): %15.7F\n",etime_ycopy);
      printf("%% YCopy             Called : %15lld\n",timer->ycopy_count);
      printf("%% YCopy           SIZE(MiB): %15.7F\n",timer->ycopy_dsize/1024/1024);
      printf("%% YCopy           (MiB/sec): %15.7F\n",timer->ycopy_dsize/etime_ycopy/1024/1024);
      printf("%% YCopy_WAIT     ETIME(sec): %15.7F\n",etime_ycopy_wait);
    }
    printf("%%==============================================================================\n");
  }

  deleteTimer(timer->solv_timer);
  deleteTimer(timer->mult_timer);
  deleteTimer(timer->mult_w_comm_timer);
  deleteTimer(timer->mult_pre_timer);
  deleteTimer(timer->mult_in_timer);
  deleteTimer(timer->mult_pst_timer);
  deleteTimer(timer->ycopy_timer);
  deleteTimer(timer->gsend_timer);
  deleteTimer(timer->ysend_timer);
  deleteTimer(timer->yrecv_timer);

  deleteTimer(timer->binv_timer);
  deleteTimer(timer->sap_mult_timer);
  deleteTimer(timer->ras_mult_timer);
  deleteTimer(timer->sap_prec_timer);
  deleteTimer(timer->ras_prec_timer);
  deleteTimer(timer->ras_ycopy_timer);

  deleteTimer(timer->prod_timer);
  deleteTimer(timer->abs2_timer);
  deleteTimer(timer->update_timer);

  deleteTimer(timer->mem_timer);
  deleteTimer(timer->assign_timer);
  deleteTimer(timer->reduct_timer);

}

static mic_timer solver_timer;

extern "C" void start_timer_gsend() { startTimer(solver_timer.gsend_timer); }
extern "C" void start_timer_ysend() { startTimer(solver_timer.ysend_timer); }
extern "C" void start_timer_yrecv() { startTimer(solver_timer.yrecv_timer); }

extern "C" void  stop_timer_gsend(const size_t size) {  stopTimer(solver_timer.gsend_timer); solver_timer.gsend_count++; solver_timer.gsend_dsize += size; }
extern "C" void  stop_timer_ysend(const size_t size) {  stopTimer(solver_timer.ysend_timer); solver_timer.ysend_count++; solver_timer.ysend_dsize += size; }
extern "C" void  stop_timer_yrecv(const size_t size) {  stopTimer(solver_timer.yrecv_timer); solver_timer.yrecv_count++; solver_timer.yrecv_dsize += size; }

#endif
