#if !defined(_OMP_SCHED_H)
#define _OMP_SCHED_H

#include <stdint.h>
#include <omp.h>
#include <assert.h>

#if defined(LONG)
typedef uint64_t iter_t;
#else
typedef uint32_t iter_t;
#endif

// aligned so we can eventually use CMPXCHG16 */
__declspec(align(16))
struct Iters
{
    iter_t start; // first iteration on this core
    iter_t end;   // last iteration on this core
};

#if 0
static inline bool
machine_cmpswp16(volatile uint64_t *addr, Iters value, Iters comparand)
{
    bool result;
    __asm__ volatile(
        "lock; cmpxchg16b %1; setz %0"
        : "=q" (result), "+m" (*addr), "+d" (value.end), "+a" (value.start)
        : "c" (comparand.end), "b" (comparand.start)
        : "cc"
    );
    return result;
}
static inline bool
machine_cmpswp8(volatile uint32_t *addr, Iters value, Iters comparand)
{
    bool result;
    __asm__ volatile(
        "lock; cmpxchg8b %1; setz %0"
        : "=q" (result), "+m" (*addr), "+d" (value.end), "+a" (value.start)
        : "c" (comparand.end), "b" (comparand.start)
        : "cc"
    );
    return result;
}
#if defined(LONG)
#define CMPSWP(a, b, c) machine_cmpswp16((volatile uint64_t *)(a), b, c)
#else
#define CMPSWP(a, b, c) machine_cmpswp8((volatile uint32_t *)(a), b, c)
#endif
#endif

# define LIKELY(x)    __builtin_expect(!!(x),1)
# define UNLIKELY(x)  __builtin_expect(!!(x),0)

static inline void cpu_pause()
{
    // pause, sleep, or delay this thread
    _mm_delay_32(4);
}

static inline uint32_t machine_swp4(uint32_t volatile *ptr, uint32_t value)
{
    __asm__ volatile("xchg  %1,%0"
                     : "+r"(value), "+m"(*ptr)
                     :
                     : "memory");
    return value;
}


static inline int testSimpleLock(uint32_t * lock, uint32_t lockedValue, uint32_t freeValue)
{
#if (HLE_ENABLED)
    return machine_swp4_xacquire(lock, lockedValue) == freeValue;
#else
    return machine_swp4(lock, lockedValue) == freeValue;
#endif
}

static void waitForSimpleLock(uint32_t * lock, uint32_t lockedValue, uint32_t freeValue)
{
    // A test and test-and-set scheme with some exponential backoff
    int delay = 1;

    do {
        // test ...
        while (*(uint32_t volatile *)lock != freeValue)
        {
            for (int i=delay; i != 0; --i)
                cpu_pause();
            delay = ((delay << 1) | 1) & 7;
        }
        // and test-and-set (since we just saw the lock free, we hope that it is unlikely that it is now locked again)
    } while (UNLIKELY(machine_swp4(lock, lockedValue) != freeValue));
}

static inline void doSimpleLock(uint32_t * lock, uint32_t lockedValue, uint32_t freeValue)
{
    if (*(uint32_t volatile *)lock == freeValue)
        if (LIKELY(testSimpleLock(lock, lockedValue, freeValue)))
            return;
    
    waitForSimpleLock(lock, lockedValue, freeValue);
}

const int maxCores = 72;
const int nHT = 4;

__declspec(align(64))
struct Percore
{
    volatile uint32_t lock;
    volatile Iters iters;
    volatile uint32_t barrier[2];
    volatile iter_t myiter[2];
};

__declspec(align(64)) struct Sched
{
    int nthreads;
    int ncores;
    int mythread;
    int myht;
    int mycore;
    int barrier_id;
    int waitval;
    int flags[2];
    Percore *core;
    Percore *base;

    iter_t getiter(Percore &p)
    {
        int retval = -1;
        if (p.iters.start >= p.iters.end)
            goto ret;
        doSimpleLock((uint32_t *)&p.lock, 1, 0);
        if (p.iters.start < p.iters.end)
            retval = p.iters.start++;
        else
            retval = -1;
        p.lock = 0;
    ret:
        return retval;
    }

    iter_t stealiter(Percore &p)
    {
        int retval = -1;
        if (p.iters.start >= p.iters.end)
            goto ret;
        doSimpleLock((uint32_t *)&p.lock, 1, 0);
        if (p.iters.start < p.iters.end)
            retval = --p.iters.end;
        else
            retval = -1;
        p.lock = 0;
    ret:
        return retval;
    }

    void init(Percore *xbase)
    {
        nthreads = omp_get_num_threads();
        assert((nthreads % nHT) == 0);
        ncores = nthreads / nHT;
        mythread = omp_get_thread_num();
        myht = mythread % nHT;
        mycore = mythread / nHT;
        base = xbase;
        core = base + mycore;
        flags[0] = flags[1] = 1;
        barrier_id = 0;
        waitval = 0x01010101;
    }

    void
    corebarrier()
    {
        // determine which barrier and values to use
        int flag = flags[barrier_id];
        int wv = (flag) ? waitval : 0;

        // loop until all threads have written to this barrier
        ((char *)&core->barrier[barrier_id])[myht] = flag;
        while (core->barrier[barrier_id] != wv)
            cpu_pause();

        // toggle the barrier/flag for next time
        flags[barrier_id] = 1 - flag;
        barrier_id = 1 - barrier_id;
    }

    Sched(iter_t niters, Percore *cores)
    {
        init(cores);
        iter_t nper = niters / ncores;
        iter_t rem  = niters % ncores;
        iter_t start = 0;
        iter_t end = 0;
        if (mycore < rem)
        {
            start = mycore * nper + mycore;
            end = start + nper + 1;
        }
        else
        {
            start = mycore * nper + rem;
            end = start + nper;
        }
        //printf("Thread=%d Core=%d start=%d end=%d\n", mythread, mycore,
        //        start, end);
        if (myht == 0)
        {
            core->iters.start = start;
            core->iters.end = end;
            core->barrier[0] = 0;
            core->barrier[1] = 0;
            core->myiter[0] = -1;
            core->myiter[1] = -1;
            core->lock = 0;
        }
        //corebarrier();
    }

    // return the next iteration on my core, or steal one. Return -1 if done.
    iter_t _nextiter()
    {
        iter_t ret = getiter(*core);
        if (ret != -1)
            return ret;

        // need to steal
        //for (int i = mycore+1; i != mycore; ++i)
        int startcore = _rdtsc() % ncores;
        for (int i = startcore; i != startcore; ++i)
        {
            // wraparound
            if (i == ncores)
            {
                i = 0;
                //if (mycore == 0)
                if (startcore == 0)
                    break;
            }
            ret = stealiter(base[i]);
            if (ret != -1)
                return ret;
        }
        return -1;
    }

    iter_t nextiter()
    {
        if (myht == 0)
        {
            core->myiter[barrier_id] = _nextiter();
        }
        corebarrier();
        return core->myiter[1-barrier_id];
    }

};

//__declspec(align(64)) __thread Sched::Tls Sched::mytls;
#endif

