#include "scif_interface.h"
#include <err.h>
#include <time.h>
#include <string.h>
#include <poll.h>
#include <sys/mman.h>

//////////////////////////////////////////////////////////////////////////
// SCIF interface API for the reverse-offloading mechanism
//////////////////////////////////////////////////////////////////////////

/**
 * Constructor
 */
#pragma offload_attribute target(mic)
SCIF_Interface::SCIF_Interface() {
    sciferr = 0;
}

/**
 * Create interface for sending messages
 */
void SCIF_Interface::create_send_port(int node, uint16_t pn)
{
    portID.node = node;
    portID.port = pn;

    /*scif_open(): call to open an endpoint to which the remote process will
     * connect
     */
    #ifdef DEBUG
    printf("Open the scif driver\n");
    #endif
    if ((epd = scif_open()) < 0)
        err(1,"create_send_port: scif_open error %d", (int)epd);

    /*scif_connect(): API to request a connection to a remote node
      epd:        The end point address returned from scif_open()
      portID:         Remote note address information contains the remote node and portID
    */
    int ret;
    timespec req = {0, 10000};
    const int ntries = 100;
    for (int i = 0; i < ntries; ++i)
    {
        if ((ret = scif_connect(epd, &portID)) >= 0)
            break;
        printf("scif_connect: errno = %d\n", errno);
        if (errno == ECONNREFUSED)
        {
            nanosleep(&req, 0);
            continue;
        }
        err(1, "create_send_port: scif_connect");
    }
    assert(ret > 0);

    #ifdef DEBUG
    printf("scif_connect success\n");
    #endif

    // Allocate the registered buffers
    void *p;
    p = mmap(0, rbufz, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
    if (p == MAP_FAILED)
        err(1, "create_send_port: mmap");
    rbuf = (char *)p;
    roff = register_buffer(rbuf, rbufz, false);
    rcur = 0;

    // receive the remote offset
    receive_message((char *)&qoff, sizeof(qoff), true);
    // receive the remote registered buffer offset
    receive_message((char *)&roffrem, sizeof(roffrem), true);
    // map the message queue buffer
    void *q = mmap_buffer(qlen+4096, qoff, false);
    qbuf = (char *)q;

    sender = true;
}

/**
 * Create a SCIF port for listening and return it
 */
uint16_t SCIF_Interface::create_receive_port(int &node)
{
    uint16_t self;


    /*scif_open(): call to open an endpoint to which the remote process will
     * connect
     */
    #ifdef DEBUG
    printf("Open the scif driver\n");
    #endif
    if ((epd = scif_open()) < 0)
        err(1, "create_receive_port: scif_open error %d", (int)epd);

    /*scif_bind(): Bind the port and get an endpoint
    */
    #ifdef DEBUG
    printf("scif_bind to port 0\n");
    #endif
    if ((con_pn = scif_bind(epd, 0)) < 0)
        err(1, "create_receive_port: scif_bind error %d\n", con_pn);
    #ifdef DEBUG
    printf("scif_bind returns %d\n", con_pn);
    #endif

    int nnodes = get_nodes(0, 0, &self);
    node = self;
    return con_pn;
}

/**
 * Wait for a connection on the server side
 */
void SCIF_Interface::server()
{
    /*scif_listen(): to mark the endpoint, epd, as a listening endpoint and can only be used to
      accept incoming connection requests.  backlog defines the maximum number of pending connection requests
    */
    #ifdef DEBUG
    printf("scif_listen\n");
    #endif
    if ((scif_listen(epd, BACKLOG)) < 0)
    {
        printf("scif_listen failed with error %d\n", errno);
        exit(3);
    }

    /*scif_accept(): to accept a connection request from the remote node where:
      epd:        The end point returned from scif_open()
      portID:     Filled in with peer node and port information
      newepd:     New endpoint created for connection and if successful,
      this new endpoint being passed back to the caller for future reference.
      flags:      Indicates sychronous or asynchronous mode: SCIF_ACCEPT_SYNC is set then this API
      block until a connection is presented. Otherwise, it will fail with an error.
    */
    #ifdef DEBUG
    printf("scif_accept in syncronous mode\n");
    #endif
    if (((scif_accept(epd, &portID, &newepd, SCIF_ACCEPT_SYNC)) < 0) && (errno != EAGAIN))
    {
        printf("scif_accept failed with errno %d\n", errno);
        exit(4);
    }

    #ifdef DEBUG
    printf("scif_connect success\n");
    #endif
    tmpepd = epd; // this endpoint stays open in listening state - let's save it in case we need it
    epd = newepd; // this is the endpoint we should use for communication

    // Register the message queue
    void *p;
    int ret = posix_memalign(&p, 4096, qlen+4096);
    assert(ret == 0);
    qbuf = (char *)p;
    // first two 64-bit words are start and end (to be interpreted modulo
    // qlen)
    ((uint64_t *)qbuf)[0] = 0;
    ((uint64_t *)qbuf)[1] = 0;
    qoff = register_buffer(qbuf, qlen+4096, false);

    // setup the registered buffer area
    ret = posix_memalign(&p, 4096, rbufz);
    p = mmap(0, rbufz, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS|MAP_HUGETLB, -1, 0);
    if (p == MAP_FAILED)
        err(1, "server: mmap");
    rbuf = (char *)p;
    roff = register_buffer(rbuf, rbufz, false);
    roffrem = roff;
    rcur = 0;

    // send the offset to the host
    send_message((char *)&qoff, sizeof(qoff), true);
    // set the registered buffer offset to the host
    send_message((char *)&roff, sizeof(roff), true);

    sender = false;
}


/**
 * Send message in given buffer
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::send_message(char *buffer, unsigned msg_size, bool block)
{
    int ret;

    if ((ret = scif_send(epd, buffer, msg_size,
        block ? SCIF_SEND_BLOCK : 0)) <= 0)
    {
        if (ret < 0) {
            warn("send_message: scif_send");
            fflush(0);
        }
    }

    if (ret != msg_size)
    {
        printf("Only portion of message sent!\n");
        assert(0);
    }
}


/**
 * Receive message into given buffer
 **/
#pragma offload_attribute target(mic)
void SCIF_Interface::receive_message(char *buffer, unsigned msg_size, bool block)
{
    int ret;

    if ((ret = scif_recv(epd, buffer, msg_size,
        block ? SCIF_RECV_BLOCK : 0)) <= 0)
    {
        if (ret < 0) {
            warn("receive_message: scif_recv");
            fflush(0);
        }
    }

    if (ret != msg_size)
    {
        printf("Only portion of message received!\n");
        assert(0);
    }
}

/**
 * Register a buffer with the SCIF RMA mechanism
 */
#pragma offload_attribute target(mic)
off_t SCIF_Interface::register_buffer(char *buffer, unsigned buffer_size, bool read_only)
{
    off_t offset;

    offset = scif_register(epd,
                           buffer,                 //starting virtual address self_addr
                           buffer_size,            //continue the window for msg_size bytes
                           0,                      //offset of window, 0 means let SCIF pick
                           (read_only)? SCIF_PROT_READ : (SCIF_PROT_READ | SCIF_PROT_WRITE),
                           0); // don't force SCIF to map to fixed offset
    if (offset < 0)
        err(1, "register_buffer: scif_register");

    return offset;
}

/**
 * Write message from given registered buffer to remote buffer at given offset
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::write(off_t local_offset, off_t remote_offset, unsigned buffer_size)
{
    int ret;

    /*scif_writeto(): Copy to a remote address space */
    if ((ret = scif_writeto(epd,
                            local_offset,     //offset in local registered address space from which to copy
                            buffer_size,      //amount to copy
                            remote_offset,    //offset in remote registered address space to which to copy
                            0)))
    {
        err(1, "write: scif_writeto");
    }
}

/**
 * Read message from remote buffer at given offset into given registered buffer
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::read(off_t local_offset, off_t remote_offset, unsigned buffer_size)
{
    int ret;

    /* scif_readfrom(): copy from a remote address space */
    if ((ret = scif_readfrom(epd,
                             local_offset,         //offset in local registered address space to which to copy
                             buffer_size,          //length of range to copy
                             remote_offset,        //offset in remote registered address space from which to copy
                             0)))
    {
        err(1, "read: scif_readfrom");
    }
}

/**
 * Wait for all outstanding RMAs to finish.  If self_fence is false, waits for peer's RMAs.
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::fence(bool self_fence)
{
    int ret;
    int mark;

    if (ret = scif_fence_mark(epd,
                              self_fence? SCIF_FENCE_INIT_SELF : SCIF_FENCE_INIT_PEER,
                              &mark))
    {
        err(1, "fence: scif_fence_mark");
    }

    /*scif_fence_wait() returns after all RMAs marked with mark have completed.  */
    if (ret = scif_fence_wait(epd, mark))
    {
        err(1, "fence: scif_fence_wait");
    }
}

/**
 * mmap a buffer with the SCIF remote mmap mechanism
 */
void *SCIF_Interface::mmap_buffer(size_t buffer_size, off_t off, bool read_only)
{
    void *addr;

    addr = scif_mmap(0,                 // virtual address
                    buffer_size,        // size
                    (read_only)? SCIF_PROT_READ : (SCIF_PROT_READ | SCIF_PROT_WRITE),
                    0,                  // map flags
                    epd,
                    off);               // remote offset
    if (addr == SCIF_MMAP_FAILED)
        err(1, "mmap_buffer: scif_mmap");

    return addr;
}

void SCIF_Interface::checkep()
{
    pollfd fds[1] = {{epd, 0, 0}};
    int ret = poll(fds, 1, 0);
    if (ret == -1)
        err(1, "poll");
    else if (fds->revents != 0)
        errx(1, "Endpoint lost");
}

// dequeue a message.
// If len is 0, no message is available, otherwise data points to the
// message data
void SCIF_Interface::dequeue(int &data, uint64_t &msgstart, size_t &len,
    bool block)
{
    uint64_t start = ((uint64_t *)qbuf)[0];
    uint64_t end = ((volatile uint64_t *)qbuf)[1];

    len = 0;
    if (start >= end)
    {
        if (!block)
            return;
        int retry = 0;
        while (start >= ((volatile uint64_t *)qbuf)[1])
        {
            #if defined(__MIC__)
            _mm_delay_32(100);
            #else
            _mm_pause();
            #endif
            ++retry;
            if ((retry % 1000) == 0)
                checkep();
        }
    }
    msgstart = start;
    int offs = start % qlen;
    assert(offs+sizeof(Command) <= qlen);
    Command *cmd = (Command *)&qbuf[4096 + offs];
    data = cmd->cmd;
    len = cmd->cmdlen;
}

// Copy whole message including the command
void SCIF_Interface::qcopy(Command *msg, uint64_t msgstart, size_t len)
{
    assert(len <= qlen);
    int offs = msgstart % qlen;
    char *first = &qbuf[4096 + offs];
    size_t rem = qlen - offs;
    size_t l = len < rem ? len : rem;
    msgstart += len;
    memcpy((void *)msg, (void *)first, l);
    if (rem < len)
    {
        len -= rem;
        memcpy((void *)((char *)msg+rem), (void *)&qbuf[4096], len);
    }
    ((uint64_t *)qbuf)[0] = msgstart;
}

// enqueue a message. Return true for success, false if there's no space
#pragma offload_attribute target(mic)
bool SCIF_Interface::enqueue(Command *msg, size_t len, bool block)
{
    assert(len <= qlen);
    assert((len % sizeof(Command)) == 0);
    uint64_t end = ((uint64_t *)qbuf)[1];
    uint64_t start = ((volatile uint64_t *)qbuf)[0];

    if (qlen - (end - start) < len)
    {
        if (!block)
            return false;
        int retry = 0;
        while ((qlen - (end - ((volatile uint64_t *)qbuf)[0])) < len)
        {
            #if defined(__MIC__)
            _mm_delay_32(100);
            #else
            _mm_pause();
            #endif
            ++retry;
            if ((retry % 1000) == 0)
                checkep();
        }
    }
    int offs = end % qlen;
    size_t rem = qlen - offs;
    size_t l = len < rem ? len : rem;
    end += len;
    memcpy((void *)&qbuf[4096+offs], (void *)msg, l);
    if (rem < len)
    {
        len -= rem;
        memcpy((void *)&qbuf[4096], (void *)((char *)msg+rem), len);
    }
    ((uint64_t *)qbuf)[1] = end;
    return true;
}

// check if queue is empty
#pragma offload_attribute target(mic)
bool SCIF_Interface::qempty(bool block)
{
    uint64_t end = ((uint64_t *)qbuf)[1];
    uint64_t start = ((volatile uint64_t *)qbuf)[0];
    int size = end - start;
    assert(size >= 0);
    if (size == 0)
        return true;
    if (!block)
        return false;
    int retry = 0;
    while ( ( end - ((volatile uint64_t *)qbuf)[0] ) > 0)
    {
        #if defined(__MIC__)
        _mm_delay_32(100);
        #else
        _mm_pause();
        #endif
        ++retry;
        if ((retry % 1000) == 0)
            checkep();
    }
    return true;
}

// Allocate a new buffer and return offset (card only)
bool SCIF_Interface::allocbuf(int len, off_t &off, void *&buf)
{
    // cache line
    int mylen = (len + 63) & ~63;
    if (rbufz - rcur < mylen)
    {
        fprintf(stderr, "rbuf failed: ask %d avail %d\n",
            mylen, rbufz-rcur);
        exit(1);
        return false;
    }
    off = roff + rcur;
    buf = &rbuf[rcur];
    rcur += mylen;
    return true;
}
//
// Get the buffer address. off is the card-side offset
void *SCIF_Interface::getbuf(off_t off)
{
    return &rbuf[off - roffrem];
}

// Read a card buffer into the corresponding buffer on the host
void SCIF_Interface::readbuf(off_t off, size_t len)
{
    off_t diff = off - roffrem;
    read(roff+diff, roffrem+diff, len);
}

// Write a card buffer from the corresponding buffer on the host
void SCIF_Interface::writebuf(off_t off, size_t len)
{
    off_t diff = off - roffrem;
    write(roff+diff, roffrem+diff, len);
}

/**
 * Close SCIF driver
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::close(void)
{
    scif_close(epd);
}

/**
 * Get list of nodes in system
 */
#pragma offload_attribute target(mic)
unsigned SCIF_Interface::get_nodes(uint16_t *nodes, int len, uint16_t *self)
{
    int val = scif_get_nodeIDs(nodes, len, self);

    if (val == -1)
        err(1, "get_nodeIDs");
    printf("I am SCIF node %d\n", *self);
    return (unsigned)val;
}

/**
 * Initialize the signal for asynchronous fences.
 */
#pragma offload_attribute target(mic)
off_t SCIF_Interface::init_asynch_fence(void)
{
    posix_memalign(&signal_addr, 4096, 4096);
    clear_asynch_signal();

    off_t offset = register_buffer((char *)signal_addr, 4096, false);

    return offset;
}

/**
 * Mark outstanding RMAs, and when all done write "value" to remote offset
 */
#pragma offload_attribute target(mic)
void SCIF_Interface::asynch_fence(off_t offset, unsigned value)
{
    int ret;
    ret = scif_fence_signal(epd, 0, 0, offset, value, SCIF_FENCE_INIT_SELF | SCIF_SIGNAL_REMOTE);
    if (ret < 0)
        err(1, "async_fence: scif_fence_signal");
}

