#include <config.h>

#define _COL3

!
! Measure Polyakov loop
!
!   Initialization of ploop_measurement object, pl_obj, is done
!   in two ways depending on context;
!
!     call new( pl_obj, params_filename, hmc_status_obj ) in on-line measurement
!     call new( pl_obj, params_filename, conf_id ) in ordinal measurement
!
!   where conf_id is some id associated with configuration
!   "new" must be called before the first call of "is_loop_end"
!   in "hmc_job_manager" object
!
module ploop_class
  use comlib
  use lattice_class
  use timer_class
  use logfile_class
  use hmc_logfile_class
  use hmc_status_class
  use error_class
  use field_gauge_class
  implicit none
  private

  public :: ploop_measurement
  public :: new
  public :: delete
  public :: measure

  type(timer), save :: copy_pl_time ! time spent to copy variables on boundary
  logical, save :: pl_is_initialized = .FALSE. ! initialization status 
  
  !
  !  boundary copy related variables
  !
  integer, parameter :: PBUFF_SIZE(NDIM-1)  &
 &            = (/COL**2*(NTH+1)*NY*NZ,    & ! p-loop wrapped along x-direction
 &                COL**2*(NTH+1)*NZ*NX,    & ! p-loop wrapped along y-direction
 &                COL**2*(NTH+1)*NX*NY  /)   ! p-loop wrapped along z-direction
  integer, parameter :: PBUFF_MAX_SIZE  &
 &            = MAX(PBUFF_SIZE(1), PBUFF_SIZE(2), PBUFF_SIZE(3))
  ! Polyakov loop ghost sites buffers
  complex(DP), save, target :: pbuffe(PBUFF_MAX_SIZE,2)
  complex(DP), save, target :: pbuffo(PBUFF_MAX_SIZE,2)
  ! comlib data for ghost site exchange.
  type(comlib_data), save   :: idpsende(NDIM-1)
  type(comlib_data), save   :: idpsendo(NDIM-1)

  type :: Polyakov_loop
  !
  ! Polyakov loop initialization status
  !
    integer, private :: idummy(2) = 0
  end type Polyakov_loop

  type ploop_measurement
    ! path to which outputs are printed out
    character(LEN=CHARLEN) :: path_to_output = ""
    ! output file name header
    character(LEN=CHARLEN) :: output_fname_header = ""
    ! number of digits for run id
    integer :: num_digits_id ! 6 for "000123", 4 for "0123"
    integer :: dummy
    type(logfile), pointer :: log_ploop => NULL()
     ! In context of HMC
    type(hmc_status), pointer :: status => NULL()
    type(hmc_logfile), pointer :: hmc_log_ploop => NULL()
  end type ploop_measurement

  interface new
    module procedure new_ploop_measurement_hmc
    module procedure new_ploop_measurement_ordinal
  end interface new

  interface read
    module procedure read_ploop_measurement
  end interface read

  interface delete
    module procedure delete_ploop_measurement
  end interface delete

  interface measure
    module procedure measure_ploop_measurement
  end interface measure

  interface run
    module procedure run_ploop_measurement
  end interface run

contains

subroutine new_Polyakov_loop( this )
!
! initialize environment of communication among nodes 
!
  use comlib
  implicit none
  type(Polyakov_loop), intent(inout) :: this
!-----------------------------------------------------------
  integer :: mu
  type(lattice_world) :: lattice
 
  if (.not. is_initialized(lattice)) then
    write(*,'("Lattice is not initialized!")')
    stop
  endif

  if (pl_is_initialized) then
    if ( nodeid == 0 ) then
      write(*, &
&       '("Communication environment for ploop was already initialized!")')
    endif
    return
  endif
 
  call new(copy_pl_time)

#ifndef _singlePU
  pbuffe(:,1) = Z0
  pbuffe(:,2) = Z0
  pbuffo(:,1) = Z0
  pbuffo(:,2) = Z0
  if ( NPU > 1 ) then
    do mu = 1, (NDIM-1)
      call comlib_make2( idpsende(mu), nodeiddn(mu),   &
                         pbuffe(:,1), pbuffe(:,2),   &
                         PBUFF_SIZE(mu) )
      write(idpsende(mu)%name,'("COPY_PL_EVEN:",I2)') mu
      call comlib_make2( idpsendo(mu), nodeiddn(mu),   &
                         pbuffo(:,1), pbuffo(:,2),   &
                         PBUFF_SIZE(mu) )
      write(idpsendo(mu)%name,'("COPY_PL_ODD :",I2)') mu
    enddo ! mu
  endif
#endif

  pl_is_initialized = .TRUE. 

  return
end subroutine new_Polyakov_loop


subroutine read_ploop_measurement( this, infname )
  !
  ! read job parameters from file
  !
  use comlib, only : comlib_bcast
  use file_tools_class
  use iso_c_functions
  implicit none
  type(ploop_measurement), intent(inout) :: this
  character(LEN=*), intent(in) :: infname ! parameter file name
!------------------------------------------------------------
  integer :: unit_id, ierr
  character(LEN=CHARLEN) :: str

  if (0 == nodeid) then
    unit_id = search_free_file_unit()
    open(unit_id,file=TRIM(infname),status='old',form='formatted')
    read(unit_id,'(A)') str
    read(unit_id,'(A)') this%path_to_output        ! ploop history file output directory path
    read(unit_id,'(A)') str
    read(unit_id,'(A)') this%output_fname_header   ! ploop history file name header
    read(unit_id,'(A)') str
    read(unit_id,*) this%num_digits_id             ! ploop history file index-id number of digits "ex. 6 means fname.000023"
    close(unit_id)

!    write(*,'("path:",A)') this%path_to_output 
!    call f_mkdir( this%path_to_output, ierr )

  endif
#ifndef _singlePU
  call comlib_bcast(this%path_to_output,0)
  call comlib_bcast(this%output_fname_header,0)
  call comlib_bcast(this%num_digits_id,0)
#endif

  if (0 == nodeid) then
    write(*,'(80("="))')
    write(*,'(" Ploop parameter read OK : ",A)') TRIM(ADJUSTL(infname))
    write(*,'(" Ploop output path       : ",A)') TRIM(ADJUSTL(this%path_to_output))
    write(*,'(" Ploop output fname      : ",A)') TRIM(ADJUSTL(this%output_fname_header))
    write(*,'(80("="))')
  endif

  return
end subroutine read_ploop_measurement


function get_fname_ploop_measurement( path, header, digits_id, id ) result( fname )
  use iso_c_functions
  implicit none
  character(LEN=CHARLEN), intent(in) :: path, header
  integer, intent(in) :: digits_id, id
  character(LEN=CHARLEN) :: fname
!-----------------------------------------------
  integer :: j
  character(LEN=digits_id) :: cid
 
  if ( digits_id == 4 ) then
    write(cid, '(I4)') id
  else if ( digits_id == 5 ) then
    write(cid, '(I5)') id
  else if ( digits_id == 6 ) then
    write(cid, '(I6)') id
  else
    call error_stop( "Program cannot deal with specified number of digits" )
  endif
  
  do j = 1, digits_id
    if ( cid(j:j) == ' ' ) cid(j:j) = '0'
  enddo

  ! make full output file name
  call f_mkdir( path, j )
  fname = TRIM( TRIM(ADJUSTL(path)) // '/' // TRIM(ADJUSTL(header)) )
  fname = TRIM( TRIM(ADJUSTL(fname)) // '.' // TRIM(ADJUSTL(cid)) )

  return
end function get_fname_ploop_measurement


subroutine new_ploop_measurement_hmc( this, infname, status_hmc )
!
!  for on-line measurement
!
  implicit none
  type(ploop_measurement), intent(inout) :: this
  character(LEN=*), intent(in) :: infname ! parameter file name
  type(hmc_status), intent(in), target :: status_hmc
!-----------------------------------------------------------
  character(LEN=CHARLEN) :: fheader
  type(Polyakov_loop) :: pl_obj

  this%status => status_hmc

  call new_Polyakov_loop(pl_obj) ! initialization of communication

  call read( this, infname )

  if ( .not. associated( this%hmc_log_ploop ) ) then
    allocate( this%hmc_log_ploop )
    fheader = TRIM(TRIM(ADJUSTL(this%path_to_output)) // '/' // &
 &                 TRIM(ADJUSTL(this%output_fname_header)))
    !fheader = TRIM(TRIM(ADJUSTL(fheader)) // '.')
    call new( this%hmc_log_ploop, fheader )
  endif

  return
end subroutine new_ploop_measurement_hmc


subroutine new_ploop_measurement_ordinal( this, infname, id )
!
!  for ordinal measurement
!
  implicit none
  type(ploop_measurement), intent(inout) :: this
  character(LEN=*), intent(in) :: infname ! parameter file name
  integer, intent(in) :: id
!-----------------------------------------------------------
  character(LEN=CHARLEN) :: fname, str
  type(Polyakov_loop) :: pl_obj

  call new_Polyakov_loop(pl_obj) ! initialization of communication

  call read( this, infname )

  if ( .not. associated( this%log_ploop ) ) then
    allocate( this%log_ploop )
    fname = get_fname_ploop_measurement( this%path_to_output, &
 &                  this%output_fname_header, this%num_digits_id, id )
    call new( this%log_ploop, fname )
    call print( this%log_ploop, "# Polyakov loop" )
    write(str, &
 &'("#dummy",9X,"Re(Px)",18X,"Im(Px)",18X,"Re(Py)",18X,"Im(Py)",18X,"Re(Pz)",18X,"Im(Pz)",18X,"Re(Pt)",18X,"Im(Pt)")')
    call print( this%log_ploop, TRIM(str) )
  endif

  return
end subroutine new_ploop_measurement_ordinal


subroutine delete_ploop_measurement( this )
  implicit none
  type(ploop_measurement), intent(inout) :: this
  !--------------------------------------------
#ifndef _singlePU
  real(DP) :: copytime

  copytime = get_elapse( copy_pl_time )
#endif

  this%status => NULL()
  if ( associated( this%log_ploop ) ) then
    call delete( this%log_ploop)
    deallocate( this%log_ploop )
    this%log_ploop => NULL()
  endif
  if ( associated( this%hmc_log_ploop ) ) then
    call delete( this%hmc_log_ploop)
    deallocate( this%hmc_log_ploop )
    this%log_ploop => NULL()
  endif
  if (0==nodeid) then
#ifndef _singlePU
    write(*,'(80("="))')
    write(*,'("   Communication Time (ploop) : ", F12.4, " [sec]" )') copytime
#endif
    write(*,'(80("="))')
  endif

  call delete( copy_pl_time )

  return
end subroutine delete_ploop_measurement


subroutine measure_ploop_measurement( this, u )
  implicit none
  type(ploop_measurement), intent(inout) :: this
  type(vfield_gluon_wg), intent(in) :: u
!-----------------------------------------------
  complex(DP) :: plavr(NDIM)
  character(CHARLEN) :: str
  integer :: sweep_count, itraj, mu

  call run( this, u, plavr )

!--- print result ---
  if ( nodeid == 0 ) then
    if ( associated( this%status ) ) then ! onfly measurement
      sweep_count = get_sweep_count( this%status )
      ! This part must be here because run id is revised in the first call of
      ! "is_loop_end"
      if ( sweep_count == 1 ) then
        call print( this%hmc_log_ploop, "# Polyakov loop" )
        write(str, &
 &'("# traj",9X,"Re(Px)",18X,"Im(Px)",18X,"Re(Py)",18X,"Im(Py)",18X,"Re(Pz)",18X,"Im(Pz)",18X,"Re(Pt)",18X,"Im(Pt)")')
        call print( this%hmc_log_ploop, TRIM(str) )
      endif
      itraj = get_trajectory_number(this%status)
      write( str, '(I6,8E24.16)' ) itraj, &
 &      (REAL(plavr(mu)),AIMAG(plavr(mu)),mu=1,NDIM)
      call print( this%hmc_log_ploop, TRIM(str) )
    else
      itraj = 0 ! dummy
      write( str, '(I6,8E24.16)' ) itraj, &
 &      (REAL(plavr(mu)),AIMAG(plavr(mu)),mu=1,NDIM)
      call print( this%log_ploop, TRIM(str) )
    endif
  endif

end subroutine measure_ploop_measurement

subroutine run_ploop_measurement( this, u, plavr )
  implicit none
  type(ploop_measurement), intent(inout) :: this
  type(vfield_gluon_wg), intent(in) :: u
  complex(DP), intent(out) :: plavr(NDIM)
!-----------------------------------------------------------
  type(su3fm) :: wrke, wrko
  complex(DP) :: ctmp0
#ifndef _singlePU
  type(su3fm) ::  ple(PBUFF_MAX_SIZE), plo(PBUFF_MAX_SIZE)
  integer :: isite, idim, itemp, offset
#endif
  integer :: mu, ix, iy, iz, ieoxyzt, itb, ic, jc, itt

!-------- Polyakov loop wrapped along t-direction ----------
  mu = 4
  ctmp0 = Z0
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,wrke,wrko,ic,itt,itb,ieoxyzt,jc) &
!$OMP REDUCTION(+:ctmp0)
  do ix = 1, NX
  do iy = 1, NY
  do iz = 1, NZ

    wrke%u(:,:) = Z0
    wrko%u(:,:) = Z0
    do ic=1,COL
      wrke%u(ic,ic) = Z1
      wrko%u(ic,ic) = Z1
    enddo

    do itt = 1, NTT
      itb = itt/2
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      if ( ieoxyzt == 0 ) then
        do jc=1,COL
        do ic=1,COL
#ifdef _COL3
          wrke%u(ic,jc) = &
 &            wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &          + wrko%u(ic,3) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Debug by yamada from here !!!!!!!!!!!!!!!!!!!!!!!!!!!
!          write (*,*) ic,jc,ix,iy,iz,itb,wrke%u(ic,jc)
!          write (*,*) wrko%u(ic,1),u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc)
!          write (*,*) wrko%u(ic,2),u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
!          write (*,*) wrko%u(ic,3),u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
!          stop
!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Debug by yamada to   here !!!!!!!!!!!!!!!!!!!!!!!!!!!
#else
          wrke%u(ic,jc) = &
 &            wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      else
        do jc=1,COL
        do ic=1,COL
#ifdef _COL3
          wrko%u(ic,jc) = &
 &            wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &          + wrke%u(ic,3) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
          wrko%u(ic,jc) = &
 &            wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      endif
    enddo ! itt

    itt = NTT
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2) + wrke%u(3,3)
#else
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2) + wrko%u(3,3)
#else
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2)
#endif
    endif

  enddo ! iz
  enddo ! iy
  enddo ! ix

#ifndef _singlePU
  call comlib_sumcast(ctmp0)
#endif
  plavr(mu) = ctmp0 / REAL( COL*NTX*NTY*NTZ, KIND=DP)

!-------- Polyakov loop wrapped along z-direction ----------
  mu = 3
  ctmp0 = Z0
#ifndef _singlePU
!$OMP PARALLEL DO PRIVATE(ix,iy,itt,itb,wrke,wrko,ic,iz,ieoxyzt,jc,isite) &
!$OMP REDUCTION(+:ctmp0)
#else
!$OMP PARALLEL DO PRIVATE(ix,iy,itt,itb,wrke,wrko,ic,iz,ieoxyzt,jc) &
!$OMP REDUCTION(+:ctmp0)
#endif
  do ix = 1, NX
  do iy = 1, NY
  do itt = 1, NTT
    itb = itt/2
    wrke%u(:,:) = Z0
    wrko%u(:,:) = Z0
    do ic = 1, COL
      wrke%u(ic,ic) = Z1
      wrko%u(ic,ic) = Z1
    enddo

    do iz = 1, NZ
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      if ( ieoxyzt == 0 ) then
        do jc = 1, COL
        do ic = 1, COL
#ifdef _COL3
           wrke%u(ic,jc) = &
 &             wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &           + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &           + wrko%u(ic,3) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
           wrke%u(ic,jc) = &
 &             wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &           + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      else
        do jc = 1, COL
        do ic = 1, COL
#ifdef _COL3
           wrko%u(ic,jc) = &
 &             wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &           + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &           + wrke%u(ic,3) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
           wrko%u(ic,jc) = &
 &             wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &           + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
         enddo
         enddo
      endif
    enddo ! iz

    iz = NZ
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
#if _NDIMZ == 1
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2) + wrke%u(3,3)
#else
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2) + wrko%u(3,3)
#else
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2)
#endif
    endif
#else
    isite = itb+1 + (iy-1)*(NTH+1) +(ix-1)*(NTH+1)*NY
    if ( ieoxyzt == 0 ) then
      do jc=1,COL
      do ic=1,COL
        pbuffe( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrke%u(ic,jc)
           ple(isite)%u(ic,jc)                           = wrke%u(ic,jc)
      enddo
      enddo
    else
      do jc=1,COL
      do ic=1,COL
        pbuffo( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrko%u(ic,jc)
           plo(isite)%u(ic,jc)                           = wrko%u(ic,jc)
      enddo
      enddo
    endif
#endif

  enddo ! itt
  enddo ! iy
  enddo ! ix

#if _NDIMZ == 1

#ifndef _singlePU
  call comlib_sumcast(ctmp0)
#endif

!!! case of (_NDIMZ > 1)
#else
  do idim = 1, NDIMZ-1

    call tic( copy_pl_time )
    call comlib_barrier
    call comlib_sendrecv( idpsende(mu) )
    call comlib_sendrecv( idpsendo(mu) )
    call toc( copy_pl_time )

    itemp = mod( idim * NZ, 2 )
!$OMP PARALLEL PRIVATE(ix,iy,itt,itb,iz,ieoxyzt,isite,jc,offset,ic,wrke,wrko)
!$OMP DO
    do ix = 1, NX
    do iy = 1, NY
    do itt = 1, NTT
      itb = itt/2

      iz = NZ
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      isite = itb+1 + (iy-1)*(NTH+1) + (ix-1)*(NTH+1)*NY

      if ( ieoxyzt == 0 ) then  ! even
        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
               wrke%u(ic,jc) = &
 &                 ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &               + ple(isite)%u(ic,2) * pbuffe(2+offset,2) &
 &               + ple(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
               wrke%u(ic,jc) = &
 &                 ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &               + ple(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
            else
#ifdef _COL3
               wrke%u(ic,jc) = &
 &                 ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &               + ple(isite)%u(ic,2) * pbuffo(2+offset,2) &
 &               + ple(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
               wrke%u(ic,jc) = &
 &                 ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &               + ple(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
            endif
          enddo ! ic
        enddo ! jc
        ple(isite)%u(:,:) = wrke%u(:,:)

      else ! odd

        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
              wrko%u(ic,jc) = &
     &            plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
     &          + plo(isite)%u(ic,2) * pbuffo(2+offset,2) &
     &          + plo(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
              wrko%u(ic,jc) = &
     &            plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
     &          + plo(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
            else
#ifdef _COL3
              wrko%u(ic,jc) = &
     &            plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
     &          + plo(isite)%u(ic,2) * pbuffe(2+offset,2) &
     &          + plo(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
              wrko%u(ic,jc) = &
     &            plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
     &          + plo(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
            endif
          enddo ! ic
        enddo ! jc
        plo(isite)%u(:,:) = wrko%u(:,:)
      endif

    enddo ! itt
    enddo ! iy
    enddo ! ix
!$OMP END DO
!$OMP END PARALLEL

!--- send dn
    if ( idim < NDIMZ-1 ) then
!$OMP PARALLEL PRIVATE(offset)
!$OMP DO
      do offset = 1, PBUFF_SIZE(mu)
        pbuffe(offset,1) = pbuffe(offset,2)
        pbuffo(offset,1) = pbuffo(offset,2)
      enddo
!$OMP END DO
!$OMP END PARALLEL
    endif

  enddo ! idim

  ctmp0 = Z0
!$OMP PARALLEL DO PRIVATE(ix,iy,itt,itb,iz,ieoxyzt,isite) &
!$OMP REDUCTION(+:ctmp0)
  do ix = 1, NX
  do iy = 1, NY
  do itt = 1, NTT
    itb = itt/2
    iz = NZ
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
    isite = itb+1 + (iy-1)*(NTH+1) +(ix-1)*(NTH+1)*NY
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2) &
 &                  + ple(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2) &
 &                  + plo(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2)
#endif
    endif
  enddo ! itt
  enddo ! iy
  enddo ! ix

  call comlib_sumcast(ctmp0)
  ctmp0 = ctmp0 /REAL( NDIMZ, KIND=DP )
! end of else of "#if _NDIMZ == 1"
#endif
  plavr(mu) = ctmp0 /REAL( COL*NTX*NTY*NTT, KIND=DP )

!------ Polyakov loop wrapped around y-direction --------
  mu = 2
  ctmp0 = Z0
#ifndef _singlePU
!$OMP PARALLEL DO PRIVATE(ix,iz,itt,itb,wrke,wrko,ic,iy,ieoxyzt,jc,isite) &
!$OMP REDUCTION(+:ctmp0)
#else
!$OMP PARALLEL DO PRIVATE(ix,iz,itt,itb,wrke,wrko,ic,iy,ieoxyzt,jc) &
!$OMP REDUCTION(+:ctmp0)
#endif
  do ix = 1, NX
  do iz = 1, NZ
  do itt = 1, NTT
    itb = itt /2
    wrke%u(:,:) = Z0
    wrko%u(:,:) = Z0
    do ic = 1, COL
       wrke%u(ic,ic) = Z1
       wrko%u(ic,ic) = Z1
    enddo

    do iy = 1, NY
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      if ( ieoxyzt == 0 ) then
        do jc = 1, COL
        do ic = 1, COL
#ifdef _COL3
          wrke%u(ic,jc) =  &
 &              wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &            + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &            + wrko%u(ic,3) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
          wrke%u(ic,jc) =  &
 &              wrko%u(ic,1) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &            + wrko%u(ic,2) *u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      else
        do jc=1,COL
        do ic=1,COL
#ifdef _COL3
          wrko%u(ic,jc) = &
 &              wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &            + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &            + wrke%u(ic,3) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
          wrko%u(ic,jc) = &
 &              wrke%u(ic,1) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &            + wrke%u(ic,2) *u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      endif
    enddo ! iy

    iy = NY
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
#if _NDIMY == 1
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2) + wrke%u(3,3)
#else
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2) + wrko%u(3,3)
#else
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2)
#endif
    endif
#else
    isite = itb+1 + (iz-1)*(NTH+1) + (ix-1)*(NTH+1)*NZ
    if ( ieoxyzt == 0 ) then
      do jc=1,COL
      do ic=1,COL
        pbuffe( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrke%u(ic,jc)
           ple(isite)%u(ic,jc)                           = wrke%u(ic,jc)
      enddo
      enddo
    else
      do jc=1,COL
      do ic=1,COL
        pbuffo( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrko%u(ic,jc)
           plo(isite)%u(ic,jc)                           = wrko%u(ic,jc)
      enddo
      enddo
    endif
#endif
  enddo ! itt
  enddo ! iz
  enddo ! ix

#if _NDIMY == 1

#ifndef _singlePU
  call comlib_sumcast(ctmp0)
#endif

!!! case of (_NDIMY > 1)
#else
  do idim = 1, NDIMY-1

    call tic( copy_pl_time )
    call comlib_barrier
    call comlib_sendrecv( idpsende(mu) )
    call comlib_sendrecv( idpsendo(mu) )
    call toc( copy_pl_time )

    itemp = mod( idim * NY, 2 )
!$OMP PARALLEL PRIVATE(ix,iz,itt,itb,iy,ieoxyzt,isite,jc,offset,ic,wrke,wrko)
!$OMP DO
    do ix = 1, NX
    do iz = 1, NZ
    do itt = 1, NTT
      itb = itt/2

      iy = NY
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      isite = itb+1 + (iz-1)*(NTH+1) + (ix-1)*(NTH+1)*NZ

      if ( ieoxyzt == 0 ) then  ! even
        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
              wrke%u(ic,jc) = &
 &                ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &              + ple(isite)%u(ic,2) * pbuffe(2+offset,2) &
 &              + ple(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
              wrke%u(ic,jc) = &
 &                ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &              + ple(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
            else
#ifdef _COL3
              wrke%u(ic,jc) = &
 &                ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &              + ple(isite)%u(ic,2) * pbuffo(2+offset,2) &
 &              + ple(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
              wrke%u(ic,jc) = &
 &                ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &              + ple(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
            endif
          enddo ! ic
        enddo ! jc
        ple(isite)%u(:,:) = wrke%u(:,:)

      else ! odd

        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
              wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffo(2+offset,2) &
 &             + plo(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
              wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
            else
#ifdef _COL3
               wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffe(2+offset,2) &
 &             + plo(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
               wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
            endif
          enddo ! ic
        enddo ! jc
        plo(isite)%u(:,:) = wrko%u(:,:)
      endif

    enddo ! itt
    enddo ! iz
    enddo ! ix
!$OMP END DO
!$OMP END PARALLEL

!--- send dn
    if ( idim < NDIMY-1 ) then
!$OMP PARALLEL PRIVATE(offset)
!$OMP DO
      do offset = 1, PBUFF_SIZE(mu)
        pbuffe(offset,1) = pbuffe(offset,2)
        pbuffo(offset,1) = pbuffo(offset,2)
      enddo
!$OMP END DO
!$OMP END PARALLEL
    endif

  enddo ! idim

  ctmp0 = Z0
!$OMP PARALLEL DO PRIVATE(ix,iz,itt,itb,iy,ieoxyzt,isite) &
!$OMP REDUCTION(+:ctmp0)
  do ix = 1, NX
  do iz = 1, NZ
  do itt = 1, NTT
    itb = itt/2
    iy = NY
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
    isite = itb+1 + (iz-1)*(NTH+1) + (ix-1)*(NTH+1)*NZ
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2) &
 &                  + ple(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2) &
 &                  + plo(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2)
#endif
    endif
  enddo ! itt
  enddo ! iz
  enddo ! ix

  call comlib_sumcast(ctmp0)
  ctmp0 = ctmp0 /REAL( NDIMY, KIND=DP )
! end of else of "#if _NDIMY == 1"
#endif
  plavr(mu) = ctmp0 /REAL( COL*NTX*NTZ*NTT, KIND=DP )

!------ Polyakov loop wrapped around x-direction --------
  mu = 1
  ctmp0 = Z0
#ifndef _singlePU
!$OMP PARALLEL DO PRIVATE(iy,iz,itt,itb,wrke,wrko,ic,ix,ieoxyzt,jc,isite) &
!$OMP REDUCTION(+:ctmp0)
#else
!$OMP PARALLEL DO PRIVATE(iy,iz,itt,itb,wrke,wrko,ic,ix,jc,ieoxyzt) &
!$OMP REDUCTION(+:ctmp0)
#endif
  do iy = 1, NY
  do iz = 1, NZ
  do itt = 1, NTT
    itb = itt /2
    wrke%u(:,:) = Z0
    wrko%u(:,:) = Z0
    do ic = 1, COL
      wrke%u(ic,ic) = Z1
      wrko%u(ic,ic) = Z1
    enddo

    do ix = 1, NX
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      if ( ieoxyzt == 0 ) then
        do jc = 1, COL
        do ic = 1, COL
#ifdef _COL3
          wrke%u(ic,jc) = &
 &           wrko%u(ic,1) * u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &         + wrko%u(ic,2) * u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &         + wrko%u(ic,3) * u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
          wrke%u(ic,jc) = &
 &           wrko%u(ic,1) * u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &         + wrko%u(ic,2) * u%eo(0)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      else
        do jc=1,COL
        do ic=1,COL
#ifdef _COL3
           wrko%u(ic,jc) = &
 &            wrke%u(ic,1) * u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrke%u(ic,2) * u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc) &
 &          + wrke%u(ic,3) * u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(3,jc)
#else
           wrko%u(ic,jc) = &
 &            wrke%u(ic,1) * u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(1,jc) &
 &          + wrke%u(ic,2) * u%eo(1)%mu(mu)%s(itb,iz,iy,ix)%u(2,jc)
#endif
        enddo
        enddo
      endif
    enddo ! ix

    ix = NX
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
#if _NDIMX == 1
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2) + wrke%u(3,3)
#else
      ctmp0 = ctmp0 + wrke%u(1,1) + wrke%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2) + wrko%u(3,3)
#else
      ctmp0 = ctmp0 + wrko%u(1,1) + wrko%u(2,2)
#endif
    endif
#else
    isite = itb+1 + (iz-1)*(NTH+1) + (iy-1)*(NTH+1)*NZ
    if ( ieoxyzt == 0 ) then
      do jc=1,COL
      do ic=1,COL
        pbuffe( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrke%u(ic,jc)
           ple(isite)%u(ic,jc)                           = wrke%u(ic,jc)
      enddo
      enddo
    else
      do jc=1,COL
      do ic=1,COL
        pbuffo( ic + (jc-1)*COL + COL*COL*(isite-1), 1 ) = wrko%u(ic,jc)
          plo(isite)%u(ic,jc)                            = wrko%u(ic,jc)
      enddo
      enddo
    endif
#endif
  enddo ! itt
  enddo ! iz
  enddo ! iy

#if _NDIMX == 1

#ifndef _singlePU
  call comlib_sumcast(ctmp0)
#endif

!!! case of (_NDIMX > 1)
#else
  do idim = 1, NDIMX-1

    call tic( copy_pl_time )
    call comlib_barrier
    call comlib_sendrecv( idpsende(mu) )
    call comlib_sendrecv( idpsendo(mu) )
    call toc( copy_pl_time )

    itemp = mod( idim * NX, 2 )
!$OMP PARALLEL PRIVATE(iy,iz,itt,itb,ix,ieoxyzt,isite,jc,offset,ic,wrke,wrko)
!$OMP DO
    do iy = 1, NY
    do iz = 1, NZ
    do itt = 1, NTT
      itb = itt/2

      ix = NX
      ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
      isite = itb+1 + (iz-1)*(NTH+1) + (iy-1)*(NTH+1)*NZ

      if ( ieoxyzt == 0 ) then ! even
        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
              wrke%u(ic,jc) = &
 &               ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + ple(isite)%u(ic,2) * pbuffe(2+offset,2) &
 &             + ple(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
              wrke%u(ic,jc) = &
 &               ple(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + ple(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
            else
#ifdef _COL3
              wrke%u(ic,jc )= &
 &               ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + ple(isite)%u(ic,2) * pbuffo(2+offset,2) &
 &             + ple(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
              wrke%u(ic,jc )= &
 &               ple(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + ple(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
            endif
          enddo ! ic
        enddo ! jc
        ple(isite)%u(:,:) = wrke%u(:,:)

      else ! odd

        do jc=1,COL
          offset = (jc-1)*COL + (isite-1)*COL*COL
          do ic=1,COL
            if ( itemp == 0 ) then
#ifdef _COL3
              wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffo(2+offset,2) &
 &             + plo(isite)%u(ic,3) * pbuffo(3+offset,2)
#else
              wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffo(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffo(2+offset,2)
#endif
             else
#ifdef _COL3
               wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffe(2+offset,2) &
 &             + plo(isite)%u(ic,3) * pbuffe(3+offset,2)
#else
               wrko%u(ic,jc) = &
 &               plo(isite)%u(ic,1) * pbuffe(1+offset,2) &
 &             + plo(isite)%u(ic,2) * pbuffe(2+offset,2)
#endif
             endif
          enddo ! ic
        enddo ! jc
        plo(isite)%u(:,:) = wrko%u(:,:)
      endif

    enddo ! itt
    enddo ! iz
    enddo ! iy
!$OMP END DO
!$OMP END PARALLEL

!-- send dn
    if ( idim < NDIMX-1 ) then
!$OMP PARALLEL PRIVATE(offset)
!$OMP DO
      do offset = 1, PBUFF_SIZE(mu)
        pbuffe(offset,1) = pbuffe(offset,2)
        pbuffo(offset,1) = pbuffo(offset,2)
      enddo
!$OMP END DO
!$OMP END PARALLEL
    endif

  enddo ! idim

  ctmp0 = Z0
!$OMP PARALLEL DO PRIVATE(iy,iz,itt,itb,ix,ieoxyzt,isite) &
!$OMP REDUCTION(+:ctmp0)
  do iy = 1, NY
  do iz = 1, NZ
  do itt = 1, NTT
    itb = itt/2
    ix = NX
    ieoxyzt = mod(ipeo+ix+iy+iz+itt,2)
    isite = itb+1 + (iz-1)*(NTH+1) + (iy-1)*(NTH+1)*NZ
    if ( ieoxyzt == 0 ) then
#ifdef _COL3
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2) &
 &                  + ple(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + ple(isite)%u(1,1) + ple(isite)%u(2,2)
#endif
    else
#ifdef _COL3
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2) &
 &                  + plo(isite)%u(3,3)
#else
      ctmp0 = ctmp0 + plo(isite)%u(1,1) + plo(isite)%u(2,2)
#endif
    endif
  enddo ! itt
  enddo ! iz
  enddo ! iy

  call comlib_sumcast(ctmp0)
  ctmp0 = ctmp0 / REAL( NDIMX, KIND=DP )
#endif
  plavr(mu) = ctmp0 / REAL( COL*NTY*NTZ*NTT, KIND=DP )

  return
end subroutine run_ploop_measurement


end module ploop_class
