#include <config.h>
module quark_clover_class
!
!= Define Wilson-Clover quark action parameters and related routines
!
!== version
!
! $Id: quark_clover_class.F90,v 1.25 2011/09/25 06:13:11 ishikawa Exp $
!
  use lattice_class
  use error_class
  use hmc_logfile_class
  use hmc_status_class
  use field_gauge_class
  use field_fermion_class
  use field_clover_class
  use quark_wilson_class
  implicit none
#ifdef _CLOVER
  private

  public :: quark_wilson
  public :: quark_clover
  public :: new
  public :: delete
  public :: read
  public :: save_config
  public :: read_config
  public :: print

  public :: get_id
  public :: set_id
  public :: get_kappa
  public :: set_kappa
  public :: get_nflavor
  public :: set_nflavor
  public :: get_ptr_inverse_clover_term
  public :: get_csw
  public :: set_csw
#ifdef _SF
  public :: get_cttilde
  public :: set_cttilde
#endif
  public :: assign
  public :: set_gaussian_noise
  public :: assign_mult_hop
  public :: assign_mult_eoprec_wd
  public :: assign_mult_eoprec_wd_dag
  public :: assign_mult_Qee
  public :: assign_mult_Oee
  public :: assign_mult_clover_term
  public ::  accum_mult_clover_term

  public :: force_clover_terms, force_clover_trlog, force_hmc_clover
  public :: hamil_clover_trlog
  public ::   make_clover_leaf
  public :: delete_clover_leaf
  public ::   make_inverse_clover_term
  public :: delete_inverse_clover_term

  character(len=*), parameter :: ACTION_NAME = "Clover Quark"

  !
  ! clover leaf term (field strength). module data.
  !
  type(field_clover_leaf), save, pointer :: m_ucl  => NULL()
  integer, parameter :: NHSITE=NTH*NZ*NY*NX

  type f_c_m_eo
    complex(DP) :: m(CLSP/2,CLSP/2,NHSITE)
    integer     :: ieo
  end type

  type, extends(quark_wilson) :: quark_clover
  !
  ! Clover quark action parameters (extends from quark_wilson)
  !
    real(DP),                         private :: csw                  ! C_SW parameter
#ifdef _SF
    real(DP),                         private :: cttilde = 1.0_DP     ! SF boundary O(a)-improvement coefficient
#endif
    type(field_clover_term), pointer, private :: fclinv => NULL()     ! inverse clover term
    type(hmc_logfile),                private :: clover_term_log      ! logfile for clover term inversion
    character(CHARLEN),               private :: clover_term_log_fname = "cq_clv_term_log."
  end type

  interface new
    module procedure new_quark_clover
  end interface

  interface delete
    module procedure delete_quark_clover
  end interface

  interface read
  !
  ! read quark action parameters in formatted form from a unit.
  !
    module procedure read_quark_clover
  end interface

  interface save_config
  !
  ! save quark action parameters to configuration file in unformatted form
  !
    module procedure save_config_quark_clover
  end interface

  interface read_config
  !
  ! read quark action parameters to configuration file in unformatted form
  !
    module procedure read_config_quark_clover
  end interface

  interface assign
  !
  ! copy and assign clover quark parameters
  !
    module procedure assign_quark_clover
  end interface

  interface print
  !
  ! print out clover quark action parameters on display
  !
    module procedure print_quark_clover
  end interface


  interface assign_mult_clover_term
  !
  ! multiply inverse clover term and assign
  !
  !  we <= Fee \ ye
  !
    module procedure assign_mult_clover_term_eo
    module procedure assign_mult_clover_term_all
  end interface

  interface accum_mult_clover_term
  !
  ! multiply inverse clover term (overwrite)
  !
  !  ye <= Fee \ ye
  !
    module procedure accum_mult_clover_term_eo
    module procedure accum_mult_clover_term_all
  end interface

  interface assign_mult_hop
  !
  ! multiply hopping matrix and inverse clover term
  !
  !  Mye <= Fee \ Meo yo
  !
    module procedure assign_mult_clover_hopping_eo
  end interface

  interface assign_mult_eoprec_wd
  !
  ! multiply even/odd-site preconditioned Wilson-Dirac(Clover) operator
  !
  !  Dye <= (1 - kappa^2 Fee \ Meo Foo \ Moe)ye
  !
    module procedure assign_mult_eoprec_cd_eo
  end interface

  interface assign_mult_eoprec_wd_dag
  !
  ! multiply Hermitian conjugate of even/odd-site preconditioned
  ! Wilson-Dirac(Clover) operator
  !
  !  Dye <= (1 - kappa^2 gamma5 Meo Foo \ Moe gamma5 Fee\ )ye
  !
    module procedure assign_mult_eoprec_cq_eo_dag
  end interface

  interface assign_mult_Qee
  !       v     v
  ! Qee = Dee^+ Dee
  !
    module procedure assign_mult_DdagD_ee_cq
  end interface

  interface assign_mult_Oee
  !       v   v
  ! Oee = Dee Dee^+
  !
    module procedure assign_mult_DDdag_ee_cq
  end interface

  interface get_ptr_inverse_clover_term
    module procedure get_ptr_inverse_clover_term
  end interface

  interface get_csw
    module procedure get_csw
  end interface

  interface set_csw
    module procedure set_csw
  end interface

#ifdef _SF
  interface get_cttilde
    module procedure get_cttilde
  end interface

  interface set_cttilde
    module procedure set_cttilde
  end interface
#endif

  interface get_kappa
    module procedure get_kappa_quark_clover
  end interface

  interface set_kappa
    module procedure set_kappa_quark_clover
  end interface

  interface set_nflavor
    module procedure set_nflavor_quark_clover
  end interface

  interface get_nflavor
    module procedure get_nflavor_quark_clover
  end interface

contains

function get_ptr_inverse_clover_term(this) result(ptr)
  implicit none
  type(quark_clover),       intent(in) :: this
  type(field_clover_term),  pointer    :: ptr
  ptr => this%fclinv
  return
end function

function get_csw(this) result(csw)
  implicit none
  type(quark_clover), intent(in) :: this
  real(DP) :: csw
  csw = this%csw
  return
end function

subroutine set_csw(this,csw)
  implicit none
  type(quark_clover), intent(inout) :: this
  real(DP), intent(in) :: csw
  this%csw = csw
  return
end subroutine

#ifdef _SF
function get_cttilde(this) result(cttilde)
  implicit none
  type(quark_clover), intent(in) :: this
  real(DP) :: cttilde
  cttilde = this%cttilde
  return
end function

subroutine set_cttilde(this,cttilde)
  implicit none
  type(quark_clover), intent(inout) :: this
  real(DP), intent(in) :: cttilde
  this%cttilde = cttilde
  return
end subroutine
#endif

function get_kappa_quark_clover(this) result(kappa)
  implicit none
  type(quark_clover), intent(in) :: this
  real(DP) :: kappa
  kappa = get_kappa(this%quark_wilson)
  return
end function

subroutine set_kappa_quark_clover(this,kappa)
  implicit none
  type(quark_clover), intent(inout) :: this
  real(DP), intent(in) :: kappa
  call set_kappa(this%quark_wilson,kappa)
  return
end subroutine

function get_nflavor_quark_clover(this) result(nf)
  implicit none
  type(quark_clover), intent(in) :: this
  integer :: nf
  nf = get_nflavor(this%quark_wilson)
  return
end function

subroutine set_nflavor_quark_clover(this,nf)
  implicit none
  type(quark_clover), intent(inout) :: this
  integer,            intent(in)    :: nf
  call set_nflavor(this%quark_wilson,nf)
  return
end subroutine

subroutine assign_quark_clover(qp1,qp2)
  implicit none
  type(quark_clover), intent(inout) :: qp1
  type(quark_clover), intent(in)    :: qp2
  call assign(qp1%quark_wilson,qp2%quark_wilson)
  qp1%csw                   =  qp2%csw
  qp1%fclinv                => qp2%fclinv
  qp1%clover_term_log       =  qp2%clover_term_log
  qp1%clover_term_log_fname =  qp2%clover_term_log_fname
  return
end subroutine

subroutine new_quark_clover(this,id)
  implicit none
  type(quark_clover), intent(inout) :: this
  integer,            intent(in)    :: id
  character(len=CHARLEN) :: str
  call new(this%quark_wilson,id)
  this%action_name = ACTION_NAME
  write(str,'(I2.2)')id
  this%clover_term_log_fname = TRIM(this%clover_term_log_fname)//TRIM(str)
  call new(this%clover_term_log,this%clover_term_log_fname)
  return
end subroutine

subroutine delete_quark_clover(this)
  implicit none
  type(quark_clover), intent(inout) :: this
  call delete(this%quark_wilson)
  call delete(this%clover_term_log)
  return
end subroutine

subroutine read_quark_clover(this,iout)
  use comlib
  implicit none
  type(quark_clover), intent(inout) :: this
  integer, intent(in) :: iout
  call read(this%quark_wilson,iout)
  if (0 == nodeid) then
#ifdef _SF
    read(iout,*)this%csw, this%cttilde
#else
    read(iout,*)this%csw
#endif
  endif
#ifndef _singlePU
  call comlib_bcast(this%csw,0)
#endif
  return
end subroutine

subroutine save_config_quark_clover(this,iout)
  implicit none
  type(quark_clover), intent(in) :: this
  integer, intent(in) :: iout
  call save_config(this%quark_wilson,iout)
#ifdef _SF
  write(iout)this%csw, this%cttilde
#else
  write(iout)this%csw
#endif
  return
end subroutine

subroutine read_config_quark_clover(this,iout)
  implicit none
  type(quark_clover), intent(inout) :: this
  integer, intent(in) :: iout
  call read_config(this%quark_wilson,iout)
#ifdef _SF
  read(iout)this%csw, this%cttilde
#else
  read(iout)this%csw
#endif
  return
end subroutine

subroutine print_quark_clover(this)
  implicit none
  type(quark_clover), intent(inout) :: this
  call print(this%quark_wilson)
    write(*,'(14X,"              Csw :",F12.8)')this%csw
#ifdef _SF
  write(*,'(12X,"SF boundary cttilde :",F12.8)')this%cttilde
#endif
  return
end subroutine

subroutine force_clover_terms(BB,XX,u,wl34,wl98)
!
! Compute Force for clover term contributions.
!
!   BB_{mu}(n) = Sum[vcl_{mu,nu}(n),nu=!mu]
!
!         X4---<---3X       Xn       2X
!          |       |         |       |         
!   vcl =  v       ^    -    ^       v
!          |       |         |       |         
!         Xn       2X       X6---<---5X
!
!   X : X_{mu,nu}(n) - X^+_{mu,nu}(n),
!   where
!       X_{mu,nu}(n) = \sigma_{mu,nu} (force contributions at site n)
!
  implicit none
  type(vfield_gluon_wg), intent(inout) :: BB      ! pre-force contirubution (\dot{u}) accumulated.
  type(tfield_gluon_wg), intent(inout) :: XX      ! pre-force contirubution (\dot{\sigma_{mu,nu}F_{mu,nu}) accumulated.
  type(vfield_gluon_wg), intent(in)    ::  u          ! gauge field
  type(tfield_gluon_wg), intent(in)    :: wl34,wl98   ! 2-link fields prepared by make_two_links

  integer :: ipl

  do ipl=1,NDIM*(NDIM-1)/2
    call copy_boundary(XX%eo(0)%munu(ipl))
    call copy_boundary(XX%eo(1)%munu(ipl))
  enddo

  call force_clover_terms_eo(BB%eo(0),XX%eo(0),XX%eo(1),u%eo(0),u%eo(1),wl34%eo(0),wl98%eo(1))
  call force_clover_terms_eo(BB%eo(1),XX%eo(1),XX%eo(0),u%eo(1),u%eo(0),wl34%eo(1),wl98%eo(0))

  return
end subroutine


subroutine force_clover_terms_eo(BBe,XXe,XXo,ue,uo,wl34e,wl98o)
!
! Force for clover term contributions. (even/odd site only)
!
!  BB_{mu}(n) = Sum[vcl_{mu,nu}(n),nu=!mu]
!
!        X4---<---3X       Xn       2X
!         |       |         |       |         
!  vcl =  v       ^    -    ^       v
!         |       |         |       |         
!        Xn       2X       X6---<---5X
!
!  X : X_{mu,nu}(n) - X^+_{mu,nu}(n),
!  where
!   X_{mu,nu}(n) = \sigma_{mu,nu} (force contributions at site n)
! 
  implicit none
  type(vfield_gluon_eo_wg), intent(inout) :: BBe  ! even/odd site force contirubution (\dot{u})
  type(tfield_gluon_eo_wg), intent(in) :: XXe,XXo ! even/odd site force contirubution (\dot{\sigma_{mu,nu}F_{mu,nu})
  type(vfield_gluon_eo_wg), intent(in) :: ue,uo   ! even/odd site gauge link
  type(tfield_gluon_eo_wg), intent(in) :: wl34e   ! even/odd site 2-links conecting 2-3-4 sites
  type(tfield_gluon_eo_wg), intent(in) :: wl98o   ! odd/even site 2-links conecting 6-9-8 sites

  integer :: mu,nu,ipl
  type(sfield_gluon_eo_wg), allocatable :: w1e,w2e,w3e
  type(sfield_gluon_eo_wg), allocatable :: X1e,X2e,X3e,X4e,X5e,X6e
  type(sfield_gluon_eo_wg), allocatable :: vcle

  if (.not. ((BBe%mu(1)%ieo   ==   XXe%munu(1)%ieo) .and.  &
 &           (BBe%mu(1)%ieo   ==    ue%mu(1)%ieo)   .and.  &
 &           (BBe%mu(1)%ieo   == wl34e%munu(1)%ieo) .and.  &
 &           (BBe%mu(1)%ieo   /= wl98o%munu(1)%ieo) .and.  &
 &           ( ue%mu(1)%ieo   /=    uo%mu(1)%ieo)   .and.  &
 &           (XXe%munu(1)%ieo /=   XXo%munu(1)%ieo)) ) then
    call error_stop("error in force_clover_terms_eo(even/odd).")
  endif

  allocate(w1e,w2e,w3e,X1e,X2e,X3e,X4e,X5e,X6e,vcle)
  call new(vcle,BBe%ieo)

!*************************************************************************
!* Calc force from clover term of multi boson and inverse clover term boson
!*
!*  *4---<---3*       *n       2*
!*   |       |         |       |         
!*   v       ^    -    ^       v  = vcl
!*   |       |         |       |         
!*  *n       2*       *6---<---5*
!*
!*  * : X_{mu,nu} - X^+_{mu,nu}
!*  where
!*   X_{mu,nu} = \sigma_{mu,nu} (I csw kappa) (force contrib.)
!*
!*************************************************************************
  do mu=1,NDIM
    call clear(vcle)
    do nu=1,NDIM
    if (mu == nu) cycle

      ipl=imunu(mu,nu)

      if ( ipl > 0 ) then
        call       assign(w1e,XXe%munu( ipl))
        call assign_conjg(w2e,w1e)
        call   assign_sub(X1e,w1e,w2e)
        call assign_shift(w1e,XXo%munu( ipl),mu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X2e,w1e,w2e)
        call assign_shift(w1e,XXe%munu( ipl),mu,nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X3e,w1e,w2e)
        call assign_shift(w1e,XXo%munu( ipl),nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X4e,w1e,w2e)
        call assign_shift(w1e,XXe%munu( ipl),mu,-nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X5e,w1e,w2e)
        call assign_shift(w1e,XXo%munu( ipl),-nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X6e,w1e,w2e)
      else
        call       assign(w1e,XXe%munu(-ipl))
        call assign_conjg(w2e,w1e)
        call   assign_sub(X1e,w2e,w1e)
        call assign_shift(w1e,XXo%munu(-ipl),mu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X2e,w2e,w1e)
        call assign_shift(w1e,XXe%munu(-ipl),mu,nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X3e,w2e,w1e)
        call assign_shift(w1e,XXo%munu(-ipl),nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X4e,w2e,w1e)
        call assign_shift(w1e,XXe%munu(-ipl),mu,-nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X5e,w2e,w1e)
        call assign_shift(w1e,XXo%munu(-ipl),-nu)
        call assign_conjg(w2e,w1e)
        call   assign_sub(X6e,w2e,w1e)
      endif

!***********************
!*  X2w34X4 = (X2-X2^+)w34 + w34(X4-X4^+)
!*  vcl <= (X2w34X4) u^{+}_nu(n)
!***********************
      if ( ipl > 0 ) then
        call       assign(w1e,wl34e%munu( ipl))
      else
        call assign_conjg(w1e,wl34e%munu(-ipl))
      endif ! end of if ipl
      call        assign_mult(w2e,X2e,w1e)
      call        assign_mult(w3e,w1e,X4e)
      call         assign_add(w1e,w2e,w3e)
      call assign_conjg2_mult(w2e,w1e,ue%mu(nu))
      call accum_add(vcle,w2e)

!******************
!*
!*   4---<---3       
!*   |               
!*   v  w41
!*   |               
!*   n       2       
!*
!******************
      call assign_shift(w1e,uo%mu(mu),nu)
      call  assign_mult(w2e,ue%mu(nu),w1e)
      call assign_conjg(w1e,w2e)
!***********************
!*  X3w41X1 = (X3-X3^+)w41 + w41(X1-X1^+)
!*  vcl <= u_nu(n+mu) (X3w41X1)
!***********************
      call  assign_mult(w2e,X3e,w1e)
      call  assign_mult(w3e,w1e,X1e)
      call   assign_add(w1e,w2e,w3e)
      call assign_shift(w2e,uo%mu(nu),mu)
      call  assign_mult(w3e,w2e,w1e)
      call accum_add(vcle,w3e)

!******************
!*
!*   n       2
!*           |
!*      w56  v 
!*           |
!*   6---<---5
!*
!******************
      call assign_shift(w1e,uo%mu(mu),-nu)
      call assign_shift(w2e,ue%mu(nu),-nu,mu)
      call  assign_mult(w3e,w1e,w2e)
      call assign_conjg(w1e,w3e)
!***********************
!*  X2w56X6 = (X2-X2^+)w56 + w56(X6-X6^+)
!*  vcl <= -(X2w56X6) u_nu(n-nu)
!***********************
      call  assign_mult(w2e,X2e,w1e)
      call  assign_mult(w3e,w1e,X6e)
      call   assign_add(w1e,w2e,w3e)
      call assign_shift(w2e,uo%mu(nu),-nu)
      call  assign_mult(w3e,w1e,w2e)
      call accum_sub(vcle,w3e)

!***********************
!*  X5w61X1 = (X5-X5^+)w61 + w61(X1-X1^+)
!*  vcl <= - u^{+}_nu(n+mu-nu) (X5w61X1)
!***********************
      if (ipl > 0) then
        call       assign_shift(w1e,wl98o%munu( ipl),mu)
      else
        call assign_conjg_shift(w1e,wl98o%munu(-ipl),mu)
      endif
      call  assign_mult(w2e,X5e,w1e)
      call  assign_mult(w3e,w1e,X1e)
      call   assign_add(w1e,w2e,w3e)
      call assign_shift(w2e,ue%mu(nu),mu,-nu)
      call assign_conjg1_mult(w3e,w2e,w1e)
      call accum_sub(vcle,w3e)

    enddo    ! end of do nu

!**********************************
!*  BB = BB - vcl
!**********************************
    call accum_sub(BBe%mu(mu),vcle)

  enddo ! end of do mu

  deallocate(w1e,w2e,w3e,X1e,X2e,X3e,X4e,X5e,X6e,vcle)

  return
end subroutine

subroutine force_clover_trlog(this,XX)
!
! Calc force contiribution from log determinant of clover term derived
! from even/odd site preconditioning.
!
! Action:
!
!      SF = - Nf log[det[F]]
!
  implicit none
  type(quark_clover),    intent(in)    :: this ! quark parameters
  type(tfield_gluon_wg), intent(inout) :: XX   ! pre-force contribution (\dot{\sigma_{mu,nu}F_{mu,nu}})

  call force_clover_trlog_eo(this,this%fclinv%eo(0),XX%eo(0))
  call force_clover_trlog_eo(this,this%fclinv%eo(1),XX%eo(1))

  return
end subroutine

#define _X_ itb0,iz,iy,ix
subroutine force_clover_trlog_eo(this,fclinve,XXe)
!
! Calc force contiribution from log determinant of clover term derived
! from even/odd site preconditioning (even/odd site only).
!
! Action:
!
!      SF = - Nf log[det[F]]
!
  implicit none
  type(quark_clover),         intent(in)    :: this    ! quark parameters
  type(field_clover_term_eo), intent(in)    :: fclinve ! even site inverse clover term matrix (chiral rep, linear form)
  type(tfield_gluon_eo_wg),   intent(inout) :: XXe     ! even site force contribution (\dot{\sigma_{mu,nu}F_{mu,nu}})

  complex(DP) :: zickd8
  real(DP) :: rrckd8
  real(DP) :: kappa,csw
  integer :: ix,iy,iz,itb,ieoxyz,itb0
  integer :: nf

  if (.not. ((fclinve%ieo == XXe%munu(1)%ieo) .and. &
 &           (fclinve%ieo == XXe%munu(2)%ieo) .and. &
 &           (fclinve%ieo == XXe%munu(3)%ieo) .and. &
 &           (fclinve%ieo == XXe%munu(4)%ieo) .and. &
 &           (fclinve%ieo == XXe%munu(5)%ieo) .and. &
 &           (fclinve%ieo == XXe%munu(6)%ieo)) ) then
    call error_stop("error in force_clover_trlog_eo(even/odd)")
  endif
!*******************************
!* Nf/2 (i csw kappa/8) * 2
!*******************************
  nf    = get_nflavor(this%quark_wilson)
  kappa =   get_kappa(this%quark_wilson)
  csw   = this%csw
  zickd8=+nf*zi*csw*kappa/8.0_DP

!*******************************
!* Nf/2 (i csw kappa/8) * 2 * i
!*******************************
  rrckd8=-nf*   csw*kappa/8.0_DP

!*************************************************************
!*
!*  XX(n) = (Nf/2)(i csw kappa /8) tr[sigma_{mu,nu}F^-1]
!*
!*************************************************************
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,itb0) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+fclinve%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      itb0 = itb+ieoxyz

!**** ipl = 1, (mu,nu)=(1,2)
      XXe%munu(1)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(1)%s(itb,iz,iy,ix)%u(1,1)  &
  &      -zickd8 *       ( fclinve%l( 1,1,_X_)+fclinve%l( 1,2,_X_)  &
  &                       -fclinve%l(16,1,_X_)-fclinve%l(16,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(1)%s(itb,iz,iy,ix)%u(1,2)  &
  &      -zickd8 *       ( fclinve%l( 2,1,_X_)+fclinve%l( 2,2,_X_)  &
  &                       -fclinve%l(17,1,_X_)-fclinve%l(17,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(1)%s(itb,iz,iy,ix)%u(1,3)  &
  &      -zickd8 *       ( fclinve%l( 3,1,_X_)+fclinve%l( 3,2,_X_)  &
  &                       -fclinve%l(18,1,_X_)-fclinve%l(18,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(1)%s(itb,iz,iy,ix)%u(2,1)  &
  &      -zickd8 *  conjg( fclinve%l( 2,1,_X_)+fclinve%l( 2,2,_X_)  &
  &                       -fclinve%l(17,1,_X_)-fclinve%l(17,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(1)%s(itb,iz,iy,ix)%u(2,2)  &
  &      -zickd8 *       ( fclinve%l( 7,1,_X_)+fclinve%l( 7,2,_X_)  &
  &                       -fclinve%l(19,1,_X_)-fclinve%l(19,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(1)%s(itb,iz,iy,ix)%u(2,3)  &
  &      -zickd8 *       ( fclinve%l( 8,1,_X_)+fclinve%l( 8,2,_X_)  &
  &                       -fclinve%l(20,1,_X_)-fclinve%l(20,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(1)%s(itb,iz,iy,ix)%u(3,1)  &
  &      -zickd8 *  conjg( fclinve%l( 3,1,_X_)+fclinve%l( 3,2,_X_)  &
  &                       -fclinve%l(18,1,_X_)-fclinve%l(18,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(1)%s(itb,iz,iy,ix)%u(3,2)  &
  &      -zickd8 *  conjg( fclinve%l( 8,1,_X_)+fclinve%l( 8,2,_X_)  &
  &                       -fclinve%l(20,1,_X_)-fclinve%l(20,2,_X_))
      XXe%munu(1)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(1)%s(itb,iz,iy,ix)%u(3,3)  &
  &      -zickd8 *       ( fclinve%l(12,1,_X_)+fclinve%l(12,2,_X_)  &
  &                       -fclinve%l(21,1,_X_)-fclinve%l(21,2,_X_))

!**** ipl = 6, (mu,nu)=(3,4)
      XXe%munu(6)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(6)%s(itb,iz,iy,ix)%u(1,1)  &
  &      -zickd8 *       ( fclinve%l( 1,2,_X_)-fclinve%l( 1,1,_X_)  &
  &                      -(fclinve%l(16,2,_X_)-fclinve%l(16,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(6)%s(itb,iz,iy,ix)%u(1,2)  &
  &      -zickd8 *       ( fclinve%l( 2,2,_X_)-fclinve%l( 2,1,_X_)  &
  &                      -(fclinve%l(17,2,_X_)-fclinve%l(17,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(6)%s(itb,iz,iy,ix)%u(1,3)  &
  &      -zickd8 *       ( fclinve%l( 3,2,_X_)-fclinve%l( 3,1,_X_)  &
  &                      -(fclinve%l(18,2,_X_)-fclinve%l(18,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(6)%s(itb,iz,iy,ix)%u(2,1)  &
  &      -zickd8 *  conjg( fclinve%l( 2,2,_X_)-fclinve%l( 2,1,_X_)  &
  &                      -(fclinve%l(17,2,_X_)-fclinve%l(17,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(6)%s(itb,iz,iy,ix)%u(2,2)  &
  &      -zickd8 *       ( fclinve%l( 7,2,_X_)-fclinve%l( 7,1,_X_)  &
  &                      -(fclinve%l(19,2,_X_)-fclinve%l(19,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(6)%s(itb,iz,iy,ix)%u(2,3)  &
  &      -zickd8 *       ( fclinve%l( 8,2,_X_)-fclinve%l( 8,1,_X_)  &
  &                      -(fclinve%l(20,2,_X_)-fclinve%l(20,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(6)%s(itb,iz,iy,ix)%u(3,1)  &
  &      -zickd8 *  conjg( fclinve%l( 3,2,_X_)-fclinve%l( 3,1,_X_)  &
  &                      -(fclinve%l(18,2,_X_)-fclinve%l(18,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(6)%s(itb,iz,iy,ix)%u(3,2)  &
  &      -zickd8 *  conjg( fclinve%l( 8,2,_X_)-fclinve%l( 8,1,_X_)  &
  &                      -(fclinve%l(20,2,_X_)-fclinve%l(20,1,_X_)))
      XXe%munu(6)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(6)%s(itb,iz,iy,ix)%u(3,3)  &
  &      -zickd8 *       ( fclinve%l(12,2,_X_)-fclinve%l(12,1,_X_)  &
  &                      -(fclinve%l(21,2,_X_)-fclinve%l(21,1,_X_)))

!**** ipl = 3, (mu,nu)=(1,4)
      XXe%munu(3)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(3)%s(itb,iz,iy,ix)%u(1,1)  &
  &      -zickd8 *       ( fclinve%l( 4,2,_X_)-fclinve%l( 4,1,_X_)  &
  &                + conjg(fclinve%l( 4,2,_X_)-fclinve%l( 4,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(3)%s(itb,iz,iy,ix)%u(1,2)  &
  &      -zickd8 *       ( fclinve%l( 5,2,_X_)-fclinve%l( 5,1,_X_)  &
  &                + conjg(fclinve%l( 9,2,_X_)-fclinve%l( 9,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(3)%s(itb,iz,iy,ix)%u(1,3)  &
  &      -zickd8 *       ( fclinve%l( 6,2,_X_)-fclinve%l( 6,1,_X_)  &
  &                + conjg(fclinve%l(13,2,_X_)-fclinve%l(13,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(3)%s(itb,iz,iy,ix)%u(2,1)  &
  &      -zickd8 *  conjg( fclinve%l( 5,2,_X_)-fclinve%l( 5,1,_X_)  &
  &                + conjg(fclinve%l( 9,2,_X_)-fclinve%l( 9,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(3)%s(itb,iz,iy,ix)%u(2,2)  &
  &      -zickd8 *       ( fclinve%l(10,2,_X_)-fclinve%l(10,1,_X_)  &
  &                + conjg(fclinve%l(10,2,_X_)-fclinve%l(10,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(3)%s(itb,iz,iy,ix)%u(2,3)  &
  &      -zickd8 *       ( fclinve%l(11,2,_X_)-fclinve%l(11,1,_X_)  &
  &                + conjg(fclinve%l(14,2,_X_)-fclinve%l(14,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(3)%s(itb,iz,iy,ix)%u(3,1)  &
  &      -zickd8 *  conjg( fclinve%l( 6,2,_X_)-fclinve%l( 6,1,_X_)  &
  &                + conjg(fclinve%l(13,2,_X_)-fclinve%l(13,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(3)%s(itb,iz,iy,ix)%u(3,2)  &
  &      -zickd8 *  conjg( fclinve%l(11,2,_X_)-fclinve%l(11,1,_X_)  &
  &                + conjg(fclinve%l(14,2,_X_)-fclinve%l(14,1,_X_)))
      XXe%munu(3)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(3)%s(itb,iz,iy,ix)%u(3,3)  &
  &      -zickd8 *       ( fclinve%l(15,2,_X_)-fclinve%l(15,1,_X_)  &
  &                + conjg(fclinve%l(15,2,_X_)-fclinve%l(15,1,_X_)))

!**** ipl = 4, (mu,nu)=(2,3)
      XXe%munu(4)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(4)%s(itb,iz,iy,ix)%u(1,1)  &
  &      -zickd8 *       ( fclinve%l( 4,2,_X_)+fclinve%l( 4,1,_X_)  &
  &                + conjg(fclinve%l( 4,2,_X_)+fclinve%l( 4,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(4)%s(itb,iz,iy,ix)%u(1,2)  &
  &      -zickd8 *       ( fclinve%l( 5,2,_X_)+fclinve%l( 5,1,_X_)  &
  &                + conjg(fclinve%l( 9,2,_X_)+fclinve%l( 9,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(4)%s(itb,iz,iy,ix)%u(1,3)  &
  &      -zickd8 *       ( fclinve%l( 6,2,_X_)+fclinve%l( 6,1,_X_)  &
  &                + conjg(fclinve%l(13,2,_X_)+fclinve%l(13,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(4)%s(itb,iz,iy,ix)%u(2,1)  &
  &      -zickd8 *  conjg( fclinve%l( 5,2,_X_)+fclinve%l( 5,1,_X_)  &
  &                + conjg(fclinve%l( 9,2,_X_)+fclinve%l( 9,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(4)%s(itb,iz,iy,ix)%u(2,2)  &
  &      -zickd8 *       ( fclinve%l(10,2,_X_)+fclinve%l(10,1,_X_)  &
  &                + conjg(fclinve%l(10,2,_X_)+fclinve%l(10,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(4)%s(itb,iz,iy,ix)%u(2,3)  &
  &      -zickd8 *       ( fclinve%l(11,2,_X_)+fclinve%l(11,1,_X_)  &
  &                + conjg(fclinve%l(14,2,_X_)+fclinve%l(14,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(4)%s(itb,iz,iy,ix)%u(3,1)  &
  &      -zickd8 *  conjg( fclinve%l( 6,2,_X_)+fclinve%l( 6,1,_X_)  &
  &                + conjg(fclinve%l(13,2,_X_)+fclinve%l(13,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(4)%s(itb,iz,iy,ix)%u(3,2)  &
  &      -zickd8 *  conjg( fclinve%l(11,2,_X_)+fclinve%l(11,1,_X_)  &
  &                + conjg(fclinve%l(14,2,_X_)+fclinve%l(14,1,_X_)))
      XXe%munu(4)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(4)%s(itb,iz,iy,ix)%u(3,3)  &
  &      -zickd8 *       ( fclinve%l(15,2,_X_)+fclinve%l(15,1,_X_)  &
  &                + conjg(fclinve%l(15,2,_X_)+fclinve%l(15,1,_X_)))

!**** ipl = 2, (mu,nu)=(1,3)
      XXe%munu(2)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(2)%s(itb,iz,iy,ix)%u(1,1)  &
  &      +rrckd8 *       ( fclinve%l( 4,1,_X_)+fclinve%l( 4,2,_X_)  &
  &                - conjg(fclinve%l( 4,1,_X_)+fclinve%l( 4,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(2)%s(itb,iz,iy,ix)%u(1,2)  &
  &      +rrckd8 *       ( fclinve%l( 5,1,_X_)+fclinve%l( 5,2,_X_)  &
  &                - conjg(fclinve%l( 9,1,_X_)+fclinve%l( 9,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(2)%s(itb,iz,iy,ix)%u(1,3)  &
  &      +rrckd8 *       ( fclinve%l( 6,1,_X_)+fclinve%l( 6,2,_X_)  &
  &                - conjg(fclinve%l(13,1,_X_)+fclinve%l(13,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(2)%s(itb,iz,iy,ix)%u(2,1)  &
  &      -rrckd8 *  conjg( fclinve%l( 5,1,_X_)+fclinve%l( 5,2,_X_)  &
  &                - conjg(fclinve%l( 9,1,_X_)+fclinve%l( 9,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(2)%s(itb,iz,iy,ix)%u(2,2)  &
  &      +rrckd8 *       ( fclinve%l(10,1,_X_)+fclinve%l(10,2,_X_)  &
  &                - conjg(fclinve%l(10,1,_X_)+fclinve%l(10,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(2)%s(itb,iz,iy,ix)%u(2,3)  &
  &      +rrckd8 *       ( fclinve%l(11,1,_X_)+fclinve%l(11,2,_X_)  &
  &                - conjg(fclinve%l(14,1,_X_)+fclinve%l(14,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(2)%s(itb,iz,iy,ix)%u(3,1)  &
  &      -rrckd8 *  conjg( fclinve%l( 6,1,_X_)+fclinve%l( 6,2,_X_)  &
  &                - conjg(fclinve%l(13,1,_X_)+fclinve%l(13,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(2)%s(itb,iz,iy,ix)%u(3,2)  &
  &      -rrckd8 *  conjg( fclinve%l(11,1,_X_)+fclinve%l(11,2,_X_)  &
  &                - conjg(fclinve%l(14,1,_X_)+fclinve%l(14,2,_X_)))
      XXe%munu(2)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(2)%s(itb,iz,iy,ix)%u(3,3)  &
  &      +rrckd8 *       ( fclinve%l(15,1,_X_)+fclinve%l(15,2,_X_)  &
  &                - conjg(fclinve%l(15,1,_X_)+fclinve%l(15,2,_X_)))

!**** ipl = 5, (mu,nu)=(2,4)
      XXe%munu(5)%s(itb,iz,iy,ix)%u(1,1)=XXe%munu(5)%s(itb,iz,iy,ix)%u(1,1)  &
  &      -rrckd8 *       ( fclinve%l( 4,2,_X_)-fclinve%l( 4,1,_X_)  &
  &                - conjg(fclinve%l( 4,2,_X_)-fclinve%l( 4,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(1,2)=XXe%munu(5)%s(itb,iz,iy,ix)%u(1,2)  &
  &      -rrckd8 *       ( fclinve%l( 5,2,_X_)-fclinve%l( 5,1,_X_)  &
  &                - conjg(fclinve%l( 9,2,_X_)-fclinve%l( 9,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(1,3)=XXe%munu(5)%s(itb,iz,iy,ix)%u(1,3)  &
  &      -rrckd8 *       ( fclinve%l( 6,2,_X_)-fclinve%l( 6,1,_X_)  &
  &                - conjg(fclinve%l(13,2,_X_)-fclinve%l(13,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(2,1)=XXe%munu(5)%s(itb,iz,iy,ix)%u(2,1)  &
  &      +rrckd8 *  conjg( fclinve%l( 5,2,_X_)-fclinve%l( 5,1,_X_)  &
  &                - conjg(fclinve%l( 9,2,_X_)-fclinve%l( 9,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(2,2)=XXe%munu(5)%s(itb,iz,iy,ix)%u(2,2)  &
  &      -rrckd8 *       ( fclinve%l(10,2,_X_)-fclinve%l(10,1,_X_)  &
  &                - conjg(fclinve%l(10,2,_X_)-fclinve%l(10,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(2,3)=XXe%munu(5)%s(itb,iz,iy,ix)%u(2,3)  &
  &      -rrckd8 *       ( fclinve%l(11,2,_X_)-fclinve%l(11,1,_X_)  &
  &                - conjg(fclinve%l(14,2,_X_)-fclinve%l(14,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(3,1)=XXe%munu(5)%s(itb,iz,iy,ix)%u(3,1)  &
  &      +rrckd8 *  conjg( fclinve%l( 6,2,_X_)-fclinve%l( 6,1,_X_)  &
  &                - conjg(fclinve%l(13,2,_X_)-fclinve%l(13,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(3,2)=XXe%munu(5)%s(itb,iz,iy,ix)%u(3,2)  &
  &      +rrckd8 *  conjg( fclinve%l(11,2,_X_)-fclinve%l(11,1,_X_)  &
  &                - conjg(fclinve%l(14,2,_X_)-fclinve%l(14,1,_X_)))
      XXe%munu(5)%s(itb,iz,iy,ix)%u(3,3)=XXe%munu(5)%s(itb,iz,iy,ix)%u(3,3)  &
  &      -rrckd8 *       ( fclinve%l(15,2,_X_)-fclinve%l(15,1,_X_)  &
  &                - conjg(fclinve%l(15,2,_X_)-fclinve%l(15,1,_X_)))


    enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz

  return
end subroutine
#undef _X_

subroutine force_hmc_clover(XX,zickd8,fx,fz)
!
! Calc force from inverse clover term in even/odd preconditioned Wilson-Dirac(Clover) hopping operator
!                         
!  XX(n) =  XX(n) + zickd8 * tr[ sigma_{mu,nu} fz(n) fx(n)^{+}]
!                         
! where,   sigma_{mu,nu} : Dirac rep.
!
  implicit none
  type(tfield_gluon_wg), intent(inout) :: XX   ! pre-force contribution for clvoer term (\dot{T})
  complex(DP),           intent(in) :: zickd8  ! force coefficient
  type(field_quark_wg),  intent(in) ::  fx     ! pseudo fermion
  type(field_quark_wg),  intent(in) ::  fz     ! pseudo fermion

  call force_hmc_clover_eo(XX%eo(0),zickd8,fx%eo(0),fz%eo(0))
  call force_hmc_clover_eo(XX%eo(1),zickd8,fx%eo(1),fz%eo(1))

  return
end subroutine

subroutine force_hmc_clover_eo(XXe,zickd8,fxe,fze)
!
! Calc force from inverse clover term  in even/odd preconditioned Wilson-Dirac(clover) hopping operator
!
!  XX(n) =  XX(n) + zickd8 * tr[ sigma_{mu,nu} fz(n) fx(n)^{+}]
!
! where,   sigma_{mu,nu} : Dirac rep.
!
  implicit none
  type(tfield_gluon_eo_wg), intent(inout) :: XXe ! even/odd site force contribution for clvoer term (\dot{T})
  complex(DP),              intent(in) :: zickd8 ! force coefficient
  type(field_quark_eo_wg),  intent(in) ::  fxe   ! even/odd site pseudo fermion
  type(field_quark_eo_wg),  intent(in) ::  fze   ! even/odd site pseudo fermion

  integer :: ix,iy,iz,itb,ieoxyz,ic,jc

  if (.not. ((XXe%munu(1)%ieo == fxe%ieo) .and. &
 &           (XXe%munu(1)%ieo == fze%ieo)) ) then
    call error_stop("error in force_hmc_clover_eo(even/odd)")
  endif

!*********************************************************************
!*  XX(n) =  XX(n) + c * tr[ sigma_{mu,nu} (fze,fzo)(n) fx(n)^{+}]
!*********************************************************************
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,ic,jc) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+fxe%ieo,2)

!********************************************************
!* ipl=1
!*          zsg1=-z1;zsg2= z1;zsg3=-z1;zsg4= z1
!*          isg1=1; isg2=2; isg3=3; isg4=4
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(1)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(1)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &               +zickd8*( -       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         -       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
!********************************************************
!* ipl=2
!*          zsg1=-zi;zsg2= zi;zsg3=-zi;zsg4= zi
!*          isg1=2; isg2=1; isg3=4; isg4=3
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(2)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(2)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &             +zi*zickd8*(-       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         -       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
!********************************************************
!* ipl=3
!*          zsg1=-z1;zsg2=-z1;zsg3=-z1;zsg4=-z1
!*          isg1=4; isg2=3; isg3=2; isg4=1
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(3)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(3)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &               -zickd8*( +       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
!********************************************************
!* ipl=4
!*          zsg1=-z1;zsg2=-z1;zsg3=-z1;zsg4=-z1
!*          isg1=2; isg2=1; isg3=4; isg4=3
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(4)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(4)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &               -zickd8*( +       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
!********************************************************
!* ipl=5
!*          zsg1= zi;zsg2=-zi;zsg3= zi;zsg4=-zi
!*          isg1=4; isg2=3; isg3=2; isg4=1
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(5)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(5)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &             +zi*zickd8*(+       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         -       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         -       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
!********************************************************
!* ipl=6
!*          zsg1=-z1;zsg2= z1;zsg3=-z1;zsg4= z1
!*          isg1=3; isg2=4; isg3=1; isg4=2
!********************************************************
      do itb=1-ieoxyz,NTH-ieoxyz
        do jc=1,COL
        do ic=1,COL
        XXe%munu(6)%s(itb,iz,iy,ix)%u(ic,jc)=XXe%munu(6)%s(itb,iz,iy,ix)%u(ic,jc)  &
 &               +zickd8*( -       fze%s(itb,iz,iy,ix)%y(ic, 3)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 1))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 4)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 2))    &
 &                         -       fze%s(itb,iz,iy,ix)%y(ic, 1)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 3))    &
 &                         +       fze%s(itb,iz,iy,ix)%y(ic, 2)     &
 &                          *conjg(fxe%s(itb,iz,iy,ix)%y(jc, 4)) )
        enddo
        enddo
      enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz

  return
end subroutine

subroutine make_clover_leaf(u,wl34,wl98)
!
! Calculate clover leaf (field strength) for Clover fermion
!
! clover field strength is class/module variable and is calculated using the two-link fields.
!
  implicit none
  type(vfield_gluon_wg), intent(in) :: u         ! gauge link
  type(tfield_gluon_wg), intent(in) :: wl34,wl98 ! 2-links prepared by make_two_links

  integer :: iflag

  iflag = 0
  if (.not.associated(m_ucl)) then
    allocate(m_ucl)
  endif
  call make_clover_leaf_eo(m_ucl%eo(0),u%eo(0),u%eo(1),wl34%eo(0),wl34%eo(1),wl98%eo(0),wl98%eo(1))
  call make_clover_leaf_eo(m_ucl%eo(1),u%eo(1),u%eo(0),wl34%eo(1),wl34%eo(0),wl98%eo(1),wl98%eo(0))
  return
end subroutine

subroutine delete_clover_leaf()
!
! delete clover leaf term.
!
  implicit none
  if (associated(m_ucl)) then
    deallocate(m_ucl)
  endif
  return
end subroutine

subroutine make_clover_leaf_eo(ucle,ue,uo,wl34e,wl34o,wl98e,wl98o)
!
! Calculate clover leaf (field strength) field on even/odd sites.
!
!   ucle = F_{mu,nu}
!
  implicit none
  type(field_clover_leaf_eo), intent(inout) :: ucle  !  even/odd site clover leaf
  type(vfield_gluon_eo_wg),   intent(in)    :: ue    ! even/odd site gauge link
  type(vfield_gluon_eo_wg),   intent(in)    :: uo    ! odd/even site gauge link
  type(tfield_gluon_eo_wg),   intent(in)    :: wl34e,wl34o   ! 2-links field even/odd,odd/even
  type(tfield_gluon_eo_wg),   intent(in)    :: wl98e,wl98o   ! 2-links field even/odd,odd/even

  type(sfield_gluon_eo_wg), allocatable :: w1,w2,w3,v
  integer :: ix,iy,iz,ieoxyz,itb,isite
  integer :: mu,nu,ipl,ic,jc
  complex(DP) :: zid8

  if (.not. ((ue%mu(1)%ieo == ue%mu(2)%ieo) .and.  &
 &           (ue%mu(1)%ieo == ue%mu(3)%ieo) .and.  &
 &           (ue%mu(1)%ieo == ue%mu(4)%ieo) .and.  &
 &           (uo%mu(1)%ieo == uo%mu(2)%ieo) .and.  &
 &           (uo%mu(1)%ieo == uo%mu(3)%ieo) .and.  &
 &           (uo%mu(1)%ieo == uo%mu(4)%ieo) .and.  &
 &           (ue%mu(1)%ieo /= uo%mu(1)%ieo)) ) then
    call error_stop("error in make_clover_term_eo(even/odd)")
  endif

  allocate(w1,w2,w3,v)

  ucle%ieo=ue%mu(1)%ieo

  zid8=-zi/8.0_DP
!************************************************************************
!*  super loop over hyperplanes (extends to end of subroutine)
!*  mu < nu
!************************************************************************
  do mu=1,NDIM-1
  do nu=mu+1,NDIM
    ipl = imunu(mu,nu)

    call new(v,ucle%ieo)
    call clear(v)

!====================================================
!
!  4---<---3
!  |       |
!  v       ^  = u(n)_mu*wl34(n)_{mu,nu}*u(n)^+_nu
!  |       |
!  n--->---2
!
!====================================================
    call assign_mult(w1,ue%mu(mu),wl34e%munu(ipl))
    call assign_conjg2_mult(v,w1,ue%mu(nu))


!====================================================
!
!  8---<---n
!  |       |
!  v       ^  = u(n-mu)^+_mu*(wl98(n)_{mu,nu})^+*u(n-nu)_nu
!  |       |
!  9--->---5
!
!====================================================
    call assign_shift(w2,uo%mu(mu),-mu)
    call assign_conjg(w3,wl98e%munu(ipl))
    call assign_conjg1_mult(w1,w2,w3)
    call assign_shift(w3,uo%mu(nu),-nu)
    call assign_mult(w2,w1,w3)
    call accum_add(v,w2)

!====================================================
!
!  7---<---4
!  |       |
!  v       ^  = wl34(n-mu)_{mu,nu}*wl98(n+nu)^+_{mu,nu}
!  |       |
!  8--->---n
!
!====================================================
    call assign_shift(w2,wl34o%munu(ipl),-mu)
    call assign_shift(w3,wl98o%munu(ipl),+nu)
    call assign_conjg2_mult(w1,w2,w3)
    call accum_add(v,w1)

!====================================================
!
!  n---<---2
!  |       |
!  v       ^  = wl98(n+mu)^+_{mu,nu}*wl34(n-nu)_{mu,nu}
!  |       |
!  5--->---6
!
!====================================================
    call assign_shift(w2,wl98o%munu(ipl),+mu)
    call assign_shift(w3,wl34o%munu(ipl),-nu)
    call assign_conjg1_mult(w1,w2,w3)
    call accum_add(v,w1)

!************************************************************************
!*  total clover (with counter clockwise plaquettes)
!*  ucl(ipl) = -zi ( v - v^{+} ) / 8
!************************************************************************
!OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,isite,ic,jc) _CLPS_
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz = mod(ipeo+iz+iy+ix+ucle%ieo,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        isite=(itb+ieoxyz)+(iz-1)*NTH+(iy-1)*NTH*NZ+(ix-1)*NTH*NZ*NY
        do jc=1,COL
        do ic=1,COL
          ucle%t(ic,jc,isite,ipl)=(      v%s(itb,iz,iy,ix)%u(ic,jc)  &
  &                               -conjg(v%s(itb,iz,iy,ix)%u(jc,ic)))*zid8
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo ! end of do ix,iy,iz

  enddo ! end of do nu
  enddo ! end of do mu

  deallocate(v,w1,w2,w3)

#ifdef _SF
  call set_sf_boundary(ucle)
#endif

  return
end subroutine

subroutine hamil_clover_trlog(this,action)
!
! Calculate Hamiltonian of log determinant clover term.
! This also computes inverse clover term by side effect.
!
!   Action :
!
!     SF = - Nf log[det[F]]
!
  implicit none
  type(quark_clover), intent(inout) :: this    !  quark parameters, and contains 
                                               !  inverse clover term (in chiral rep., linear form)
  real(DP),           intent(out)   :: action  ! action value
  real(DP) :: logdetfcl

!**************************************************
!* calc. inverse clover term and log[det[Fcl]]
!**************************************************
  call make_inverse_clover_term(this,1,logdetfcl)

  action=-logdetfcl*get_nflavor(this%quark_wilson)

  return
end subroutine

subroutine make_inverse_clover_term(this,idet,logdetfcl)
!
! Calculate inverse clover term matrix and log-determinant of clover term.
!
!   fcl    = [ 1 - csw kappa/2 sigma_{mu,nu} F_{mu,nu}]
!   fclinv =  fcl^-1
!   log[det[fcl]]
!
! This uses module variable, m_ucl : clover leaf field.
!
!   [ 1 - csw kappa/2 sigma_{mu,nu} F_{mu,nu}]^-1  = diag(fclinv(1),fclinv(2)),
!
!   where
!
!     |   1   2   3   4   5   6  |
!     |       7   8   9  10  11  |
!     |          12  13  14  15  |
!     |              16  17  18  | * 0.5
!     |                  19  20  |
!     |                      21  |
! 
!   are stored in fclinv(i).
!
!  Note: 0.5 is multiplied to simplify conversion from chiral to Dirac rep.
!
  use comlib
  implicit none
  type(quark_clover), intent(inout) :: this      ! quark parameter contains 
                                                 ! inverse clover term (in chiral rep. linear storage)
 
  integer,            intent(in)    :: idet      ! if idet == 1 then logdetfcl is computed.
  real(DP),           intent(out)   :: logdetfcl ! log determinat of clover term.

  type(f_c_m_eo), allocatable :: f(:),finv(:)
  real(DP) :: logdetf1cl(0:1)
  real(DP) :: logdetf2cl(0:1)
  integer :: ieo,iout
  real(DP) :: kappa,csw

   iout = get_file_unit(this%clover_term_log)
  kappa = get_kappa(this%quark_wilson)
    csw = this%csw

  logdetfcl=0.0_DP

  if (.not.associated(m_ucl)) then
    call error_stop("Clover leaf field is not created, Stop!")
  endif
  if (.not.associated(this%fclinv)) then
    allocate(this%fclinv)
  endif
  call new(this%fclinv)

  allocate(f(2),finv(2))

  do ieo=0,1

!*****************************
!* calc clover term matrix in chiral rep.
!*
!*  fcl = diag[f1cl,f2cl] = 1 - csw kappa/2 sigma_{mu,nu}F_{mu,nu}
!*
!* solve inverse matrix (even part)
!*
!*****************************
    call clover_f1f2(kappa,csw,f(1),f(2),m_ucl%eo(ieo))
#ifdef _SF
    call set_sf_boundary_clover_f1f2_eo(kappa,this%cttilde,f(1),f(2))
#endif
    call _CLV_SOLVER(f(1)%m,iout,idet,finv(1)%m,logdetf1cl(ieo))
    call _CLV_SOLVER(f(2)%m,iout,idet,finv(2)%m,logdetf2cl(ieo))
    finv(1)%ieo=f(1)%ieo
    finv(2)%ieo=f(2)%ieo

!*****************************
!* chiral full matrix -> chiral linear vector
!*****************************
    call conv_matrix_to_linear_eo(finv(1),finv(2),this%fclinv%eo(ieo))

  enddo

  if (idet == 1) then
    logdetfcl=logdetf1cl(0)+logdetf2cl(0)  &
 &           +logdetf1cl(1)+logdetf2cl(1)
#ifndef _singlePU
    call comlib_sumcast(logdetfcl)
#endif
  endif

  deallocate(f,finv)

  return
end subroutine

subroutine delete_inverse_clover_term(this)
!
! delete inverse clover term.
!
  implicit none
  type(quark_clover), intent(inout) :: this
  if (associated(this%fclinv)) then
    deallocate(this%fclinv)
  endif
  return
end subroutine

subroutine assign_mult_clover_hopping_eo(this,FMye,yo,u)
!
! Multiply clover term preconditioned Hopping operator (odd->even or even->odd hopping only)
!
!   FMye <= Fee \ Meo yo
! 
  use counter_class
  implicit none
  type(quark_clover),      intent(in)    :: this  ! clover quark parameter and inverse clover term
  type(field_quark_eo_wg), intent(out)   :: FMye  ! even/odd site fermion field
  type(field_quark_eo_wg), intent(inout) ::   yo  ! odd/even site fermion field
  type(vfield_gluon_wg),   intent(in)    ::    u  ! gauge field

  call copy_boundary(yo)
  call mult_clv_hopping_tzyx_eo(FMye,yo,u,this%fclinv)
  call inc(mult_iter)

!  call assign_mult_hop(this%quark_wilson,FMye,yo,u)
!  call accum_mult_clover_term(FMye,this)

  return
end subroutine

subroutine assign_mult_clover_term_all(yd,this,y)
!
! Multiply inverse clover term matrix on fermion field
!
!    yd <= F \ y
!
  implicit none
  type(field_quark_wg), intent(inout) :: yd
  type(quark_clover),   intent(in)    :: this
  type(field_quark_wg), intent(in)    ::  y
  call assign_mult_clover_term(yd%eo(0),this,y%eo(0))
  call assign_mult_clover_term(yd%eo(1),this,y%eo(1))
  return
end subroutine

subroutine assign_mult_clover_term_eo(yde,this,ye)
!
! Multiply inverse clover term matrix on fermion field (even/odd sites only)
!
!    yde <= Fee \ ye
!
  implicit none
  type(field_quark_eo_wg), intent(inout) :: yde  ! even/odd site fermion field
  type(quark_clover),      intent(in)    :: this ! clover quark parameters
  type(field_quark_eo_wg), intent(in)    ::  ye  ! even/odd site fermion field

  complex(DP):: a1,a2,a3,a4,a5,a6
  complex(DP):: b1,b2,b3,b4,b5,b6
  complex(DP):: p1,p2,p3,p4,p5,p6
  complex(DP):: m1,m2,m3,m4,m5,m6
  complex(DP):: y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc
  integer :: ix,iy,iz,itb,ieoxyz,ieo,itb0

  ieo     = ye%ieo
  yde%ieo = ye%ieo

!***********************************************************
!* Calc:
!*        
!*  yde = F^-1ee ye
!*      
!***********************************************************
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,itb,itb0, &
!$OMP         a1,a2,a3,a4,a5,a6,b1,b2,b3,b4,b5,b6, &
!$OMP         p1,p2,p3,p4,p5,p6,m1,m2,m3,m4,m5,m6, &
!$OMP         y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+yde%ieo,2)
#ifdef _SF
!===============================
! Set SF boundary condition
! itt = 0,1 and NTT+1 is zero
!===============================
    yde%s(0,iz,iy,ix)%y(:,:) = Z0
    do itb=1,NTH-ieoxyz
#else
    do itb=1-ieoxyz,NTH-ieoxyz
#endif
      itb0 = itb+ieoxyz

#define _YE_    ye%s(itb,iz,iy,ix)
#define _YDE_  yde%s(itb,iz,iy,ix)
#define _FCLV_ this%fclinv
#include "mult_clover_term_eo_kernel.h90"

    enddo

#ifdef _SF
    if (ieoxyz==1) then
      yde%s(NTH,iz,iy,ix)%y(:,:) = Z0
    endif
#endif

  enddo
  enddo
  enddo ! end of do ix,iy,iz

  return
end subroutine

subroutine accum_mult_clover_term_all(y,this)
!
! Multiply inverse clover term matrix on fermion field (overwritten)
!
!    y <= F \ y
!
  implicit none
  type(field_quark_wg), intent(inout) :: y    ! fermion field (overwritten)
  type(quark_clover),   intent(in)    :: this ! clover quark paramters
  call accum_mult_clover_term(y%eo(0),this)
  call accum_mult_clover_term(y%eo(1),this)
  return
end subroutine

subroutine accum_mult_clover_term_eo(ye,this)
!
! Multiply inverse clover term matrix on fermion field (overwritten) (even/odd sites only)
!
!    ye <= Fee \ ye
!
  implicit none
  type(field_quark_eo_wg), intent(inout) :: ye   ! even/odd site fermion field (overwritten)
  type(quark_clover),      intent(in)    :: this ! clover quark paramters

  complex(DP):: a1,a2,a3,a4,a5,a6
  complex(DP):: b1,b2,b3,b4,b5,b6
  complex(DP):: p1,p2,p3,p4,p5,p6
  complex(DP):: m1,m2,m3,m4,m5,m6
  complex(DP):: y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc
  integer :: ix,iy,iz,itb,ieoxyz,itb0
  integer :: ieo

  ieo = ye%ieo

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,itb,itb0, &
!$OMP         y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc, &
!$OMP         p1,p2,p3,p4,p5,p6,m1,m2,m3,m4,m5,m6, &
!$OMP         a1,a2,a3,a4,a5,a6,b1,b2,b3,b4,b5,b6) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
#ifdef _SF
!===============================
! Set SF boundary condition
! itt = 0,1 and NTT+1 is zero
!===============================
    ye%s(0,iz,iy,ix)%y(:,:) = Z0
    do itb=1,NTH-ieoxyz
#else
    do itb=1-ieoxyz,NTH-ieoxyz
#endif
      itb0 = itb+ieoxyz

#define _YE_   ye%s(itb,iz,iy,ix)
#define _YDE_  ye%s(itb,iz,iy,ix)
#define _FCLV_ this%fclinv
#include "mult_clover_term_eo_kernel.h90"

    enddo

#ifdef _SF
    if (ieoxyz==1) then
      ye%s(NTH,iz,iy,ix)%y(:,:) = Z0
    endif
#endif

  enddo
  enddo
  enddo ! end of do ix,iy,iz

  return
end subroutine

subroutine assign_mult_eoprec_cd_eo(this,yde,ye,u)
!
! Multiply even/odd-site and clover term preconditioned Wilson-Clover-Dirac operator(even/odd sites only):
!
!               v
!        yde <= Dee ye,
!
!  where  v
!         Dee = 1ee - kappa^2 Fee \ Meo Foo \ Moe
!                                          
  implicit none
  type(quark_clover),      intent(in)    :: this ! clover quark parameters and inverse clover term
  type(field_quark_eo_wg), intent(inout) :: yde  ! even/odd site fermion field
  type(field_quark_eo_wg), intent(inout) ::  ye  ! even/odd site fermion field
  type(vfield_gluon_wg),   intent(in)    :: u    ! gauge field

  type(field_quark_eo_wg), allocatable :: Myo
  real(DP) :: kappa2,kappa

  allocate(Myo)
  kappa = get_kappa(this%quark_wilson)
  kappa2=-(kappa**2)

!***********************************
!*  Myo = F^-1oo Moe ye
!*  yde = F^-1ee Meo Myo = F^-1ee Meo F^-1oo Moe ye
!***********************************
  call assign_mult_hop(this,Myo, ye,u)
  call assign_mult_hop(this,yde,Myo,u)

!***********************************
!*  yde = ye - kappa^2 yde
!*      = ye - kappa^2 Meo Moe ye
!***********************************
  call accum_mult_add(yde,kappa2,ye)

  deallocate(Myo)

  return
end subroutine

subroutine assign_mult_eoprec_cq_eo_dag(phys_param,dye,ye,u)
!
!= Multiply Hermitian conjugate of the even/odd-site preconditioned wilson-dirac operator
!
!        v
!  dye = Dee^+ ye
!
!  v
!  Dee^+ = 1ee - gamma_5 Meo Foo \ Moe Fee \ gamma_5
!
  implicit none
  type(quark_clover), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: dye
  type(field_quark_eo_wg), intent(in)    ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable :: wo
  type(quark_wilson) :: wilson_dummy
  real(DP) :: kappa,kappa2

  kappa = get_kappa(phys_param)
  call set_kappa(wilson_dummy,kappa)

  !=============================
  ! dye = Fee \ gamma_5 ye
  !=============================
  call assign_mult_gamma5(dye,ye)
  call accum_mult_clover_term(dye,phys_param)
  
  !=============================
  ! wo = Foo \ Moe dye
  !=============================
  allocate(wo)
  call assign_mult_hop(  phys_param,wo,dye,u)

  !=============================
  ! dye = gamma_5 Meo wo
  !=============================
  call assign_mult_hop(wilson_dummy,dye,wo,u)
  call accum_mult_gamma5(dye)
  deallocate(wo)

  !=============================
  ! dye = -kappa^2 dye + ye
  !=============================
  kappa2 = -kappa**2
  call accum_mult_add(dye,kappa2,ye)

  return
end subroutine

subroutine mult_clv_hopping_tzyx_eo(yde,yo,u,fclinv)
!
! Multiply clover-term preconditioned hopping matrix (odd->even or even->odd hopping only)
!
  use field_gauge_class
  implicit none
  type(field_quark_eo_wg), intent(inout):: yde  ! even/odd site fermion vector (output)
  type(field_quark_eo_wg), intent(in)   ::  yo  ! odd/even site fermion vector (input)
  type(vfield_gluon_wg),   intent(in)   :: u    ! gauge field
  type(field_clover_term), intent(in)   :: fclinv  ! inverse clover term in chiral-linear form.

  type(su3fv_spinor) :: yt
  complex(DP):: w(COL,SPIN/2),y(COL,SPIN/2)
  complex(DP):: a1,a2,a3,a4,a5,a6
  complex(DP):: b1,b2,b3,b4,b5,b6
  complex(DP):: p1,p2,p3,p4,p5,p6
  complex(DP):: m1,m2,m3,m4,m5,m6
  complex(DP):: y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc
  integer :: ix,iy,iz,itb,itbup,itbdn,ieoxyz,ieo,ioe,itb0

  ioe = yo%ieo
  ieo = 1-ioe
  yde%ieo = ieo

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,itb,itb0,itbup,itbdn,yt,w,y, &
!$OMP         a1,a2,a3,a4,a5,a6,b1,b2,b3,b4,b5,b6, &
!$OMP         p1,p2,p3,p4,p5,p6,m1,m2,m3,m4,m5,m6, &
!$OMP         y1,y2,y3,y4,y5,y6,y7,y8,y9,ya,yb,yc) _CLPS_
  do ix=1,NX
  do iy=1,NY         
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+yde%ieo,2)
    itbup=   +ieoxyz 
    itbdn=-(1-ieoxyz)

#ifdef _SF
!===============================
! Set SF boundary condition
! itt = 0,1 and NTT+1 is zero
!===============================
    yde%s(0,iz,iy,ix)%y(:,:) = Z0
    do itb=1,NTH-ieoxyz
#else
    do itb=1-ieoxyz,NTH-ieoxyz
#endif

#include "../FieldFermionClass/mult_hop_kernel.h90"

      itb0 = itb+ieoxyz
#define _YE_    yt
#define _YDE_   yde%s(itb,iz,iy,ix)
#define _FCLV_  fclinv
#include "mult_clover_term_eo_kernel.h90"

    enddo

#ifdef _SF
    if (ieoxyz==1) then
      yde%s(NTH,iz,iy,ix)%y(:,:) = Z0
    endif
#endif

  enddo
  enddo
  enddo

  return
end subroutine

subroutine clover_f1f2(kappa,csw,f1cle,f2cle,ucle)
!
! Calculate clover term matrix in chiral rep form clover-leaf field on even/odd sites.
!
! f1cl = 1 - csw*kappa*[(-F23+F14)sig1 + (+F13+F24)sig2 + (-F12+F34)sig3]
! f2cl = 1 - csw*kappa*[(-F23-F14)sig1 + (+F13-F24)sig2 + (-F12-F34)sig3]
!
! F12 = ucl(1), F13 = ucl(2), F14 = ucl(3)
! F23 = ucl(4), F24 = ucl(5), F34 = ucl(6)  : clover leaf field tensor
!
! sig1, sig2, sig3 : Pauli matrix
!
  implicit none
  real(DP),                   intent(in)    :: kappa ! hopping parameter
  real(DP),                   intent(in)    :: csw   ! C_sw
  type(f_c_m_eo),             intent(inout) :: f1cle ! upper-2-spinor component of clover term in chiral rep.
  type(f_c_m_eo),             intent(inout) :: f2cle ! lower-2-spinor component of clover term in chiral rep.
  type(field_clover_leaf_eo), intent(in)    :: ucle  ! clover leaf field tensor

  complex(DP) :: zunit(COL,COL)
  real(DP) :: cswkappa
  integer :: ixyzt,ic,jc

  f1cle%ieo=ucle%ieo
  f2cle%ieo=ucle%ieo

  cswkappa=csw*kappa

  do jc=1,COL
  do ic=1,COL
    zunit(ic,jc) = z0
  enddo
  enddo
  do ic=1,COL
    zunit(ic,ic) = z1
  enddo

!*************************************************************************
!*
!* f1cl=1-csw*kappa*[(-F23+F14)sig1+(+F13+F24)sig2+(-F12+F34)sig3]
!* f2cl=1-csw*kappa*[(-F23-F14)sig1+(+F13-F24)sig2+(-F12-F34)sig3]
!*
!*  F12 = ucl(1),F13 = ucl(2),F14 = ucl(3)
!*  F23 = ucl(4),F24 = ucl(5),F34 = ucl(6)
!*
!*************************************************************************
!$OMP PARALLEL DO PRIVATE(ixyzt,ic,jc)
  do ixyzt=1,(NTH*NZ*NY*NX)
    do jc=1,COL
    do ic=1,COL
      f1cle%m(ic    ,jc    ,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,1)  &
 &                               +ucle%t(ic,jc,ixyzt,6))+zunit(ic,jc)
      f1cle%m(ic+COL,jc+COL,ixyzt)=-cswkappa*  &
 &                              (+ucle%t(ic,jc,ixyzt,1)  &
 &                               -ucle%t(ic,jc,ixyzt,6))+zunit(ic,jc)
    enddo
    enddo

    do jc=1,COL
    do ic=1,COL
      f1cle%m(ic    ,jc+COL,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,4)  &
 &                               +ucle%t(ic,jc,ixyzt,3)  &
 &                            -zi*ucle%t(ic,jc,ixyzt,2)  &
 &                            -zi*ucle%t(ic,jc,ixyzt,5))
      f1cle%m(ic+COL,jc    ,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,4)  &
 &                               +ucle%t(ic,jc,ixyzt,3)  &
 &                            +zi*ucle%t(ic,jc,ixyzt,2)  &
 &                            +zi*ucle%t(ic,jc,ixyzt,5))
    enddo
    enddo

    do jc=1,COL
    do ic=1,COL
      f2cle%m(ic    ,jc    ,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,1)  &
 &                               -ucle%t(ic,jc,ixyzt,6))+zunit(ic,jc)
      f2cle%m(ic+COL,jc+COL,ixyzt)=-cswkappa*  &
 &                              (+ucle%t(ic,jc,ixyzt,1)  &
 &                               +ucle%t(ic,jc,ixyzt,6))+zunit(ic,jc)
    enddo
    enddo

    do jc=1,COL
    do ic=1,COL
      f2cle%m(ic    ,jc+COL,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,4)  &
 &                               -ucle%t(ic,jc,ixyzt,3)  &
 &                            -zi*ucle%t(ic,jc,ixyzt,2)  &
 &                            +zi*ucle%t(ic,jc,ixyzt,5))
      f2cle%m(ic+COL,jc    ,ixyzt)=-cswkappa*  &
 &                              (-ucle%t(ic,jc,ixyzt,4)  &
 &                               -ucle%t(ic,jc,ixyzt,3)  &
 &                            +zi*ucle%t(ic,jc,ixyzt,2)  &
 &                            -zi*ucle%t(ic,jc,ixyzt,5))
    enddo
    enddo
  enddo ! end of do ixyzt

  return                                                            
end subroutine

subroutine clvinv_ldl(Ae,iout,idet,Be,logdetAe)
!
! Calcluate the inverse of 6x6 clover term matrix using LDL decomposition,
!
!  Solve:
!            Ae Be = 1 -> Be = Ae^-1
!
  implicit none
  integer, parameter :: NSITE=NTH*NZ*NY*NX
  complex(DP), intent(in)  :: Ae(CLSP/2,CLSP/2,NSITE)  ! even/odd site clover term matrix (input) (col*spin/2 x col*spin/2)
  integer,     intent(in)  :: iout                     ! logfile unit number
  integer,     intent(in)  :: idet                     ! if idet == 1 then logdetAe is computed.
  complex(DP), intent(out) :: Be(CLSP/2,CLSP/2,NSITE)  ! even/odd site inverse clover term matrix (output)
  real(DP),    intent(out) :: logdetAe                 ! log[det[Ae]] : log determinant of Ae 

  complex(DP) :: zunit(CLSP/2,CLSP/2)
  complex(DP), allocatable ::   AAe(:,:,:)
  complex(DP), allocatable :: ctmp1(:,:,:)
  complex(DP), allocatable :: ctmp2(:,:,:)
  complex(DP), allocatable :: ctmp3(:,:,:)
  real(DP),    allocatable :: d(:)
  integer :: icount
  integer :: ixyzt,ics,jcs

  allocate(  AAe(CLSP/2,CLSP/2,NSITE))
  allocate(ctmp1(CLSP/2,CLSP/2,NSITE))
  allocate(ctmp2(CLSP/2,CLSP/2,NSITE))
  allocate(ctmp3(CLSP/2,CLSP/2,NSITE))
  allocate(d(NSITE))

  do jcs=1,CLSP/2
  do ics=1,CLSP/2
    zunit(ics,jcs) = z0
  enddo
  enddo
  do ics=1,CLSP/2
    zunit(ics,ics) = z1
  enddo

!*********************************
!* set unit matrix for source b
!*********************************
!$OMP PARALLEL DO PRIVATE(ixyzt,ics,jcs)
  do ixyzt=1,NSITE
    do jcs=1,CLSP/2
    do ics=1,CLSP/2
       Be(ics,jcs,ixyzt) = z0
      AAe(ics,jcs,ixyzt) = Ae(ics,jcs,ixyzt)
    enddo
    enddo
  enddo
!$OMP PARALLEL DO PRIVATE(ixyzt,ics)
  do ixyzt=1,NSITE
    do ics=1,CLSP/2
       Be(ics,ics,ixyzt) = z1
    enddo
  enddo

  call ldldcmp(AAe)

!*************************************
!* calculate det[AAe] when idet = 1.
!*************************************
  if (idet == 1) then
!$OMP PARALLEL DO PRIVATE(ixyzt)
    do ixyzt=1,NSITE
      d(ixyzt)=real(AAe(1,1,ixyzt))  &
 &            *real(AAe(2,2,ixyzt))  &
 &            *real(AAe(3,3,ixyzt))  &
 &            *real(AAe(4,4,ixyzt))  &
 &            *real(AAe(5,5,ixyzt))  &
 &            *real(AAe(6,6,ixyzt))
    enddo
    logdetAe=0.0_DP
!$OMP PARALLEL DO PRIVATE(ixyzt) REDUCTION(+:logdetAe)
    do ixyzt=1,NSITE
      logdetAe=logdetAe+log(d(ixyzt))
    enddo
  else
    logdetAe=0.0_DP
  endif

!*************************************
!* calculate b = A^-1
!*************************************
  call ldlbksb(AAe,Be)

!*************************************
!* check  b * A = A^-1 * A = I ?
!*************************************
!$OMP PARALLEL DO PRIVATE(ixyzt,ics,jcs)
  do ixyzt=1,NSITE
    do jcs=1,CLSP/2
    do ics=1,CLSP/2
      ctmp1(ics,jcs,ixyzt)=  &
 &                  Ae(ics,1,ixyzt)*Be(1,jcs,ixyzt)  &
 &                 +Ae(ics,2,ixyzt)*Be(2,jcs,ixyzt)  &
 &                 +Ae(ics,3,ixyzt)*Be(3,jcs,ixyzt)  &
 &                 +Ae(ics,4,ixyzt)*Be(4,jcs,ixyzt)  &
 &                 +Ae(ics,5,ixyzt)*Be(5,jcs,ixyzt)  &
 &                 +Ae(ics,6,ixyzt)*Be(6,jcs,ixyzt)
    enddo
    enddo
  enddo
!***************
!*  ctmp =  A^-1*A - I
!***************
!$OMP PARALLEL DO PRIVATE(ixyzt,ics,jcs)
  do ixyzt=1,NSITE
    do jcs=1,CLSP/2
    do ics=1,CLSP/2
      ctmp1(ics,jcs,ixyzt)=ctmp1(ics,jcs,ixyzt)  &
 &                        -zunit(ics,jcs)
    enddo
    enddo
  enddo

!$OMP PARALLEL DO PRIVATE(ixyzt)
  do ixyzt=1,NSITE
    d(ixyzt)=0.0_DP
  enddo

!$OMP PARALLEL DO PRIVATE(ixyzt)
  do ixyzt=1,NSITE
    d(ixyzt)=real(ctmp1(1,1,ixyzt))**2+aimag(ctmp1(1,1,ixyzt))**2  &
 &          +real(ctmp1(2,1,ixyzt))**2+aimag(ctmp1(2,1,ixyzt))**2  &
 &          +real(ctmp1(3,1,ixyzt))**2+aimag(ctmp1(3,1,ixyzt))**2  &
 &          +real(ctmp1(4,1,ixyzt))**2+aimag(ctmp1(4,1,ixyzt))**2  &
 &          +real(ctmp1(5,1,ixyzt))**2+aimag(ctmp1(5,1,ixyzt))**2  &
 &          +real(ctmp1(6,1,ixyzt))**2+aimag(ctmp1(6,1,ixyzt))**2  &
 &          +real(ctmp1(1,2,ixyzt))**2+aimag(ctmp1(1,2,ixyzt))**2  &
 &          +real(ctmp1(2,2,ixyzt))**2+aimag(ctmp1(2,2,ixyzt))**2  &
 &          +real(ctmp1(3,2,ixyzt))**2+aimag(ctmp1(3,2,ixyzt))**2  &
 &          +real(ctmp1(4,2,ixyzt))**2+aimag(ctmp1(4,2,ixyzt))**2  &
 &          +real(ctmp1(5,2,ixyzt))**2+aimag(ctmp1(5,2,ixyzt))**2  &
 &          +real(ctmp1(6,2,ixyzt))**2+aimag(ctmp1(6,2,ixyzt))**2  &
 &          +real(ctmp1(1,3,ixyzt))**2+aimag(ctmp1(1,3,ixyzt))**2  &
 &          +real(ctmp1(2,3,ixyzt))**2+aimag(ctmp1(2,3,ixyzt))**2  &
 &          +real(ctmp1(3,3,ixyzt))**2+aimag(ctmp1(3,3,ixyzt))**2  &
 &          +real(ctmp1(4,3,ixyzt))**2+aimag(ctmp1(4,3,ixyzt))**2  &
 &          +real(ctmp1(5,3,ixyzt))**2+aimag(ctmp1(5,3,ixyzt))**2  &
 &          +real(ctmp1(6,3,ixyzt))**2+aimag(ctmp1(6,3,ixyzt))**2  &
 &          +real(ctmp1(1,4,ixyzt))**2+aimag(ctmp1(1,4,ixyzt))**2  &
 &          +real(ctmp1(2,4,ixyzt))**2+aimag(ctmp1(2,4,ixyzt))**2  &
 &          +real(ctmp1(3,4,ixyzt))**2+aimag(ctmp1(3,4,ixyzt))**2  &
 &          +real(ctmp1(4,4,ixyzt))**2+aimag(ctmp1(4,4,ixyzt))**2  &
 &          +real(ctmp1(5,4,ixyzt))**2+aimag(ctmp1(5,4,ixyzt))**2  &
 &          +real(ctmp1(6,4,ixyzt))**2+aimag(ctmp1(6,4,ixyzt))**2  &
 &          +real(ctmp1(1,5,ixyzt))**2+aimag(ctmp1(1,5,ixyzt))**2  &
 &          +real(ctmp1(2,5,ixyzt))**2+aimag(ctmp1(2,5,ixyzt))**2  &
 &          +real(ctmp1(3,5,ixyzt))**2+aimag(ctmp1(3,5,ixyzt))**2  &
 &          +real(ctmp1(4,5,ixyzt))**2+aimag(ctmp1(4,5,ixyzt))**2  &
 &          +real(ctmp1(5,5,ixyzt))**2+aimag(ctmp1(5,5,ixyzt))**2  &
 &          +real(ctmp1(6,5,ixyzt))**2+aimag(ctmp1(6,5,ixyzt))**2  &
 &          +real(ctmp1(1,6,ixyzt))**2+aimag(ctmp1(1,6,ixyzt))**2  &
 &          +real(ctmp1(2,6,ixyzt))**2+aimag(ctmp1(2,6,ixyzt))**2  &
 &          +real(ctmp1(3,6,ixyzt))**2+aimag(ctmp1(3,6,ixyzt))**2  &
 &          +real(ctmp1(4,6,ixyzt))**2+aimag(ctmp1(4,6,ixyzt))**2  &
 &          +real(ctmp1(5,6,ixyzt))**2+aimag(ctmp1(5,6,ixyzt))**2  &
 &          +real(ctmp1(6,6,ixyzt))**2+aimag(ctmp1(6,6,ixyzt))**2
  enddo 

  icount=0
  do ixyzt=1,NSITE
    if (d(ixyzt) > (1.0e-15_DP*(CLSP/2))**2) icount=icount+1
  enddo

  if (icount /= 0) then
    write(iout,'(" LU decomp. inversion : ",I3)')nodeid
    write(iout,'("   # of site, (residual > 1.0d-15) :",I4)')icount
  endif

  deallocate(  AAe)
  deallocate(ctmp1)
  deallocate(ctmp2)
  deallocate(ctmp3)
  deallocate(d)

  return
end subroutine

subroutine ldlbksb(a,b)
!
! Solve A X = B with LDL^+ decomposed matrix A using forward(backward) substitution.
!
  implicit none
  integer, parameter :: NSITE=NTH*NZ*NY*NX
  complex(DP), intent(in)    :: a(CLSP/2,CLSP/2,NSITE)  ! LDL^+ decomposed matrix
  complex(DP), intent(inout) :: b(CLSP/2,CLSP/2,NSITE)  ! source matrix for input, solution matrix for output
!*****************
  complex(DP), allocatable :: sum(:,:)
  integer :: ics,jcs,kcs,ixyzt

  allocate(sum(CLSP/2,NSITE))

  do ics=2,CLSP/2
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
    do ixyzt=1,NSITE
      do jcs=1,CLSP/2
        sum(jcs,ixyzt)=b(ics,jcs,ixyzt)
      enddo
    enddo
    do kcs=1,ics-1
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
      do ixyzt=1,NSITE
        do jcs=1,CLSP/2
          sum(jcs,ixyzt) = sum(jcs,ixyzt)  &
 &                     - a(ics,kcs,ixyzt)  &
 &                      *b(kcs,jcs,ixyzt)
        enddo
      enddo
    enddo
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
    do ixyzt=1,NSITE
      do jcs=1,CLSP/2
        b(ics,jcs,ixyzt)=sum(jcs,ixyzt)
      enddo
    enddo
  enddo

!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
  do ixyzt=1,NSITE
    do jcs=1,CLSP/2
      b(CLSP/2,jcs,ixyzt) =   &
 &         b(CLSP/2,jcs,ixyzt)/real(a(CLSP/2,CLSP/2,ixyzt))
    enddo
  enddo

  do ics=CLSP/2-1,1,-1
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
    do ixyzt=1,NSITE
      do jcs=1,CLSP/2
        sum(jcs,ixyzt) =  &
 &         b(ics,jcs,ixyzt)/real(a(ics,ics,ixyzt))
      enddo
    enddo
    do kcs=CLSP/2,ics+1,-1
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
      do ixyzt=1,NSITE
        do jcs=1,CLSP/2
          sum(jcs,ixyzt) = sum(jcs,ixyzt)   &
 &              -  conjg(a(kcs,ics,ixyzt))  &
 &                      *b(kcs,jcs,ixyzt)
        enddo
      enddo
    enddo
!$OMP PARALLEL DO PRIVATE(ixyzt,jcs)
    do ixyzt=1,NSITE
      do jcs=1,CLSP/2
        b(ics,jcs,ixyzt) = sum(jcs,ixyzt)
      enddo
    enddo
  enddo

  deallocate(sum)

  return
end subroutine

subroutine ldldcmp(a)
!
! Do LDL^+ decomposition
!
  implicit none
  integer, parameter :: NSITE=NTH*NZ*NY*NX
  complex(DP), intent(inout) :: a(CLSP/2,CLSP/2,NSITE)  ! input matrix to be LDL^+ decomposed (overwritten)

  complex(DP), allocatable ::  sum(:)
  real(DP),    allocatable :: rsum(:)
  integer :: ixyzt,ics,jcs,kcs

  allocate( sum(NSITE))
  allocate(rsum(NSITE))

  do ics=1,CLSP/2
!$OMP PARALLEL DO PRIVATE(ixyzt)
    do ixyzt=1,NSITE
      rsum(ixyzt) = real(a(ics,ics,ixyzt))
    enddo
    do kcs=1,ics-1
!$OMP PARALLEL DO PRIVATE(ixyzt)
      do ixyzt=1,NSITE
        rsum(ixyzt) = rsum(ixyzt)  &
 &             -(  real(a(ics,kcs,ixyzt))**2   &
 &               +aimag(a(ics,kcs,ixyzt))**2)  &
 &                *real(a(kcs,kcs,ixyzt))
      enddo
    enddo
!$OMP PARALLEL DO PRIVATE(ixyzt)
    do ixyzt=1,NSITE
      a(ics,ics,ixyzt) = cmplx(rsum(ixyzt),0.0_DP,kind=DP)
    enddo

    do jcs=ics+1,CLSP/2
!$OMP PARALLEL DO PRIVATE(ixyzt)
      do ixyzt=1,NSITE
        sum(ixyzt) = a(ics,jcs,ixyzt)
      enddo
      do kcs=1,ics-1
!$OMP PARALLEL DO PRIVATE(ixyzt)
        do ixyzt=1,NSITE
          sum(ixyzt) = sum(ixyzt) - a(ics,kcs,ixyzt)   &
 &                           *conjg(a(jcs,kcs,ixyzt))  &
 &                            *real(a(kcs,kcs,ixyzt))
        enddo
      enddo
!$OMP PARALLEL DO PRIVATE(ixyzt)
      do ixyzt=1,NSITE
        sum(ixyzt)=  &
 &        conjg(sum(ixyzt))/real(a(ics,ics,ixyzt))
      enddo
!$OMP PARALLEL DO PRIVATE(ixyzt)
      do ixyzt=1,NSITE
        a(jcs,ics,ixyzt)=sum(ixyzt)
      enddo
    enddo
  enddo

  deallocate( sum)
  deallocate(rsum)

  return
end subroutine

subroutine conv_matrix_to_linear_eo(f1cl,f2cl,fcl)
!
! Transform clover term matrix from chiral-rep full matrix form to chiral-rep linear vector form.
!
!
!  fcl = [1-csw/2 kappa sigma_{mu,nu} F_{mu,nu}] (uppar and lower in chiral rep.)
!
  implicit none
  type(field_clover_term_eo), intent(inout) ::  fcl  ! even/odd site inverse clover term matrix in chiral rep. (linear)
  type(f_c_m_eo),             intent(in)    :: f1cl  ! even/odd site inverse clover term matrix in chiral rep. (upper)
  type(f_c_m_eo),             intent(in)    :: f2cl  ! even/odd site inverse clover term matrix in chiral rep. (lower)

  integer :: ix,iy,iz,itb,itb0,ieoxyz,isite

  if (f1cl%ieo /= f2cl%ieo) then
    call error_stop("error in conv_matrix_to_linear_eo(even/odd)")
  endif
  fcl%ieo=f1cl%ieo

!*********************************************
!* transform  full form -> linear form (hermitian matrix)
!*
!*         |  1   2   3   4   5   6 |
!*         |      7   8   9  10  11 |
!*  f1cl = |         12  13  14  15 |*0.5_DP, i = fcl%l(i,....,1)
!*         |             16  17  18 |
!*         |                 19  20 |
!*         |                     21 |
!*
!*         |  1   2   3   4   5   6 |
!*         |      7   8   9  10  11 |
!*  f2cl = |         12  13  14  15 |*0.5_DP, i = fcl%l(i,....,2)
!*         |             16  17  18 |
!*         |                 19  20 |
!*         |                     21 |
!*
!*********************************************
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,itb,itb0,isite) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+fcl%ieo,2)
    do itb=1-ieoxyz,NTH-ieoxyz
      itb0 = itb+ieoxyz
      isite = itb0+(iz-1)*NTH+(iy-1)*NTH*NZ+(ix-1)*NTH*NZ*NY
      fcl%l( 1,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,1,isite)
      fcl%l( 2,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,2,isite)
      fcl%l( 3,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,3,isite)
      fcl%l( 4,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,4,isite)
      fcl%l( 5,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,5,isite)
      fcl%l( 6,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(1,6,isite)
      fcl%l( 7,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(2,2,isite)
      fcl%l( 8,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(2,3,isite)
      fcl%l( 9,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(2,4,isite)
      fcl%l(10,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(2,5,isite)
      fcl%l(11,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(2,6,isite)
      fcl%l(12,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(3,3,isite)
      fcl%l(13,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(3,4,isite)
      fcl%l(14,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(3,5,isite)
      fcl%l(15,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(3,6,isite)
      fcl%l(16,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(4,4,isite)
      fcl%l(17,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(4,5,isite)
      fcl%l(18,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(4,6,isite)
      fcl%l(19,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(5,5,isite)
      fcl%l(20,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(5,6,isite)
      fcl%l(21,1,itb0,iz,iy,ix)= 0.5_DP*f1cl%m(6,6,isite)

      fcl%l( 1,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,1,isite)
      fcl%l( 2,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,2,isite)
      fcl%l( 3,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,3,isite)
      fcl%l( 4,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,4,isite)
      fcl%l( 5,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,5,isite)
      fcl%l( 6,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(1,6,isite)
      fcl%l( 7,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(2,2,isite)
      fcl%l( 8,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(2,3,isite)
      fcl%l( 9,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(2,4,isite)
      fcl%l(10,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(2,5,isite)
      fcl%l(11,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(2,6,isite)
      fcl%l(12,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(3,3,isite)
      fcl%l(13,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(3,4,isite)
      fcl%l(14,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(3,5,isite)
      fcl%l(15,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(3,6,isite)
      fcl%l(16,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(4,4,isite)
      fcl%l(17,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(4,5,isite)
      fcl%l(18,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(4,6,isite)
      fcl%l(19,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(5,5,isite)
      fcl%l(20,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(5,6,isite)
      fcl%l(21,2,itb0,iz,iy,ix)= 0.5_DP*f2cl%m(6,6,isite)
    enddo
  enddo
  enddo
  enddo ! end of do ix,iy,iz

  return
end subroutine


subroutine assign_mult_DdagD_ee_cq(phys_param,qye,ye,u)
!
!= Multiply Hermitian squared the even/odd-site preconditioned wilson-dirac
!  operator
!
!        v     v        v
!  qye = Dee^+ Dee ye = Qee ye
!
  type(quark_clover), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: qye
  type(field_quark_eo_wg), intent(inout) ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable   ::  we
  allocate(we)
  call assign_mult_eoprec_wd(    phys_param, we,ye,u)
  call assign_mult_eoprec_wd_dag(phys_param,qye,we,u)
  deallocate(we)
  return
end subroutine

subroutine assign_mult_DDdag_ee_cq(phys_param,qye,ye,u)
!
!= Multiply Hermitian squared the even/odd-site preconditioned wilson-dirac
!  operator
!
!        v   v          v
!  qye = Dee Dee^+ ye = Oee ye
!
  type(quark_clover), intent(in) :: phys_param
  type(field_quark_eo_wg), intent(inout) :: qye
  type(field_quark_eo_wg), intent(inout) ::  ye
  type(vfield_gluon_wg),   intent(in)    ::  u
  type(field_quark_eo_wg), allocatable   ::  we
  allocate(we)
  call assign_mult_eoprec_wd_dag(phys_param, we,ye,u)
  call assign_mult_eoprec_wd(    phys_param,qye,we,u)
  deallocate(we)
  return
end subroutine

#ifdef _SF
subroutine set_sf_boundary_clover_f1f2_eo(kappa,cttilde,f1cle,f2cle)
!
! Set SF boundary term on the clover term (chiral rep.)
!
! Input:
!
! (f1,f2)(n) = 1 - csw*kappa sum_{mu>nu} sigma_{mu,nu} F_{mu,nu}(n)
!
! Output:
! (f1,f2)(n) = 1 - csw*kappa sum_{mu>nu} sigma_{mu,nu} F_{mu,nu}(n) + 2*(cttilde-1)*kappa*B(n4)
!
! with B(n4) = 1*(delta_{n4,1} + delta_{n4,T-1})
!
!    kappa : hopping paramter
!  cttilde : Wilson-Clover SF boundary improvement coefficient
!    f1cle : (even/odd site) clover term upper-component in chiral rep.
!    f2cle : (even/odd site) clover term lower-component in chiral rep.
!
  real(DP), intent(in) :: kappa
  real(DP), intent(in) :: cttilde
  type(f_c_m_eo), intent(inout) :: f1cle 
  type(f_c_m_eo), intent(inout) :: f2cle 
  integer :: ieo,ith,ieoxyz,iz,iy,ix,ics
  integer :: isite
  real(DP) :: kct

  if (.not. (f1cle%ieo == f2cle%ieo)  ) then
    call error_stop("error in set_sf_boundary_clover_f1f2_eo(even/odd)")
  endif
  ieo = f1cle%ieo

  kct = 2*kappa*(cttilde-1.0_DP)

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,isite,ics) _CLPS_
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz = mod(ipeo+iz+iy+ix+ieo,2)

    if ( 0 == ieoxyz ) then
      !-----------------------------------
      ! itt = 2 (t=1) boundary set
      ! ith = 1-ieoxyz,...,NTH-ieoxyz
      ! itt = 2*ith + ieoxyz
      ! itt = 2 => ith = 1, ieoxyz = 0
      !-----------------------------------
      ith = 1
      isite=(ith+ieoxyz)+(iz-1)*NTH+(iy-1)*NTH*NZ+(ix-1)*NTH*NZ*NY

      !-----------------------
      ! spin color diagonal
      !-----------------------
      do ics=1,COL*SPIN/2 
        f1cle%m(ics,ics,isite) = f1cle%m(ics,ics,isite) + kct
        f2cle%m(ics,ics,isite) = f2cle%m(ics,ics,isite) + kct
      enddo

      !-----------------------------------
      ! itt = NTT (t=T-1) boundary set
      ! ith = 1-ieoxyz,...,NTH-ieoxyz
      ! itt = 2*ith + ieoxyz
      ! itt = NTT => ith = NTH, ieoxyz = 0
      !-----------------------------------

      ith = NTH
      isite=(ith+ieoxyz)+(iz-1)*NTH+(iy-1)*NTH*NZ+(ix-1)*NTH*NZ*NY

      !-----------------------
      ! spin color diagonal
      !-----------------------
      do ics=1,COL*SPIN/2
        f1cle%m(ics,ics,isite) = f1cle%m(ics,ics,isite) + kct
        f2cle%m(ics,ics,isite) = f2cle%m(ics,ics,isite) + kct
      enddo

    endif

  enddo
  enddo
  enddo

  return
end subroutine
#endif

#endif
! end of ifdef _CLOVER

end module
