#include "../config.h"
  
!#define _TEST_DUMPY_
#undef _TEST_DUMPY_

module reweighting_class
  use comlib
  use constants_module
  use random_class
  use error_class
  use counter_class
  use timer_class
  use lattice_class
  use logfile_class
  use print_status_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use field_clover_class
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use config_profile_class
  use quark_prop_class, only : assign_inv_mult_wd
  implicit none
  private

  public :: reweighting_obj
  public :: new,delete
  public :: read,save,print,print_statistics,run

  integer, parameter :: MAX_IR_SOLVER_ITER=10

  type reweighting_obj
    type(vfield_gluon_wg), allocatable :: u
    type(config_profile_obj)   :: config_prof
#ifdef _CLOVER
    type(quark_clover) :: target_quark
    type(quark_clover) :: config_quark
#else
    type(quark_wilson) :: target_quark
    type(quark_wilson) :: config_quark
#endif
    real(DP)  :: tol=1.0d-14
    type(logfile), pointer :: reweight_log => NULL()
    type(logfile), pointer :: solver_log   => NULL()
    type(timer)   :: tot_time, io_time, run_time
    type(counter) :: solver_iter

    real(DP)  :: stout_alpha = 0.0_DP
    integer   :: stout_nstep = 0

    integer :: nf=2
    integer :: nsqrt=1
    integer :: run=0
    integer :: traj_start=0
    integer :: traj_end=0
    integer :: traj_skip=0
    integer :: traj=0
    integer :: random_number_seed=1
    integer :: nnoise=1
    integer :: nbreak=1

    character(len=CHARLEN) :: input_fname=""
    character(len=CHARLEN) :: cwd=""
    character(len=CHARLEN) :: job_id=""
    character(len=CHARLEN) :: prog_version=""
    character(len=CHARLEN) :: ctime=""
    character(len=CHARLEN) :: cdate=""
    character(len=CHARLEN) :: commentline(4)=["","","",""]
    character(len=CHARLEN) :: reweight_file="reweight_hist" ! reweight filename
    character(len=CHARLEN) :: solver_file="solver_log"      ! reweight filename
    character(len=CHARLEN) :: config_path=""                ! configuration file path
    character(len=CHARLEN) :: config_file=""                ! configuration file name
    character(len=CHARLEN) :: seed_file="ReweightRandSeed"  ! seed filename
    character(len=CHARLEN) :: seed_path=""                  ! seed path
  end type

  type reweighting_factor
    real(DP) :: rho,dS,dSA,dSB,dSC
  end type

  interface new
    module procedure new_reweighting
  end interface

  interface delete
    module procedure delete_reweighting
  end interface

  interface read
    module procedure read_reweighting
  end interface

  interface print
    module procedure print_reweighting
  end interface

  interface print_statistics
    module procedure print_stat_reweighting
  end interface

  interface save
    module procedure save_reweighting
  end interface

  interface run
    module procedure run_reweighting
  end interface

contains

subroutine new_reweighting(this)
  use hmc_identity
  implicit none
  type(reweighting_obj), intent(inout) :: this
  type(lattice_world)   :: lattice

  call new(lattice)

  call new(this%tot_time)
  call tic(this%tot_time)

  call new(this%io_time)
  call new(this%run_time)

  !=====================================
  ! Initialize Status
  !=====================================
  this%input_fname=repeat(' ',len(this%input_fname))
  this%cwd        =repeat(' ',len(this%cwd))
  this%job_id     =repeat(' ',len(this%job_id))
  this%prog_version =repeat(' ',len(this%prog_version))

  call DATE_AND_TIME(date=this%cdate,time=this%ctime)
  write(this%job_id,'(A,1X,A)')TRIM(ADJUSTL(this%cdate)),TRIM(ADJUSTL(this%ctime))
  
  this%prog_version = TRIM(hmc_vers)

  if (nodeid == 0) then
    call GET_COMMAND_ARGUMENT(1,this%input_fname)
    call getcwd(this%cwd)
  endif

  if (nodeid == 0) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
    write(*,'(A)')TRIM(this%prog_version)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%input_fname,0)
    call comlib_bcast(this%cwd,0)
    call comlib_bcast(this%job_id,0)
    call comlib_bcast(this%ctime,0)
    call comlib_bcast(this%cdate,0)
  endif

  allocate(this%u)
  call new(this%u)

  this%reweight_log => NULL()
  allocate(this%reweight_log)

  this%solver_log => NULL()
  allocate(this%solver_log)

  call new(this%config_prof)

  call new(this%solver_iter)

  return
end subroutine

subroutine delete_reweighting(this)
  implicit none
  type(reweighting_obj), intent(inout) :: this
  type(lattice_world)                  :: lattice

  call delete(this%config_prof)

  call delete(this%reweight_log)
  deallocate(this%reweight_log)
  this%reweight_log => NULL()

  call delete(this%solver_log)
  deallocate(this%solver_log)
  this%solver_log => NULL()

  deallocate(this%u)

  call delete(lattice)

  return
end subroutine

subroutine read_reweighting(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(reweighting_obj), intent(inout) :: this
  integer :: iout
  character(len=CHARLEN) :: char,fname,str
  real(DP) :: target_kappa,target_csw
  real(DP) :: config_kappa,config_csw

  call tic(this%io_time)

  iout = search_free_file_unit()
!=====================================
! Read Input Parameters
!=====================================
  if (nodeid==0) then
    open(iout,file=this%input_fname,status='unknown',form='formatted')
#ifdef _CLOVER
    read(iout,*)target_kappa,target_csw
    read(iout,*)config_kappa,config_csw
#else
    read(iout,*)target_kappa
    read(iout,*)config_kappa
#endif
    read(iout,*)this%nf
    read(iout,*)this%tol

    read(iout,'(A)')this%config_path
    read(iout,'(A)')this%config_file

    read(iout,'(A)')this%seed_path
    read(iout,'(A)')this%seed_file

    read(iout,*)this%stout_alpha,this%stout_nstep

    read(iout,*)this%run
    read(iout,*)this%traj_start,this%traj_end,this%traj_skip
    read(iout,*)this%random_number_seed
    read(iout,*)this%nnoise
    read(iout,*)this%nbreak
    close(iout)
  endif

#ifndef _singlePU
  call comlib_bcast(target_kappa,0)
  call comlib_bcast(config_kappa,0)
#ifdef _CLOVER
  call comlib_bcast(target_csw,0)
  call comlib_bcast(config_csw,0)
#endif
#endif

  call set_kappa(this%config_quark,config_kappa)
  call set_kappa(this%target_quark,target_kappa)
#ifdef _CLOVER
  call set_csw(this%config_quark,config_csw)
  call set_csw(this%target_quark,target_csw)
#endif

#ifndef _STOUT
  this%stout_alpha = 0.0_DP
  this%stout_nstep = 0
#endif

#ifndef _singlePU
  call comlib_bcast(this%stout_alpha,0)
  call comlib_bcast(this%stout_nstep,0)
  call comlib_bcast(this%nf,0)
  call comlib_bcast(this%tol,0)
  call comlib_bcast(this%config_path,0)
  call comlib_bcast(this%config_file,0)
  call comlib_bcast(this%seed_path,0)
  call comlib_bcast(this%seed_file,0)
  call comlib_bcast(this%run,0)
  call comlib_bcast(this%traj_start,0)
  call comlib_bcast(this%traj_end,0)
  call comlib_bcast(this%traj_skip,0)
  call comlib_bcast(this%random_number_seed,0)
  call comlib_bcast(this%nnoise,0)
  call comlib_bcast(this%nbreak,0)
#endif

!=====================================
! Check Input Parameters
!=====================================
  if (this%traj_end < this%traj_start) then
    write(str,'(" ERROR: traj_start > traj_end !")')
    call error_stop(str)
  endif
  if (this%run < 0) then
    write(str,'(" ERROR: run < 0 !")')
    call error_stop(str)
  endif
  if (this%nnoise < 1) then
    write(str,'(" ERROR: nnoise < 1 !")')
    call error_stop(str)
  endif
  if (this%nbreak < 1) then
    write(str,'(" ERROR: nbreak < 1 !")')
    call error_stop(str)
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
  endif
  write(char,'(1X,A," : read OK ",I3)')TRIM(this%input_fname),nodeid
  call print_status(char)

!=====================================
! Initialize Random Number generator
!=====================================
  fname=fname_cont(this%run,this%seed_path,this%seed_file)
  select case (this%run)
  case (0)
    call new(g_rand,0,this%random_number_seed,NPU,nodeid)
    if (nodeid == 0) then
      write(*,'(80("-"))')
      write(*,'(1X,"Initialize Random Number")')
    endif
  case (1:)
    call new(g_rand,1,this%random_number_seed,NPU,nodeid)
    open(iout,file=fname,status='unknown',form='unformatted',action='read')
    call read(g_rand,iout)
    close(iout)
    if (nodeid == 0) then
      write(*,'(80("-"))')
    endif
    write(char,'(1X,A," : read OK ",I3)')TRIM(fname),nodeid
    call print_status(char)
  end select

!=====================================
! Open Log Files
!=====================================
!  char=TRIM(fname_plaq(this%run+1,this%hmc_status%current_directory,TRIM(this%reweight_file)))
  write(char,'(A,I6.6,"-",I6.6)') &
  TRIM(TRIM(this%cwd)//"/"//TRIM(this%reweight_file))//".",this%traj_start,this%traj_end
  call new(this%reweight_log,char)

!  char=TRIM(fname_plaq(this%run+1,this%hmc_status%current_directory,TRIM(this%solver_file)))
  write(char,'(A,I6.6,"-",I6.6)') &
  TRIM(TRIM(this%cwd)//"/"//TRIM(this%solver_file))//".",this%traj_start,this%traj_end
  call new(this%solver_log,char)

  call toc(this%io_time)

  return
end subroutine

subroutine print_reweighting(this)
  implicit none
  type(reweighting_obj), intent(inout) :: this
  character(len=CHARLEN) :: cahr

  call tic(this%io_time)
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("       Quark Parameters :")')
#ifdef _CLOVER
    write(*,'(" Target Quark Kappa,Csw :",2F24.15)')get_kappa(this%target_quark),get_csw(this%target_quark)
    write(*,'(" Config Quark Kappa,Csw :",2F24.15)')get_kappa(this%config_quark),get_csw(this%config_quark)
#else
    write(*,'(" Target Quark Kappa     :",1F24.15)')get_kappa(this%target_quark)
    write(*,'(" Config Quark Kappa     :",1F24.15)')get_kappa(this%config_quark)
#endif
    write(*,'("                     Nf :",I3)')this%nf
    write(*,'(80("-"))')
    write(*,'("      Solver Parameters :")')
    write(*,'("              Tolelance :",E24.15)')this%tol
    write(*,'(80("-"))')
    write(*,'("            Config Path :",A)')TRIM(this%config_path)
    write(*,'("            Config File :",A)')TRIM(this%config_file)
    write(*,'("              Seed Path :",A)')TRIM(this%seed_path)
    write(*,'("              Seed File :",A)')TRIM(this%seed_file)
    write(*,'(80("-"))')
    write(*,'("         Job Parameters :")')
    write(*,'("              Rand Seed :",I12)')this%random_number_seed
    write(*,'("                  Run # :",I4)')this%run
    write(*,'(" Traj. Start, End, Skip :",2I6,I4)')this%traj_start,this%traj_end,this%traj_skip
    write(*,'(80("-"))')
    write(*,'(" Reweighting Parameters :")')
    write(*,'("                Noise # :",I4)')this%nnoise
    write(*,'("              Breakup # :",I4)')this%nbreak
    write(*,'(80("="))')
  endif
  call toc(this%io_time)

  return
end subroutine

subroutine run_reweighting(this)
#ifdef _STOUT
  use hmc_stout_class
#endif
  implicit none
  type(reweighting_obj), intent(inout) :: this
#ifdef _STOUT
  type(stout_smearing_parameters) :: stout
#endif
  integer :: itraj,inoise

#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: dummy
#endif

  type(reweighting_factor) :: rfactor
  real(DP) :: plq0
  real(DP) :: dS,rho,rtmp,dSA,dSB,dSC
  complex(DP) :: ctmp
  character(len=CHARLEN) :: str

  character(len=CHARLEN) :: fbuff
  character(len=CHARLEN) :: config_path2

  this%run=this%run+1

  rho = get_kappa(this%config_quark)/get_kappa(this%target_quark)

  do itraj=this%traj_start,this%traj_end,this%traj_skip

    ! reset order of Taylor expansion for Nf=1
    this%nsqrt = 1

!=====================================
! Read Measurment Configuration
!=====================================
    call read_config(itraj,this%config_path,this%config_file,this%config_prof,this%u)

!--- Ukita ------
!    fbuff=REPEAT(' ',LEN(fbuff))
!    write(fbuff,'(I6.6)') itraj
!    config_path2=TRIM(ADJUSTL(this%config_path))//'/'//TRIM(ADJUSTL(fbuff))

!    call read(itraj,this%config_path,this%config_file,this%config_prof,this%u)

#ifdef _STOUT
!-------------------
!  Stout smearing
!-------------------
    if (this%stout_alpha > 0.0_DP .and. this%stout_nstep > 0 ) then
      call new(stout,alpha=this%stout_alpha,nstout=this%stout_nstep,is_smear_only=.TRUE.)
      call print(stout)
      call smear(stout,this%u)
      call assign(this%u,stout%usmr(1))
      call delete(stout)
      call copy_boundary(this%u)
      call plaquette(this%u,plq0)
      if (nodeid==0) write(*,'(" PLQ(STOUT):",F24.16)')plq0
    endif
#endif

#ifdef _CLOVER
!===============================
! compute inverse clover term
!===============================
    allocate(wl34,wl98)
    call make_two_links(this%u,wl34,wl98)
    call make_clover_leaf(this%u,wl34,wl98)
    deallocate(wl34,wl98)
    call make_inverse_clover_term(this%target_quark,0,dummy)
    call delete_clover_leaf()
#endif

    do inoise=1,this%nnoise

      call tic(this%run_time)

      select case(this%nf)
      case(2)
        call calc_reweighting_factor_nf2(this,rfactor)
      case(1)
        call calc_reweighting_factor_nf1(this,rfactor)
      case default
        write(str,'(" Error: invalid Nf",I3,". Stop!")')this%nf
        call error_stop(str)
      end select

      if (nodeid==0) write(*,'("%DSP",4ES24.15)') &
  &         rfactor%dSA,rfactor%dSB,rfactor%dSC,rfactor%dSA+rfactor%dSB+rfactor%dSC

      if (nodeid==0) then
        write(str,'(I6,I3,3F24.15,2ES24.15)')  &
  &     itraj,inoise,get_kappa(this%target_quark),get_kappa(this%config_quark),rho,rfactor%dS,exp(-rfactor%dS)
        call print(this%reweight_log,str)
      endif

      call toc(this%run_time)

    enddo
  enddo

#ifdef _CLOVER
  call delete_inverse_clover_term(this%target_quark)
#endif

  return
end subroutine

subroutine calc_reweighting_factor_nf2(this,rfactor)
!=======================================================================
!= Calculate reweighting factor for Nf=2
!=  dS = |W \ eta|^2 - |eta|^2
!=     W^(-1) = rho + (1-rho) D'^(-1)
!=     rho = kappa_config/kappa_target
!=======================================================================
#ifdef _TEST_DUMPY_
  use read_dumpy_class
#endif
  implicit none
  type(reweighting_obj),    intent(inout) :: this
  type(reweighting_factor), intent(out) :: rfactor
  integer :: iter
  type(field_quark_wg), allocatable :: eta,w,v
  real(DP) :: dS,dSA,dSB,dSC,rho
  real(DP) :: rtmp,tol
  integer :: iout,iiter
  complex(DP) :: ctmp
  character(len=CHARLEN) :: str
#ifdef _TEST_DUMPY_
  character(len=CHARLEN) :: fname="./catdump.bin"
#endif

  write(str,'("#SOLVER_LOG Nf=2: ",A," :",I9)')__FILE__,__LINE__
  call print(this%solver_log,str)
  iout = get_file_unit(this%solver_log)

  allocate(eta,w,v)
  call new(eta)
  call new(w)
  call new(v)

  rho = get_kappa(this%config_quark)/get_kappa(this%target_quark)
  rfactor%rho = rho

!===================================================
!  dS = | rho eta + (1-rho) D' \ eta |^2 - |eta|^2
!     = | D/D' eta|^2 - |eta|^2
!===================================================

  ! Gaussian noise : eta
  call set_gaussian_noise(eta)
#ifdef _TEST_DUMPY_
  call read_dumpy(fname,eta)
#endif

!============================================================================
! uses rho formula
!============================================================================

  tol = this%tol
  iiter = 0

  !----------------
  ! w = D' \ eta
  !----------------
  call assign_inv_mult_wd(this%target_quark,iout,tol,iiter,w,eta,this%u)
  call inc(this%solver_iter,iiter)

  call assign_mult(v,eta,rho)    ! v = eta * rho
  rtmp = 1.0_DP-rho
  call accum_add_mult(v,w,rtmp)  ! v = v + (1-rho) w

!--- Ukita ------
!  rfactor%dS = abs2(v) - abs2(eta)
  rfactor%dS = qf_dd_wg_abs2_sub(v,eta)
!----------------

!================================
! rho expansion
!================================
  rfactor%dSA = (1.0_DP-rho)**2 * abs2(w)                      ! gauge dependent term (O(1-rho)^2)
  ctmp = prod(w,eta)
  rfactor%dSB = 2*rho*(1.0_DP-rho) * real(ctmp,kind=KIND(rho)) ! gauge dependent term   (O(1-rho))
  rfactor%dSC = (rho**2-1.0_DP) * abs2(eta)                    ! gauge independent term (O(1-rho))

  deallocate(eta,w,v)

  return
end subroutine


subroutine calc_reweighting_factor_nf1(this,rfactor)
!=======================================================================
!= Calculate reweighting factor for Nf=1
!=  dS = |W^(-1/2) eta|^2 - |eta|^2
!=     W^(-1) = rho + (1-rho) D'^(-1)
!=     rho = kappa_config/kappa_target
!=======================================================================
#ifdef _TEST_DUMPY_
  use read_dumpy_class
#endif
  implicit none
  type(reweighting_obj),    intent(inout) :: this
  type(reweighting_factor), intent(out) :: rfactor
  integer :: isqrt,nsqrt,iter,iout
  type(field_quark_wg), allocatable :: eta,w,v,s,t
  real(DP) :: dS,dSA,dSB,dSC,rho
  real(DP) :: rtmp,coef,tol
  complex(DP) :: ctmp
  character(len=CHARLEN) :: str
#ifdef _TEST_DUMPY_
  character(len=CHARLEN) :: fname="./catdump.bin"
#endif

  write(str,'("#SOLVER_LOG Nf=1: ",A," :",I9)')__FILE__,__LINE__
  call print(this%solver_log,str)
  iout = get_file_unit(this%solver_log)

  allocate(eta,w)
  call new(eta)
  call new(w)

  rho = get_kappa(this%config_quark)/get_kappa(this%target_quark)
  rfactor%rho = rho

!===================================================
!  dS = | (D/D')^(1/2) eta|^2 - |eta|^2
!===================================================

  ! Gaussian noise : eta
  call set_gaussian_noise(eta)
#ifdef _TEST_DUMPY_
  call read_dumpy(fname,eta)
#endif

  nsqrt = this%nsqrt


  tol = this%tol

100 continue   ! nsqrt loop

!====================================
! w = W^(-1/2) eta
!====================================
    call assign_mult_sqrt_corr(nsqrt,iout,tol,this%solver_iter,rho,this%target_quark,w,eta,this%u)

    allocate(v,s,t)
    call new(v)
    call new(s)
    call new(t)

!====================================
! s = W^(-1/2) w \sim W^(-1) eta
!====================================
    call assign_mult_sqrt_corr(nsqrt,iout,tol,this%solver_iter,rho,this%target_quark,s,w,this%u)

!====================================
! v = W^(-1) eta
!====================================
    !---------------
    ! t = D' \ eta
    !---------------
    iter = 0
    call assign_inv_mult_wd(this%target_quark,iout,tol,iter,t,eta,this%u)
    call inc(this%solver_iter,iter)

    call assign_mult(v,eta,rho)     ! v = rho eta
    rtmp = 1.0_DP-rho
    call accum_add_mult(v,t,rtmp)   ! v = v + (1-rho) t

    !
    ! residual check
    ! s = s - v
    !
    call accum_sub(s,v)
    rtmp = sqrt(abs2(s)/abs2(v))

    !
    ! dS
    !
!--- Ukita ------
!    rfactor%dS  = abs2(w) - abs2(eta)
    rfactor%dS  = qf_dd_wg_abs2_sub(w,eta)
!----------------
    rfactor%dS  = abs2(w) - abs2(eta)
    rfactor%dSA = 0.0_DP
    rfactor%dSB = 0.0_DP
    rfactor%dSC = 0.0_DP

    if (nodeid==0) then
      write(*,'("%SQRT:",I4,2E24.15)')nsqrt,rtmp,rfactor%dS
    endif
    deallocate(v,s,t)

  if (rtmp > this%tol) then
    nsqrt = nsqrt + 1
    goto 100
  else
    if (nodeid==0) then
      write(*,'("%SQRT converged")')
    endif
  endif

  this%nsqrt = nsqrt

  deallocate(eta,w)

  return
end subroutine


subroutine assign_mult_sqrt_corr(nsqrt,iout,tol,solver_iter,rho,target_quark,wy,y,u)
!=======================================================================
! multiply matrix square root of W^(-1) using Taylor expansion
!
!  wy = Z y,
!
!  Z \sim W^(-1/2),
!
!  W^(-1) = rho + (1-rho) D'^(-1) = 1 - X,
!
!       X = (1-rho)(1 - D'^(-1))
!
!  rho = kappa/kappa'
!
!  kappa  : config kappa
!  kappa' : target kappa
!
!        nsqrt : order of Taylor expansion
!          tol : solver tolerance for D'^{-1}
!  solver_iter : solver iteration counter for D'^{-1}
!         iout : solver logfile unit
!          rho : config_kappa/target_kappa
! target_quark : target quark parameter
!
!=======================================================================
  implicit none
  integer,  intent(in)    :: nsqrt,iout
  real(DP), intent(in)    :: tol
  type(counter), intent(inout) :: solver_iter
  real(DP), intent(in)    :: rho
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: target_quark
#else
  type(quark_wilson), intent(inout) :: target_quark
#endif
  type(field_quark_wg),  intent(inout) :: wy
  type(field_quark_wg),  intent(in)    ::  y
  type(vfield_gluon_wg), intent(in)    ::  u
  type(field_quark_wg), allocatable :: v,w,s
  real(DP) :: rtmp,coef
  integer :: iiter
  integer :: isqrt,iter,iaa,ibb

  allocate(v,w,s)
  call new(wy)
  call new(v)
  call new(w)
  call new(s)

  call assign(wy,y)  ! wy = y

  do isqrt=nsqrt,1,-1

    !-------------
    ! w = D' \ wy
    !-------------
    iiter = 0
    call assign_inv_mult_wd(target_quark,iout,tol,iiter,w,wy,u)
    call inc(solver_iter,iiter)

    !
    ! wy = y + coef*(1-rho)*( wy - D'\ wy )
    !      y + coef*(1-rho)*( wy - w )
    !
    ibb = 2*isqrt
    iaa = ibb - 3
    coef = real(iaa,kind=KIND(coef))/ibb
    rtmp = coef*(1.0_DP-rho)
    call accum_sub(wy,w)            ! wy = wy - w
    call accum_mult_add(wy,rtmp,y)  ! wy = wy * rtmp + y

  enddo

  deallocate(v,w,s)

  return
end subroutine


subroutine save_reweighting(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(reweighting_obj), intent(inout) :: this
  integer :: iout
  character(len=CHARLEN) :: fname
  real(DP) :: config_kappa,config_csw
  real(DP) :: target_kappa,target_csw

  call tic(this%io_time)

  !=====================================
  ! Save Random Seed
  !=====================================
  fname=fname_cont(this%run,this%seed_path,this%seed_file)
  iout = search_free_file_unit()
  open(iout,file=fname,status='unknown',form='unformatted',action='write')
  call save(g_rand,iout)
  close(iout)

  config_kappa = get_kappa(this%config_quark)
  target_kappa = get_kappa(this%target_quark)
#ifdef _CLOVER
  config_csw = get_csw(this%config_quark)
  target_csw = get_csw(this%target_quark)
#endif

  !=====================================
  ! Save Input Parameters
  !=====================================
  if (nodeid==0) then
    open(iout,file=this%input_fname,status='unknown',form='formatted')
#ifdef _CLOVER
    write(iout,*)target_kappa,target_csw
    write(iout,*)config_kappa,config_csw
#else
    write(iout,*)target_kappa
    write(iout,*)config_kappa
#endif
    write(iout,*)this%nf
    write(iout,*)this%tol

    write(iout,'(A)')TRIM(this%config_path)
    write(iout,'(A)')TRIM(this%config_file)
    write(iout,'(A)')TRIM(this%seed_path)
    write(iout,'(A)')TRIM(this%seed_file)

    write(iout,*)this%stout_alpha,this%stout_nstep

    write(iout,*)this%run
    write(iout,*)this%traj_start,this%traj_end,this%traj_skip
    write(iout,*)this%random_number_seed
    write(iout,*)this%nnoise
    write(iout,*)this%nbreak
    close(iout)
  endif
  call toc(this%io_time)

  return
end subroutine

subroutine print_stat_reweighting(this)
  implicit none
  type(reweighting_obj), intent(inout) :: this
  real(DP) :: run_time,io_time,tot_time,iter_ave

  iter_ave = get_average(this%solver_iter)

  run_time = get_elapse(this%run_time)
   io_time = get_elapse(this%io_time)
  call toc(this%tot_time)
  tot_time = get_elapse(this%tot_time)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" Total Time :",F24.15)')tot_time
    write(*,'("    IO Time :",F24.15," (",F12.6," %)")')  io_time,100* io_time/tot_time
    write(*,'("   RUN Time :",F24.15," (",F12.6," %)")') run_time,100*run_time/tot_time
    write(*,'("Solver Iter :",F14.7)')iter_ave
    write(*,'(80("="))')
  endif

  return
end subroutine

!--- Ukita ------
! improved subtraction to avoid loss of significant digit on large lattice
!--------------------------------------------------------------------------
#include "abs2_sub.h90"


#include "read_config.h90"

end module
